//@***h* BDBP/CalculateMonthlyMeans.dpr
// 
// NAME
//   CalculateMonthlyMeans.dpr
//
// PURPOSE
//  Calculates monthly mean ozone values out of several different ozone sources that 
//  are taken from the BDBP Database.
//
// DESCRIPTION
//  The program calculates monthly mean ozone values out of several different ozone 
//  sources that are taken from the BDBP Database, from 1978 to 2006. Minimum data 
//  points to calculate a monthly mean is set to 6, and the width of the latitude bands 
//  is set to 5 degrees.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


program CalculateMonthlyMeans;

{$APPTYPE CONSOLE}
{$I UNITDIREC.INC}

{$UNDEF IMPOSE_EQLAT}
{$UNDEF BOREAS}
{$UNDEF DEBUG}
{$UNDEF TRANSFORM}
{$UNDEF FINAL}

uses
  Classes,
  Contnrs,
  DateUtils,
  Dialogs,
  Math,
  SysUtils,
  gb_Array,
  gb_DateTime,
  gb_Files,
  gb_Strings,
  gb_TMeasurementSet, 
  gb_SpatioTemporalBiasCorrector;

const
  {Options for Method are 'Quartile' or '3Sigma'}
  Method='3Sigma';
  StartYear=1978;
  EndYear=2006;
  DeltaLat=5;
  MinEntries=6;
  DivNumber=4;
  VersionStr='1.1.0.6';
  KConst=1.380662e-23;
  SetDesc='LongPatched';

  InputDir='T:\Archive\BDBP\BDBP_V'+VersionStr;
  PathForDailyData='B:\Archive\NIWACombinedTCO\Version2.7\Daily\'+SetDesc+'\T4DSynoptic';
  PathForMonthlyData='B:\Archive\NIWACombinedTCO\Version2.7\Monthly\'+SetDesc+'\TSynopticGlobal';

  {$IFDEF IMPOSE_EQLAT}
//  BaseOutputDir='D:\projects\BDBP\MonthlyMeans\V'+VersionStr+'\Eq';
  {$ELSE}
  BaseOutputDir='B:\Archive\BDBPMonthlyMeans\V'+VersionStr;
  {$ENDIF}

type
  TAltMeasSet=record
    AltIndex:byte;
    DateTime:TDateTime;
    Latitude,
    Longitude,
    OzoneError,
    Ozone:single;
  end;
  PAltMeasSet=^TAltMeasSet;

  TMeasRec=record
    Value,
    Weight,
    CosWeight:single;
    Ozonesonde:boolean;
  end;
  TMeasArray=array of TMeasRec;
  PMeasRec=^TMeasRec;

var
  DataStore:array[1..70,StartYear..EndYear,1..12] of TMeasArray;
  MonthlyMeans:array[1..70,StartYear..EndYear,1..12] of TMeasRec;
  DensCount,
  SetCount,
  LatCount:integer;
  DataSet:string;
  BiasCorrector:TSpatioTemporalBiasCorrector;

function CompareByValue(Item1,Item2:Pointer):Integer;
var
  Data1,Data2:PMeasRec;
begin
  Data1:=PMeasRec(Item1);
  Data2:=PMeasRec(Item2);
  Result:=0;
  if Data1^.Value>Data2^.Value then Result:=1;
  if Data1^.Value<Data2^.Value then Result:=-1;
end;

function Compare(Item1,Item2:pointer):integer;
var
  Hold1,
  Hold2:TAltMeasSet;
begin
  Result:=0;
  Hold1:=PAltMeasSet(Item1)^;
  Hold2:=PAltMeasSet(Item2)^;
  if Hold1.DateTime<Hold2.DateTime then Result:=-1;
  if Hold1.DateTime>Hold2.DateTime then Result:=1;
end;

function TransformOzone(Alt:word;Ozone:double):double;
begin
  {$IFDEF TRANSFORM}
  if ((Alt<=10) or (Alt>=50) or (Ozone>=1.0)) then Result:=Ozone else Result:=ln(Ozone)+1.0;
  {$ELSE}
  Result:=Ozone;
  {$ENDIF}
end;

function InverseTransformOzone(Alt:word;Value:double):double;
begin
  {$IFDEF TRANSFORM}
  if ((Alt<=10) or (Alt>=50) or (Value>=1.0)) then Result:=Value
  else Result:=exp(Value-1.0);
  if Result<1E-5 then Result:=1E-5;
  {$ELSE}
  Result:=Value;
  {$ENDIF}
end;

procedure WriteInfoFile;
var
  Dest:textfile;
begin
  assignfile(Dest,BaseOutputDir+'\Information.txt');
  rewrite(Dest);
  writeln(Dest,'The monthly means listed in these data files were calculated from BDBP');
  writeln(Dest,'version '+VersionStr+'. The method used to excluding outlying data is: '+Method);
  closefile(Dest);
end;

function TestData(Data:TMeasurementSet;Lat1,Lat2:double):boolean;
var
  Year:word;
begin
  Result:=true;
  Year:=YearOf(Data.DateTime);
  if ((Year<StartYear) or (Year>EndYear)) then Result:=false;
  {$IFDEF IMPOSE_EQLAT}
  if Data.EquivalentLatitude>Lat1 then Result:=false;
  if Data.EquivalentLatitude<Lat2 then Result:=false;
  {$ELSE}
  if Data.Latitude>Lat1 then Result:=false;
  if Data.Latitude<Lat2 then Result:=false;
  {$ENDIF}
end;

procedure FillTheDataArray;
var
  Ind1,Ind2,
  I,J,K:integer;
  Source:file;
  BiasCorr:double;
  Year,
  Month,
  NumSets,
  Alt:word;
  TempValue,
  PressValue,
  NormFac,
  MixRatioError,
  Lat1,Lat2:double;
  OldFilename,
  Filename:string;
  Data:TMeasurementSet;
  DataPtr:PAltMeasSet;
  MeasurementSetList:TList;

  {$IFDEF DEBUG}
  Dest:textfile;
  Num:word;
  {$ENDIF}
begin
  Lat1:=90-(LatCount-1)*DeltaLat;
  Lat2:=90-LatCount*DeltaLat;

  {$IFDEF IMPOSE_EQLAT}
  Ind1:=1;
  Ind2:=90;
  {$ELSE}
  Ind1:=trunc((90-Lat1)/2)+1;
  Ind2:=trunc((90-Lat2)/2)+1;
  if Ind2=91 then Ind2:=90;
  {$ENDIF}
  MeasurementSetList:=TList.Create;
  try
    {$IFDEF DEBUG}
    assignfile(Dest,BaseOutputDir+'\DebugDetail.txt');
    rewrite(Dest);
    Num:=0;
    {$ENDIF}

    {First put the TAltMeasSet objects onto the object list and sort by time.}
    for I:=Ind1 to Ind2 do
    begin
      Filename:=InputDir+'\'+DataSet+'\BDBP_'+DataSet+'_V'+ReplaceChars(VersionStr,'.','')+'_'+inttostrf(I,2,true)+'.bin';
      if not FileExists(Filename) then
      begin
        {This code below is here to rename the V1.1.0.6 files which were named}
        {incorrectly when CombineBDBPs was last run.}
        OldFilename:=InputDir+'\'+DataSet+'\TimeSeries_'+inttostrf(I,2,true)+'.bin';
        if fileexists(OldFilename) then CopyFile(OldFilename,Filename,nil);
      end;
      FileMode:=0;
      assignfile(Source,Filename);
      reset(Source,1);
      if filesize(Source)>0 then
      repeat
        try
          Alt:=0;
          NumSets:=0;
          blockread(Source,Alt,sizeof(word));
          blockread(Source,NumSets,sizeof(word));
          for J:=1 to NumSets do
          begin
            Data:=TMeasurementSet.Create;
            try
              Data.ReadFromFile(Source);
              if TestData(Data,Lat1,Lat2) then
              begin
                getmem(DataPtr,sizeof(TAltMeasSet));

                {Transfer the data from Data to DataPtr here.}
                DataPtr^.AltIndex:=Alt;
                if pos('OZS',Data.Source)>0 then DataPtr^.AltIndex:=Alt+100;
                DataPtr^.DateTime:=Data.DateTime;
                DataPtr^.Latitude:=Data.Latitude;
                DataPtr^.Longitude:=Data.Longitude;
                DataPtr^.Ozone:=Data.GetValueForDesc(Ozone_LowRes);
                if ((pos('SAGE1',Data.Source)>0) and (Alt<18)) then DataPtr^.Ozone:=-1;
                if ((pos('SAGE2',Data.Source)>0) and (Alt<10)) then DataPtr^.Ozone:=-1;
                if ((pos('LIMS',Data.Source)>0) and (Alt<25)) then DataPtr^.Ozone:=-1;
                if pos('OZS',Data.Source)>0 then
                begin
                  NormFac:=Data.GetValueForDesc(SondeCorrFactorNotApplied);
                  if ((NormFac>=0.9) and (NormFac<=1.1)) then DataPtr^.Ozone:=DataPtr^.Ozone*NormFac;
                end;
                if DataPtr^.Ozone>1E-5 then
                begin
                  DataPtr^.OzoneError:=Data.GetErrorForDesc(Ozone_LowRes)/100; {The error is now in percent.}
                  if pos('SAGE1',Data.Source)>0 then DataPtr^.OzoneError:=DataPtr^.OzoneError*10;
                  if ((pos('OZS',Data.Source)>0) or (DensCount=2)) then {Will need pressure and temperature values.}
                  begin
                    if SetCount=1 then PressValue:=Data.GetValueForDesc(Pressure) else PressValue:=1013.25*exp(-Alt/7.0);
                    TempValue:=Data.GetValueForDesc(Temperature);
                  end;
                  if pos('OZS',Data.Source)>0 then
                  begin
                    if ((TempValue>0.0) and (PressValue>0.0)) then
                    begin
                      MixRatioError:=DataPtr^.OzoneError*DataPtr^.Ozone*1E20*KConst*TempValue/PressValue; {Mix ratio in ppm}
                      if MixRatioError<0.004 then DataPtr^.OzoneError:=0.004*PressValue/(DataPtr^.Ozone*1E20*KConst*TempValue);
                    end
                    else DataPtr^.Ozone:=1E-10;
                  end;
                  if DensCount=2 then
                  begin
                    if ((TempValue>0.0) and (PressValue>0.0)) then DataPtr^.Ozone:=DataPtr^.Ozone*1E22*KConst*TempValue/PressValue {Mix ratio in ppm}
                    else DataPtr^.Ozone:=1E-10;
                  end;
                  if ((DataPtr^.Ozone>1E-5) and (DataPtr^.OzoneError<=200.0)) then
                  begin
                    {$IFDEF DEBUG}
                    if ((YearOf(Data.DateTime)=1994) and (MonthOf(Data.DateTime)=1) and (Alt=10)
                      and (Data.Latitude>=20) and (Data.Latitude<=25)) then
                    begin
                      inc(Num);
                      writeln(Dest,'Entry=',Num);
                      Data.AppendToTextFile(Dest);
                      writeln(Dest);
                    end;
                    {$ENDIF}

                    MeasurementSetList.Add(DataPtr)
                  end
                  else freemem(DataPtr,sizeof(TAltMeasSet));
                  if ((MeasurementSetList.Count+1) mod 100000)=0 then write('.');
                end
                else freemem(DataPtr,sizeof(TAltMeasSet));
              end;
            finally
              FreeAndNil(Data);
            end;
          end;
        except
          on E:Exception do
          begin
            writeln('An error occurred at Alt='+inttostr(Alt)+'  NumSets='+
            inttostr(NumSets)+' and the error message was '+E.Message+
            '. Press any key to halt the program');
            readln;
            halt;
          end;
        end;
      until eof(Source);
      closefile(Source);
      FileMode:=2;
    end;
    writeln;
    writeln('Number of measurements available: '+inttostr(MeasurementSetList.Count));
    writeln;

    {$IFDEF DEBUG}
    closefile(Dest);
    {$ENDIF}

    {Now sort by time.}
    MeasurementSetList.Sort(Compare);

    {Now work through the list and transfer the data to the DataStore array.}
    for I:=0 to MeasurementSetList.Count-1 do
    try
      DataPtr:=PAltMeasSet(MeasurementSetList[I]);
      if DataPtr^.AltIndex>100 then Alt:=DataPtr^.AltIndex-100 else Alt:=DataPtr^.AltIndex;
      Year:=YearOf(DataPtr^.DateTime);
      Month:=MonthOf(DataPtr^.DateTime);
      K:=length(DataStore[Alt,Year,Month]);
      setlength(DataStore[Alt,Year,Month],K+1);
      DataStore[Alt,Year,Month][K].Ozonesonde:=(DataPtr^.AltIndex>100);
      try
        BiasCorr:=BiasCorrector.GetCorrection(DataPtr^.DateTime,DataPtr^.Latitude,DataPtr^.Longitude);
      except
        on Exception do BiasCorr:=1.0;
      end;
      DataPtr^.Ozone:=DataPtr^.Ozone/BiasCorr;
      DataStore[Alt,Year,Month][K].Value:=TransformOzone(Alt,DataPtr^.Ozone);
      DataStore[Alt,Year,Month][K].Weight:=1/sqr(DataPtr^.OzoneError*DataPtr^.Ozone*0.01);
      DataStore[Alt,Year,Month][K].CosWeight:=cos(DegToRad(DataPtr^.Latitude))/sqr(DataPtr^.OzoneError*DataPtr^.Ozone*0.01);
      freemem(DataPtr,sizeof(TAltMeasSet));
      MeasurementSetList[I]:=nil;
    except
      on E:Exception do ShowMessage('The error occurred at I='+inttostr(I)+' and the message was: '+E.Message);
    end;
  finally
    MeasurementSetList.Clear;

    {The line below may cause the program to crash.}
    try
      FreeAndNil(MeasurementSetList);
    except
      on Exception do MeasurementSetList:=nil;
    end;
  end;
end;

function CalcMonthMeanQuartile(X:TMeasArray;Alt:word):TMeasRec;
var
  DataList:TList;
  Hold:TMeasRec;
  I,J:integer;
  Fraction,
  TotalWeightXMeas,
  TotalCosWeight,
  TotalWeight:double;
  OzonesondeOnly:boolean;
  AccumWeight:array of double;

  {$IFDEF DEBUG}
  Dest:textfile;
  {$ENDIF}
begin
  OzonesondeOnly:=true;
  for I:=1 to length(X) do
    if not X[I-1].Ozonesonde then OzonesondeOnly:=false;
  if ((OzonesondeOnly and (length(X)>=1)) or (length(X)>=MinEntries)) then
  begin
    DataList:=TList.Create;
    try
      {Sort the entries.}
      for I:=Low(X) to High(X) do DataList.Add(addr(X[I]));
      DataList.Sort(CompareByValue);

      {Fill the accumulated weight values.}
      setlength(AccumWeight,length(X));
      TotalWeight:=0.0;
      for I:=Low(X) to High(X) do
      begin
        TotalWeight:=TotalWeight+TMeasRec(DataList[I]^).Weight;
        if I=Low(X) then AccumWeight[I]:=TMeasRec(DataList[I]^).Weight
        else AccumWeight[I]:=AccumWeight[I-1]+TMeasRec(DataList[I]^).Weight;
      end;

      {Normalize the accumulated weights to 0.0 to 1.0}
      for I:=Low(X) to High(X) do AccumWeight[I]:=AccumWeight[I]/TotalWeight;

      {$IFDEF DEBUG}
      assignfile(Dest,BaseOutputDir+'\Debug.dat');
      rewrite(Dest);
      for I:=High(X) downto Low(X) do
      begin
        Hold:=TMeasRec(DataList.Items[I]^);
        writeln(Dest,Hold.Value:8:6,' ',Hold.Weight:8:2,' ',Hold.CosWeight:8:3,' ',AccumWeight[I]:8:6);
      end;
      writeln(Dest);
      {$ENDIF}

      {Blat the measurement weights in the bottom 0.25.}
      I:=0;
      while AccumWeight[I]<0.25 do
      begin
        TMeasRec(DataList[I]^).CosWeight:=0.0;
        TMeasRec(DataList[I]^).Weight:=0.0;
        inc(I);
      end;

      {Deal with the bit in the cell above 0.25.}
      if I=0 then Fraction:=(AccumWeight[I]-0.25)/AccumWeight[I]
      else Fraction:=(AccumWeight[I]-0.25)/(AccumWeight[I]-AccumWeight[I-1]);
      TMeasRec(DataList[I]^).CosWeight:=TMeasRec(DataList[I]^).CosWeight*Fraction;
      TMeasRec(DataList[I]^).Weight:=TMeasRec(DataList[I]^).Weight*Fraction;

      {Move to above 0.75.}
      while AccumWeight[I]<0.75 do inc(I);

      {Deal with the bit in the cell below 0.75.}
      if I=0 then Fraction:=0.75/AccumWeight[I] else Fraction:=(0.75-AccumWeight[I-1])/(AccumWeight[I]-AccumWeight[I-1]);
      TMeasRec(DataList[I]^).CosWeight:=TMeasRec(DataList[I]^).CosWeight*Fraction;
      TMeasRec(DataList[I]^).Weight:=TMeasRec(DataList[I]^).Weight*Fraction;

      {Blat the measurement weights above I}
      for J:=I+1 to High(X) do
      begin
        TMeasRec(DataList[J]^).CosWeight:=0.0;
        TMeasRec(DataList[J]^).Weight:=0.0;
      end;

      {$IFDEF DEBUG}
      for I:=High(X) downto Low(X) do
      begin
        Hold:=TMeasRec(DataList.Items[I]^);
        writeln(Dest,Hold.Value:8:6,' ',Hold.Weight:8:2,' ',Hold.CosWeight:8:3);
      end;
      writeln(Dest);
      {$ENDIF}

      {Now do the weighted mean as before.}
      TotalWeightXMeas:=0.0;
      TotalCosWeight:=0.0;
      TotalWeight:=0.0;
      for I:=Low(X) to High(X) do
      begin
        Hold:=TMeasRec(DataList.Items[I]^);
        TotalWeightXMeas:=TotalWeightXMeas+(Hold.Value*Hold.CosWeight);
        TotalCosWeight:=TotalCosWeight+Hold.CosWeight;
        TotalWeight:=TotalWeight+Hold.Weight;
      end;
      Result.Value:=TotalWeightXMeas/TotalCosWeight;
      Result.Value:=InverseTransformOzone(Alt,Result.Value);
      Result.Weight:=TotalWeight;
      Result.CosWeight:=TotalCosWeight;

      {$IFDEF DEBUG}
      writeln(Dest,Result.Value:8:6,' ',Result.Weight:8:2,' ',Result.CosWeight:8:3);
      closefile(Dest);
      {$ENDIF}
    finally
      DataList.Free;
    end;
  end
  else
  begin
    Result.Value:=-999.99;
    Result.Weight:=-999.99;
  end;
end;

function CalcMonthMean3Sigma(X:TMeasArray;Alt:word;var NumUsed:integer):TMeasRec;
var
  Data:TDoubleArray;
  OzonesondeOnly:boolean;
  Mean,
  Stdv:extended;
  I:integer;
  TotalWeightXMeas,
  TotalCosWeight,
  TotalWeight:double;

  {$IFDEF DEBUG}
  Dest:textfile;
  {$ENDIF}
begin
  OzonesondeOnly:=true;
  NumUsed:=0;
  for I:=1 to length(X) do if not X[I-1].Ozonesonde then OzonesondeOnly:=false;
  if ((OzonesondeOnly and (length(X)>=1)) or (length(X)>=MinEntries)) then
  begin
    if length(X)=1 then
    begin
      Result.Value:=X[0].Value;
      Result.Weight:=X[0].Weight;
      Result.CosWeight:=X[0].CosWeight;
      Exit;
    end;

    {$IFDEF DEBUG}
    assignfile(Dest,BaseOutputDir+'\Debug.txt');
    rewrite(Dest);
    for I:=0 to length(X)-1 do
    begin
      write(Dest,X[I].Value:8:4,' ',X[I].Weight:8:4,' ',X[I].CosWeight:8:4,' ');
      if X[I].Ozonesonde then writeln(Dest,'Yes') else writeln(Dest,'No');
    end;
    closefile(Dest);
    {$ENDIF}

    {Calculate the mean and standard deviation.}
    setlength(Data,length(X));
    for I:=0 to length(X)-1 do Data[I]:=X[I].Value;
    MeanAndStdDev(Data,Mean,Stdv);

    {Now do the weighted mean as before but exclude all values that are more}
    {than 3 sigma above the mean and less than 3 sigmal below the mean.}
    TotalWeightXMeas:=0.0;
    TotalCosWeight:=0.0;
    TotalWeight:=0.0;
    for I:=Low(X) to High(X) do
    begin
      if ((X[I].Value>=Mean-3*Stdv) and (X[I].Value<=Mean+3*Stdv)) then
      begin
        inc(NumUsed);
        TotalWeightXMeas:=TotalWeightXMeas+(X[I].Value*X[I].CosWeight);
        TotalCosWeight:=TotalCosWeight+X[I].CosWeight;
        TotalWeight:=TotalWeight+X[I].Weight;
      end;
    end;
    if NumUsed>0 then
    begin
      Result.Value:=TotalWeightXMeas/TotalCosWeight;
      Result.Weight:=TotalWeight;
      Result.CosWeight:=TotalCosWeight;
    end
    else
    begin
      Result.Value:=-999.99;
      Result.Weight:=-999.99;
    end;
  end
  else
  begin
    Result.Value:=-999.99;
    Result.Weight:=-999.99;
  end;
end;

procedure CalculateTheMonthlyMeans;
var
  Year,
  Month,
  Alt:word;
  NumTotal,
  NumUsed:integer;
  Filename,
  LatStr1,
  LatStr2,
  OutputDir:string;
  Dest:textfile;
begin
  LatStr1:=inttostrf(abs(90-(LatCount-1)*DeltaLat),2,true);
  if (90-(LatCount-1)*DeltaLat)<0.0 then LatStr1:=LatStr1+'S' else LatStr1:=LatStr1+'N';
  LatStr2:=inttostrf(abs(90-LatCount*DeltaLat),2,true);
  if (90-LatCount*DeltaLat)<0.0 then LatStr2:=LatStr2+'S' else LatStr2:=LatStr2+'N';
  if DensCount=1 then OutputDir:=BaseOutputDir+'\NumDens' else OutputDir:=BaseOutputDir+'\MixRatio';
  OutputDir:=OutputDir+'\'+DataSet+'\Stats';
  if not DirectoryExists(OutputDir) then ForceDirectories(OutputDir);
  for Alt:=1 to 70 do
  begin
    Filename:=OutputDir+'\BDBPMonthMeans_'+LatStr1+'To'+LatStr2+'_'+inttostrf(Alt,2,true)+'.dat';
    assignfile(Dest,Filename);
    rewrite(Dest);
    writeln(Dest,'''This file was created by B:\Contracts\Con00004\Source\CalculateMonthlyMeans\CalculateMonthlyMeans.bdsproj');
    writeln(Dest,'''Column A) Year.');
    writeln(Dest,'''Column B) Month.');
    writeln(Dest,'''Column C) Decimal year.');
    writeln(Dest,'''Column D) Total number of measurements available.');
    writeln(Dest,'''Column E) Number of measurements used.');
    writeln(Dest,'''Column F) Percentage of measurements used.');
    for Year:=StartYear to EndYear do
      for Month:=1 to 12 do
        begin
          {$IFDEF DEBUG}
          writeln('About to calculate monthly means for Year=',Year,' Month=',Month,' Alt=',Alt);
          {$ENDIF}
          if Method='Quartile' then MonthlyMeans[Alt,Year,Month]:=CalcMonthMeanQuartile(DataStore[Alt,Year,Month],Alt);
          if Method='3Sigma' then MonthlyMeans[Alt,Year,Month]:=CalcMonthMean3Sigma(DataStore[Alt,Year,Month],Alt,NumUsed);
          NumTotal:=length(DataStore[Alt,Year,Month]);
          write(Dest,Year:4,', ',Month:2,', ',Year+(Month-0.5)/12:8:3,', ',NumTotal:6,', ',NumUsed:6,', ');
          if NumTotal=0 then writeln(Dest,'        ,') else writeln(Dest,100*NumUsed/NumTotal:8:4);
        end;
  end;      
end;

procedure SaveTheMonthlyMeans;
var
  Filename,
  LatStr1,
  LatStr2,
  OutputDir:string;
  BinDest:file;
  TextDest:textfile;
  Year,
  Month,
  Alt:word;
  Error,
  ErrorCos:double;
begin
  LatStr1:=inttostrf(abs(90-(LatCount-1)*DeltaLat),2,true);
  if (90-(LatCount-1)*DeltaLat)<0.0 then LatStr1:=LatStr1+'S' else LatStr1:=LatStr1+'N';
  LatStr2:=inttostrf(abs(90-LatCount*DeltaLat),2,true);
  if (90-LatCount*DeltaLat)<0.0 then LatStr2:=LatStr2+'S' else LatStr2:=LatStr2+'N';
  if DensCount=1 then OutputDir:=BaseOutputDir+'\NumDens' else OutputDir:=BaseOutputDir+'\MixRatio';
  OutputDir:=OutputDir+'\'+DataSet;

  if not DirectoryExists(OutputDir+'\Bin') then ForceDirectories(OutputDir+'\Bin');
  Filename:=OutputDir+'\Bin\BDBPMonthMeans_'+LatStr1+'To'+LatStr2;
  assignfile(BinDest,Filename+'.bin');
  rewrite(BinDest,1);
  blockwrite(BinDest,MonthlyMeans,sizeof(MonthlyMeans));
  closefile(BinDest);

  if not DirectoryExists(OutputDir+'\Text') then ForceDirectories(OutputDir+'\Text');
  Filename:=OutputDir+'\Text\BDBPMonthMeans_'+LatStr1+'To'+LatStr2;
  assignfile(TextDest,Filename+'.txt');
  rewrite(TextDest);
  writeln(TextDest,'''This file was created by D:\projects\BDBP\Source\CalculateMonthlyMeans\CalculateMonthlyMeans.dpr');
  writeln(TextDest,'''Null data are excluded from this file so that only where data are available are they listed.');
  writeln(TextDest,'''Column A) Year.');
  writeln(TextDest,'''Column B) Month.');
  writeln(TextDest,'''Column C) Decimal Year.');
  if SetCount=1 then writeln(TextDest,'''Column D) Altitude (km).');
  if SetCount=2 then writeln(TextDest,'''Column D) Pressure (hPa).');
  if DensCount=1 then writeln(TextDest,'''Column E) Monthly mean ozone (10^18 molecules/m^3).');
  if DensCount=2 then writeln(TextDest,'''Column E) Monthly mean ozone ppm.');
  writeln(TextDest,'''Column F) Error on the monthly mean.');
  writeln(TextDest,'''Column G) Percentage error on the monthly mean.');
  writeln(TextDest,'''Column H) Error on the monthly mean cosine weighted.');
  writeln(TextDest,'''Column I) Percentage error on the monthly mean cosine weighted.');
  for Year:=StartYear to EndYear do
    for Month:=1 to 12 do
      for Alt:=1 to 70 do
        if MonthlyMeans[Alt,Year,Month].Value>0.0 then
        try
          Error:=1/sqrt(MonthlyMeans[Alt,Year,Month].Weight);
          ErrorCos:=1/sqrt(MonthlyMeans[Alt,Year,Month].CosWeight);
          write(TextDest,Year:4,' ',Month:2,' ',Year+(Month-0.5)/12:8:3,' ');
          if DataSet='LatAlt' then write(TextDest,Alt:2,' ');
          if DataSet='LatPress' then write(TextDest,1013.25*exp(-Alt/7):8:4,' ');
          writeln(TextDest,MonthlyMeans[Alt,Year,Month].Value:8:5,' ',Error:8:5,' ',
          Error*100/MonthlyMeans[Alt,Year,Month].Value:8:4,' ',ErrorCos:8:5,' ',
          ErrorCos*100/MonthlyMeans[Alt,Year,Month].Value:8:4);
        except
          on E:Exception do ShowMessage('The error message was: '+E.Message);
        end;
  closefile(TextDest);
end;

procedure ClearTheDataArray;
var
  Year,
  Month,
  Alt:word;
begin
  for Year:=StartYear to EndYear do
    for Month:=1 to 12 do
      for Alt:=1 to 70 do
        setlength(DataStore[Alt,Year,Month],0);
end;

begin
  if not DirectoryExists(BaseOutputDir) then ForceDirectories(BaseOutputDir);
  WriteInfoFile;
  BiasCorrector:=TSpatioTemporalBiasCorrector.Create(PathForDailyData,PathForMonthlyData,SetDesc,5);
  try
    for DensCount:=1 to 2 do {DensCount=1 is number density, DensCount=2 is mixing ratio.}
      for SetCount:=1 to 2 do {SetCount=1 is LatAlt, SetCount=2 is LatPress.}
      begin
        if SetCount=1 then DataSet:='LatAlt' else DataSet:='LatPress';
        for LatCount:=1 to (180 div DeltaLat) do
        begin
          writeln('DensCount=',DensCount,' SetCount=',SetCount,' LatCount=',LatCount,' out of ',(180 div DeltaLat));
          FillTheDataArray;
          CalculateTheMonthlyMeans;
          SaveTheMonthlyMeans;
          ClearTheDataArray;
        end;
      end;
  finally
    FreeAndNil(BiasCorrector);
  end;
end.
