//@***h* BDBP/gb_DelphiNetCDF.pas
// 
// NAME
//   gb_DelphiNetCDF.pas
//
// PURPOSE
//  Handles all actions for writing to and reading from netCDF files.
//
// DESCRIPTION
//  This unit handles all actions for writing to and reading from netCDF files.
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_DelphiNetCDF;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  Contnrs,
  Sysutils,
  DelphiNetCDF,
  gb_array;

type
  TConvertStrToDateTime=function(DateTimeStr:string):TDateTime;
  TConvertValueToDateTime=function(Value:double):TDateTime;
  TTransformLevel=function(Value:double):double;

  TDelphiNetCDF=class
  strict private
    FListOfObjectsToFree:TObjectList;
    FNetCDFLevelIDStr,
    FNetCDFTimeIDStr,
    FNetCDFVarStr,
    FFilename:string;
    FRead:boolean;
    FIntID:integer;
    FFunctionToConvertStrToDateTime:TConvertStrToDateTime;
    FFunctionToConvertValueToDateTime:TConvertValueToDateTime;
  public
    constructor Create(AFilename:string);
    function DefineDimension(DimName:string;DimSize:longint):integer;
    function DefineFloatVariable(VarName:string;Dims:string=''):integer;

    function GetListOfDimensions:TStringList;
    function GetListOfDimIndicesForVar(VarName:string):TByteArray;
    function GetListOfVariables:TStringList;
    function GetVarIndex(VarName:string):integer;
    function GetDimIndex(DimName:string):integer;
    function GetDimensionLength(DimIndex:integer):integer;
    function GetDatesAndTimes(TimeDimIndex,TimeVarIndex:integer):TDoubleArray;
    function GetDimensionArray(ADimIndex,AVarIndex:integer):TDoubleArray;

    function GetDataArray(VarStr:string;StartPosition,CountsPerDim:TIntegerArray):TDoubleArray;
    procedure PutVariableFloatArray(VarStr:string;StartPosition,CountsPerDim:TIntegerArray;Data:TSingleArray);

    procedure PutGlobalTextAttribute(Name,Text:string);
    procedure PutVariableTextAttribute(VarName,Name,Text:string);
    procedure PutVariableFloatAttribute(VarName,Name:string;Value:single);
    procedure PutVariableFloat(VarName:string;Data:TSingleArray);
    procedure SwitchToDataMode;
    destructor Destroy;override;

    property FunctionToConvertStrToDateTime:TConvertStrToDateTime write FFunctionToConvertStrToDateTime;
    property FunctionToConvertValueToDateTime:TConvertValueToDateTime write FFunctionToConvertValueToDateTime;
    property NetCDFLevelIDStr:string write FNetCDFLevelIDStr;
    property NetCDFTimeIDStr:string write FNetCDFTimeIDStr;
    property NetCDFVarStr:string write FNetCDFVarStr;
    property OpenRead:boolean read FRead;
  end;

  EDelphiNetCDF=class(exception);

implementation

uses
  gb_strings;

constructor TDelphiNetCDF.Create(AFilename:string);
var
  Test:integer;
begin
  FRead:=FileExists(AFilename);
  if (FRead and (nc_open(PChar(AFilename),NC_NOWRITE,FIntID)<>0)) then raise Exception.Create('Something is wrong here');
//  if (not FRead and (nc_create(PChar(AFilename),NC_WRITE,FIntID)<>0)) then fail;
  if not FRead then
  begin
    Test:=nc_create(PChar(AFilename),NC_WRITE,FIntID);
    if Test<>0 then raise Exception.Create('Something is wrong here');
  end;
  FFilename:=AFilename;
  FListOfObjectsToFree:=TObjectList.Create;
  FListOfObjectsToFree.OwnsObjects:=true;
end;

function TDelphiNetCDF.DefineDimension(DimName:string;DimSize:longint):integer;
begin
  if nc_def_dim(FIntID,PChar(DimName),DimSize,Result)<>0 then
    raise EDelphiNetCDF.Create('the call to nc_def_dim failed.');
end;

//function TDelphiNetCDF.DefineVariable(DimName:string;DimType:nc_type;DimIDs:TIntegerArray):integer;
//var
//  Hold:TVarDimArray;
//  I:integer;
//begin
//  for I:=1 to length(DimIDs) do Hold[I]:=DimIDs[I-1];
//  if nc_def_var(FIntID,PChar(DimName),DimType,length(DimIDs),Hold,Result)<>0 then
//    raise EDelphiNetCDF.Create('the call to nc_def_var failed.');
//end;

function TDelphiNetCDF.DefineFloatVariable(VarName:string;Dims:string=''):integer;
var
  Hold:TVarDimArray;
  NumDims,
  I:integer;
begin
  if Dims='' then Dims:=VarName;
  NumDims:=WordCount(Dims,',');
  for I:=1 to NumDims do nc_inq_dimid(FIntID,PChar(GetWord(Dims,',',I)),Hold[I]);
  if nc_def_var(FIntID,PChar(VarName),NC_FLOAT,NumDims,Hold,Result)<>0 then
    raise EDelphiNetCDF.Create('the call to nc_def_var failed.');
end;

function TDelphiNetCDF.GetListOfDimensions:TStringList;
var
  NumDimensions,
  I:integer;
  DimName:array[0..NC_MAX_NAME] of char;
begin
  Result:=TStringList.Create;
  FListOfObjectsToFree.Add(Result);
  if nc_inq_ndims(FIntID,NumDimensions)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_ndims failed.');
  for I:=0 to NumDimensions-1 do
  begin
    if nc_inq_dimname(FIntID,I,DimName)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_dimname failed.');
    Result.Add(DimName);
  end;
end;

function TDelphiNetCDF.GetListOfDimIndicesForVar(VarName:string):TByteArray;
var
  NumDims,
  DataIndex,
  I:integer;
  DimArray:TVarDimArray;
begin
  if nc_inq_varid(FIntID,PChar(VarName),DataIndex)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_varid failed.');
  if nc_inq_varndims(FIntID,DataIndex,NumDims)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_varndims failed.');
  setlength(Result,NumDims);
  if nc_inq_vardimid(FIntID,DataIndex,pinteger(DimArray[1]))<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_vardimid failed.');
  for I:=1 to NumDims do Result[I-1]:=DimArray[I];
end;

function TDelphiNetCDF.GetListOfVariables:TStringList;
var
  NumVars,
  I:integer;
  VarName:array[0..NC_MAX_NAME] of char;
begin
  Result:=TStringList.Create;
  FListOfObjectsToFree.Add(Result);
  if nc_inq_nvars(FIntID,NumVars)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_nvars failed.');
  for I:=0 to NumVars-1 do
  begin
    if nc_inq_varname(FIntID,I,VarName)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_varname failed.');
    Result.Add(VarName);
  end;
end;

function TDelphiNetCDF.GetVarIndex(VarName:string):integer;
begin
  if nc_inq_varid(FIntID,PChar(VarName),Result)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_varid failed.');
end;

function TDelphiNetCDF.GetDimIndex(DimName:string):integer;
begin
  if nc_inq_dimid(FIntID,PChar(DimName),Result)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_dimid failed.');
end;

function TDelphiNetCDF.GetDimensionLength(DimIndex:integer):integer;
begin
  if nc_inq_dimlen(FIntId,DimIndex,Result)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_dimlen failed.');
end;

function TDelphiNetCDF.GetDatesAndTimes(TimeDimIndex,TimeVarIndex:integer):TDoubleArray;
var
  VarType:nc_type;
  DoubleDataArray:PNetCDFDoubleArray;
  SingleDataArray:PNetCDFSingleArray;
  IntegerDataArray:PVarIndexArray;
  TimeDim,
  DimIndex,
  VarIndex,
  I:integer;
begin
  {We need to do this otherwise the parameters passed get screwed up in the nc_ calls.}
  DimIndex:=TimeDimIndex;
  VarIndex:=TimeVarIndex;

  if nc_inq_vartype(FIntID,VarIndex,VarType)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_vartype failed.');
  if nc_inq_dimlen(FIntId,DimIndex,TimeDim)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_dimlen failed.');
  setlength(Result,TimeDim);
  case VarType of
    NC_NAT:raise EDelphiNetCDF.Create('the date and time variable cannot be of no type.');
    NC_CHAR:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_BYTE:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_SHORT:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_INT:begin
      getmem(IntegerDataArray,TimeDim*sizeof(integer));
      nc_get_var_int(FIntId,VarIndex,IntegerDataArray);
      for I:=1 to TimeDim do Result[I-1]:=FFunctionToConvertValueToDateTime(IntegerDataArray^[I]);
      freemem(IntegerDataArray,TimeDim*sizeof(integer));
    end;
    NC_FLOAT:begin
      getmem(SingleDataArray,TimeDim*sizeof(single));
      nc_get_var_float(FIntId,VarIndex,SingleDataArray);
      for I:=1 to TimeDim do Result[I-1]:=FFunctionToConvertValueToDateTime(SingleDataArray^[I]);
      freemem(SingleDataArray,TimeDim*sizeof(single));
    end;
    NC_DOUBLE:begin
      getmem(DoubleDataArray,TimeDim*sizeof(double));
      nc_get_var_double(FIntId,VarIndex,DoubleDataArray);
      for I:=1 to TimeDim do Result[I-1]:=FFunctionToConvertValueToDateTime(DoubleDataArray^[I]);
      freemem(DoubleDataArray,TimeDim*sizeof(double));
    end;
  end;
end; {End of TDelphiNetCDF.GetDatesAndTimes.}

function TDelphiNetCDF.GetDimensionArray(ADimIndex,AVarIndex:integer):TDoubleArray;
var
  VarType:nc_type;
  DoubleDataArray:PNetCDFDoubleArray;
  SingleDataArray:PNetCDFSingleArray;
  LevelDim,
  DimIndex,
  VarIndex,
  I:integer;
begin
  {We need to do this otherwise the parameters passed get screwed up in the nc_ calls.}
  DimIndex:=ADimIndex;
  VarIndex:=AVarIndex;

  if nc_inq_vartype(FIntID,VarIndex,VarType)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_vartype failed.');
  if nc_inq_dimlen(FIntId,DimIndex,LevelDim)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_dimlen failed.');
  setlength(Result,LevelDim);
  case VarType of
    NC_NAT:raise EDelphiNetCDF.Create('the level variable cannot be of no type.');
    NC_CHAR:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_BYTE:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_SHORT:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_INT:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_FLOAT:begin
      getmem(SingleDataArray,LevelDim*sizeof(single));
      nc_get_var_float(FIntID,VarIndex,SingleDataArray);
      for I:=1 to LevelDim do Result[I-1]:=SingleDataArray^[I];
      freemem(SingleDataArray,LevelDim*sizeof(single));
    end;
    NC_DOUBLE:begin
      getmem(DoubleDataArray,LevelDim*sizeof(double));
      nc_get_var_double(FIntID,VarIndex,DoubleDataArray);
      for I:=1 to LevelDim do Result[I-1]:=DoubleDataArray^[I];
      freemem(DoubleDataArray,LevelDim*sizeof(double));
    end;
  end;
end; {End of TDelphiNetCDF.GetLevels.}

function TDelphiNetCDF.GetDataArray(VarStr:string;StartPosition,CountsPerDim:TIntegerArray):TDoubleArray;
var
  StartPos,
  Counts:PVarDimArray;
  DataIndex,
  DataLen,
  I:integer;
  VarType:nc_type;
  FloatData:PNetCDFSingleArray;
  DoubleData:PNetCDFDoubleArray;
begin
  nc_inq_varid(FIntID,PChar(VarStr),DataIndex);
  getmem(StartPos,length(StartPosition)*sizeof(integer));
  for I:=1 to length(StartPosition) do StartPos^[I]:=StartPosition[I-1];
  getmem(Counts,length(CountsPerDim)*sizeof(integer));
  DataLen:=1;
  for I:=1 to length(CountsPerDim) do
  begin
    Counts^[I]:=CountsPerDim[I-1];
    DataLen:=DataLen*CountsPerDim[I-1];
  end;
  setlength(Result,DataLen);
  if nc_inq_vartype(FIntID,DataIndex,VarType)<>0 then raise EDelphiNetCDF.Create('the call to nc_inq_vartype failed.');
  case VarType of
    NC_NAT:raise EDelphiNetCDF.Create('the variable type cannot be of no type.');
    NC_CHAR:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_BYTE:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_SHORT:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_INT:raise EDelphiNetCDF.Create('haven''t written the code yet.');
    NC_FLOAT:begin
      getmem(FloatData,DataLen*sizeof(single));
      if nc_get_vara_float(FIntId,DataIndex,StartPos,Counts,FloatData)<>0 then
        raise EDelphiNetCDF.Create('the call to nc_get_vara_float failed.');
      for I:=1 to DataLen do Result[I-1]:=FloatData^[I];
      freemem(FloatData,DataLen*sizeof(single));
    end;
    NC_DOUBLE:begin
      getmem(DoubleData,DataLen*sizeof(double));
      if nc_get_vara_double(FIntId,DataIndex,StartPos,Counts,DoubleData)<>0 then
        raise EDelphiNetCDF.Create('the call to nc_get_vara_float failed.');
      for I:=1 to DataLen do Result[I-1]:=DoubleData^[I];
      freemem(DoubleData,DataLen*sizeof(double));
    end;
  end;
  freemem(StartPos,length(StartPosition)*sizeof(integer));
  freemem(Counts,length(CountsPerDim)*sizeof(integer));
end;

procedure TDelphiNetCDF.PutVariableFloatArray(VarStr:string;StartPosition,CountsPerDim:TIntegerArray;Data:TSingleArray);
var
  StartPos,
  Counts:PVarDimArray;
  FloatData:PNetCDFSingleArray;
  DataIndex,
  I:integer;
begin
  nc_inq_varid(FIntID,PChar(VarStr),DataIndex);
  getmem(StartPos,length(StartPosition)*sizeof(integer));
  for I:=1 to length(StartPosition) do StartPos^[I]:=StartPosition[I-1];
  getmem(Counts,length(CountsPerDim)*sizeof(integer));
  for I:=1 to length(CountsPerDim) do Counts^[I]:=CountsPerDim[I-1];
  getmem(FloatData,length(Data)*sizeof(single));
  for I:=1 to length(Data) do FloatData^[I]:=Data[I-1];
  if nc_put_vara_float(FIntId,DataIndex,StartPos,Counts,FloatData)<>0 then
    raise EDelphiNetCDF.Create('the call to nc_put_vara_float failed.');
  freemem(StartPos,length(StartPosition)*sizeof(integer));
  freemem(Counts,length(CountsPerDim)*sizeof(integer));
end;

procedure TDelphiNetCDF.PutGlobalTextAttribute(Name,Text:string);
begin
  nc_put_att_text(FIntId,NC_GLOBAL,PChar(Name),length(Text),PChar(Text));
end;

procedure TDelphiNetCDF.PutVariableTextAttribute(VarName,Name,Text:string);
var
  VarID:integer;
begin
  nc_inq_varid(FIntID,PChar(VarName),VarID);
  nc_put_att_text(FIntId,VarID,PChar(Name),length(Text),PChar(Text));
end;

procedure TDelphiNetCDF.PutVariableFloatAttribute(VarName,Name:string;Value:single);
var
  VarID:integer;
begin
  nc_inq_varid(FIntID,PChar(VarName),VarID);
  nc_put_att_float(FIntId,VarID,PChar(Name),NC_FLOAT,1,Value);
end;

procedure TDelphiNetCDF.PutVariableFloat(VarName:string;Data:TSingleArray);
var
  VarID,
  I:integer;
  Hold:PNetCDFSingleArray;
begin
  nc_inq_varid(FIntID,PChar(VarName),VarID);
  getmem(Hold,length(Data)*sizeof(single));
  for I:=1 to length(Data) do Hold^[I]:=Data[I-1];
  if nc_put_var_float(FIntID,VarID,Hold)<>0 then raise EDelphiNetCDF.Create('the call to nc_put_var_float failed.');
  freemem(Hold,length(Data)*sizeof(single));
end;

procedure TDelphiNetCDF.SwitchToDataMode;
begin
  nc_enddef(FIntID);
end;

destructor TDelphiNetCDF.Destroy;
begin
  nc_close(FIntId);
  FreeAndNil(FListOfObjectsToFree);
  inherited Destroy;
end;

end.
