//@***h* BDBP/gb_EESCServer.pas
// 
// NAME
//   gb_EESCServer.pas
//
// PURPOSE
//  Deals with all aspects of providing a flexible EESC basis function for a multiple
//  linear regression model.
//
// DESCRIPTION
//  This unit deals with all aspects of providing a flexible EESC basis function for a 
//  multiple linear regression model.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_EESCServer;

{$I UNITDIREC.INC}

interface

uses
  SysUtils,
  gb_4Dsynoptic,
  gb_strings,
  gb_xysignal;

type
  TEESCServer=class
  private
    EESCTimeSeries:array[1..7] of TXYSignal;
    AgeOfAirData:T4DSynoptic;
  public
    constructor Create(DataFileDir:string;ClyOnly:boolean=false);
    destructor Destroy;override;
    function GetEESC(AgeOfAir,DecYear:double):double;overload;
    function GetEESC(Latitude,Altitude,DecYear:double):double;overload;
  end;

implementation

constructor TEESCServer.Create(DataFileDir:string;ClyOnly:boolean=false);
var
  I,J:integer;
  Filename,
  Linein:string;
  Source:textfile;
  Year,
  EESC:double;
begin
  {First load the EESC time series.}
  for I:=1 to 7 do
  begin
    EESCTimeSeries[I]:=TXYSignal.Create(-999.99,-999.99); {These are freed in the destructor.}
    Filename:=DataFileDir+'\EESC_Age'+floattostrf(I,ffFixed,3,1)+'_Width'+floattostrf(I/2,ffFixed,3,1);
    if ClyOnly then Filename:=Filename+'_ClyOnly.dat' else Filename:=Filename+'.dat';
    if not FileExists(Filename) then raise Exception.Create('the file named '+Filename+' couldn''t be found.');
    assignfile(Source,Filename);
    reset(Source);
    for J:=1 to 8 do readln(Source); {This skips the header.}
    repeat
      readln(Source,Linein);
      if Linein<>'' then
      begin
        Year:=strtofloat(GetWord(Linein,' ',1));
        EESC:=strtofloat(GetWord(Linein,' ',2))/1000;
        EESCTimeSeries[I].AddCoord(Year,EESC);
      end;
    until eof(Source);
    closefile(Source);
  end;

  {Now load the age of air data.}
  Filename:=DataFileDir+'\AgeOfAirData.4Dsyn';
  if not FileExists(Filename) then raise Exception.Create('the file named '+Filename+' couldn''t be found.');
  AgeOfAirData:=T4DSynoptic.Create;
  AgeOfAirData.ExtendToPole:=true;
  AgeOfAirData.ReadFromFile(Filename); {This is freed in the destructor.}
end;

destructor TEESCServer.Destroy;
var
  I:integer;
begin
  for I:=1 to 7 do FreeAndNil(EESCTimeSeries[I]);
  FreeAndNil(AgeOfAirData);
  inherited Destroy;
end;

function TEESCServer.GetEESC(AgeOfAir,DecYear:double):double;
var
  Ind1,Ind2:byte;
begin
  if ((DecYear<1950) or (DecYear>2100)) then raise Exception.Create('the DecYear value was out of range.');
  if AgeOfAir<1.0 then AgeOfAir:=1.0;
  if AgeOfAir>7.0 then AgeOfAir:=7.0;
  Ind1:=trunc(AgeOfAir);
  Ind2:=Ind1+1;
  if Ind2=8 then Ind2:=7;
  Result:=frac(AgeOfAir)*EESCTimeSeries[Ind2].InterpolateYValue(DecYear,false,false)+
    (1-frac(AgeOfAir))*EESCTimeSeries[Ind1].InterpolateYValue(DecYear,false,false);
end;

function TEESCServer.GetEESC(Latitude,Altitude,DecYear:double):double;
var
  AgeOfAir:double;
begin
  if ((Latitude<-90) or (Latitude>90)) then raise Exception.Create('the latitude was out of range.');
  if ((Altitude<10) or (Altitude>70)) then raise Exception.Create('the altitude was out of range.');
  AgeOfAir:=AgeOfAirData.GetValueByCoords(0.0,0.0,Altitude,Latitude,true)+
    AgeOfAirData.GetValueByCoords(0.0,trunc(frac(DecYear)*12)+1,Altitude,Latitude,true);
  Result:=GetEESC(AgeOfAir,DecYear);
end;

end.
