//@***h* BDBP/gb_synopticglobal.pas
// 
// NAME
//   gb_synopticglobal.pas
//
// PURPOSE
//  Defines a class which stores global synoptic data over a user 
//  specified area and at user specified latitude and longitude resolution.  
//
// DESCRIPTION
//  This unit defines a class which stores global synoptic data over a user 
//  specified area and at user specified latitude and longitude resolution. The 
//  user specificies whether the data are stored at word precision (2 bytes), 
//  in which case prescribed scaling and offset values are used to convert the 
//  word values to single precision values, or at single (4 byte) precision.   
//  
//
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_synopticglobal;

{$I UNITDIREC.INC}

interface

uses
  {This is a standard delphi unit}
  gb_array,
  SysUtils;

const
  NullValue=1E30;
  StrID='TSynopticGlobal';

type
  //needed for gb_polarplot
  TLatLongToXYTransformData=record
    minLatitude,
    longitudeOffset,
    numDeg,
    scale:double;
    height,
    width,radius:integer;
    hemi:string;
  end;

  TVertex=record
    x,y:double;
  end;

  TInterpolationFailPlan=(raiseexception,returnnull);
  TConvertLatLongToXY = function(lat, long:single; latLong : boolean;Level:byte; transformData:TLatLongToXYTransformData):TVertex;

  TCornerRec=record
    V:array[1..4] of double;
    X1,X2,Y1,Y2:integer;
    DLong,DLat:double;
  end;

  TSynopticGlobalHeader=record
    NumLong,
    NumLat:cardinal;
    TagValue, {Kept for backward compatability but never used.}
    DeltaLong,
    DeltaLat,
    StartLong,
    StartLat:double;
  end;

  TSynopticGlobal=class
  strict private
    MaxMemory:cardinal;
    FMaxLongInterpol, {The maximum number of degrees over which longitudinal interpolation must be performed.}
    FFracMissing,
    FScaling,
    FOffset:double;
    FExtendToPole,
    FAutoRescale:boolean;
    FInterpolationFailPlan:TInterpolationFailPlan;
    Header:TSynopticGlobalHeader;
    FData:TLargeArray;
    FNonUniformLats:boolean;
    FLatitudes:TDoubleArray;
    FDateTimeStamp:TDateTime;
    FDataLevel:double;
    FFunctionToUse:TConvertLatLongToXY;

    procedure CompressionOn(MaxValue,MinValue:double);
    function GetCornerValues(Latitude,Longitude:double):TCornerRec;
    procedure SetDefaultValues;
    procedure TransformLongitude(var Longitude:double);
    procedure TransformLatitude(var Latitude:double);
  public
    {Constructors}
    constructor Create;overload;
    constructor Create(Source:TSynopticGlobal);overload;
    constructor Create(HeaderData:TSynopticGlobalHeader;MaxMem:cardinal=high(cardinal));overload;
    constructor Create(Filename:string;MaxMem:cardinal=high(cardinal));overload;
    constructor Create(var Source:file;MaxMem:cardinal=high(cardinal));overload;

    {Other methods in alphabetical order.}
    function  AllNull:boolean;
    function  BiLinear(Latitude,Longitude:double):single;
    procedure Clear;
    procedure FillWithZero;
    procedure DeepCopy(Source:TSynopticGlobal);
    function  DistWeighting(Latitude,Longitude:double;Order:byte):single;
    function  GetCellArea(LatIndex:cardinal):single;
    function  GetHeader:TSynopticGlobalHeader;
    function  GetLatitude(Index:cardinal):double;
    function  GetMaximumValue:single;
    procedure GetMemoryStats(var TDisk:boolean;var TNumPages:cardinal);
    function  GetMemoryUsed:cardinal;
    function  GetMinimumValue:single;
    function  GetValue(LongIndex,LatIndex:cardinal):single;
    function  GetValueByCoords(ALongitude,ALatitude:double;interpolate:boolean):single;
    function  GetZonalMean(LatIndex:cardinal):single;
    function  IsNotNull(LongIndex,LatIndex:cardinal):boolean;
    function  IsNull(LongIndex,LatIndex:cardinal):boolean;
    function  IsNullByCoords(ALongitude,ALatitude:double):boolean;
    function  GetLongitude(Index:cardinal):double;
    function  EndLatitude:double;
    function  EndLongitude:double;
    procedure FillWithDataFrom(Source:TSynopticGlobal);
    procedure FillWithLongitudeInterpolation;
    function  GetLatitudes:TDoubleArray;
    function  GetTotalTimesArea(Scaling:double):single;
    function  GetAreaWeightedMean(LatSouth,LatNorth,LongWest,LongEast:double):double;
    function  GetMeanAtLatIndex(LatIndex:cardinal):double;
    function  Missing:double;overload;
    function  Missing(StartLatInd,EndLatInd:cardinal):double;overload;
    procedure PutValue(LongIndex,LatIndex:cardinal;Value:single);
    procedure PutValueByCoords(ALongitude,ALatitude:double;Value:single);
    function  ReadCompressed:boolean;
    procedure ReadDataGridFromFile(var Source:file);overload;
    procedure ReadDataGridFromFile(Filename:string);overload;
    procedure ReadFromFile(var Source:file);
    procedure Rescale(Value:single);
    function  SameGridStructure(Source:TSynopticGlobal):boolean;
    procedure Scale(X:single);
    procedure SetScalingAndOffset(Scaling,Offset:double);
    procedure SpatialSmooth(Weights:T2DDoubleArray);
    procedure SpecifyNonuniformLatitudes(Latitudes:TDoubleArray);
    procedure SwitchCompressionOff;
    procedure SwitchCompressionOn;overload;
    procedure SwitchCompressionOn(MaxValue,MinValue:double);overload;
    procedure WriteCompressed(Value:boolean);
    procedure WriteDataGridToFile(var Dest:file);
    procedure WriteDataToTextFile(Filename:string);
    procedure WriteToFile(var Dest:file);overload;
    procedure WriteToFile(Filename:string);overload;

    {Destructor}
    destructor Destroy;override;

    {Properties}
    property AutoRescale:boolean read FAutoRescale write FAutoRescale;
    property Compressed:boolean read ReadCompressed write WriteCompressed;
    property ExtendToPole:boolean read FExtendToPole write FExtendToPole;
    property Data[LongIndex,LatIndex:cardinal]:single read GetValue write PutValue; default;
    property DeltaLat:double read Header.DeltaLat;
    property DeltaLong:double read Header.DeltaLong;
    property MaximumLongitudinalInterpolation:double read FMaxLongInterpol write FMaxLongInterpol;
    property FracMissing:double read FFracMissing;
    property Scaling:double read FScaling;
    property Offset:double read FOffset;
    property InterpolationFailPlan:TInterpolationFailPlan read FInterpolationFailPlan write FInterpolationFailPlan;
    property Level:double read FDataLevel write FDataLevel;
    property DateTime:TDateTime read FDateTimeStamp write FDateTimeStamp;
    property NumLong:cardinal read Header.NumLong;
    property NumLat:cardinal read Header.NumLat;
    property StartLat:double read Header.StartLat;
    property StartLong:double read Header.StartLong;
    property TagValue:double read Header.TagValue write Header.TagValue;
    property NonUniformLats:boolean read FNonUniformLats;
    property PolePlotFunctionToUse:TConvertLatLongToXY read FFunctionToUse write FFunctionToUse;
  end;

  TSynopticGlobalArray=array of TSynopticGlobal;

  ESynopticGlobal=class(exception);

{General purpose TSynopticGlobal related functions and procedures.}
procedure CopySynopticGlobalHeader(Source:TSynopticGlobalHeader;out Dest:TSynopticGlobalHeader);
function LoadSynopticGlobalHeader(ANumLong,ANumLat:cardinal;ADeltaLong,
  ADeltaLat,AStartLong,AStartLat:double):TSynopticGlobalHeader;
function SynopticGlobalMultiplyAAndB(A,B:TSynopticGlobal):TSynopticGlobal;
function SynopticGlobalAddAAndB(A,B:TSynopticGlobal):TSynopticGlobal;
function SynopticGlobalDivideAByB(A,B:TSynopticGlobal):TSynopticGlobal;
function SynopticGlobalMean(ASynopticGlobalArray:TSynopticGlobalArray;NumVals:word=1):TSynopticGlobal;
function SynopticGlobalStdv(ASynopticGlobalArray:TSynopticGlobalArray;Mean:TSynopticGlobal):TSynopticGlobal;



implementation

uses
  {This is a standard delphi unit.}
  Math,

  {These are specific units}
  gb_astro,
  gb_math;

constructor TSynopticGlobal.Create;
begin
  raise ESynopticGlobal.Create('the call to a TSynopticGlobal constructor with '+
    'no parameters is not allowed.');
end; {End of TSynopticGlobal.Create.}

constructor TSynopticGlobal.Create(HeaderData:TSynopticGlobalHeader;MaxMem:cardinal=high(cardinal));
var
  I:integer; {Loop control variable through latitude steps.}
begin
  MaxMemory:=MaxMem;
  SetDefaultValues;
  Header:=HeaderData;
  if Header.NumLat=0 then raise ESynopticGlobal.Create('calling the constructor with NumLat=0 is not a good idea.');
  setlength(FLatitudes,Header.NumLat);
  for I:=0 to Header.NumLat-1 do FLatitudes[I]:=Header.StartLat+I*Header.DeltaLat;
  FData:=T2DSingleArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
  Clear;
end; {End of TSynopticGlobal.Create.}

constructor TSynopticGlobal.Create(Source:TSynopticGlobal);
begin
  if Assigned(Source) then DeepCopy(Source);
end; {End of TSynopticGlobal.Create.}

constructor TSynopticGlobal.Create(Filename:string;MaxMem:cardinal=high(cardinal));
var
  Source:file; {The file handle for reading in the data.}
begin
  MaxMemory:=MaxMem;
  SetDefaultValues;
  if not FileExists(Filename) then raise ESynopticGlobal.Create('the file named '+Filename+' doesn''t exist.');
  assignfile(Source,Filename);
  reset(Source,1);
  ReadFromFile(Source);
  closefile(Source);
end; {End of TSynopticGlobal.Create.}

constructor TSynopticGlobal.Create(var Source:file;MaxMem:cardinal=high(cardinal));
begin
  MaxMemory:=Maxmem;
  SetDefaultValues;
  ReadFromFile(Source);
end; {End of TSynopticGlobal.Create.}

procedure TSynopticGlobal.SetDefaultValues;
begin
  FScaling:=1.0;
  FOffset:=0.0;
  FDateTimeStamp:=0.0;
  FDataLevel:=0.0;
  FMaxLongInterpol:=60;
  FNonUniformLats:=false;
  FAutoRescale:=true;
  FExtendToPole:=false;
  FInterpolationFailPlan:=raiseexception;
end; {End of TSynopticGlobal.SetDefaultValues.}

procedure TSynopticGlobal.TransformLongitude(var Longitude:double);
var
  FullLongCoverage:boolean;
begin
  with Header do
  begin
    if DeltaLong*NumLong>=360.0 then FullLongCoverage:=true else FullLongCoverage:=false;
    if ((Longitude<StartLong) and (not FullLongCoverage)) then
      raise ESynopticGlobal.Create('the longitude value was less than the start longitude.');
    if ((Longitude>EndLongitude) and (not FullLongCoverage)) then
      raise ESynopticGlobal.Create('the longitude value was greater than the end longitude.');
    {The next two lines are crucial. We want the longitude between the start and}
    {end longitudes, or within Header.Deltalong degrees of the start longitude. }
    while Longitude<StartLong do Longitude:=Longitude+360.0;
    while Longitude>EndLongitude do Longitude:=Longitude-360.0;
  end;
end;

procedure TSynopticGlobal.TransformLatitude(var Latitude:double);
begin
  with Header do
  begin
    if Latitude<FLatitudes[0] then
    begin
      if FExtendToPole then Latitude:=FLatitudes[0] else raise ESynopticGlobal.Create(
      'the latitude value was less than the start latitude.');
    end;
    if Latitude>FLatitudes[High(FLatitudes)] then
    begin
      if FExtendToPole then Latitude:=FLatitudes[High(FLatitudes)] else raise ESynopticGlobal.Create(
      'the latitude value was greater than the end latitude.');
    end;
  end;
end;

procedure TSynopticGlobal.CompressionOn(MaxValue,MinValue:double);
var
  Hold:T2DWordArray;
  I,J:cardinal;
  SValue:single;
begin
  if FData is T2DWordArray then raise ESynopticGlobal.Create(
    'compression is already switched on.');

  {First we skip through the array and make sure that none of the values is   }
  {bigger than MaxValue or smaller than MinValue. If they are and FAutoRescale}
  {is true we extend MaxValue or MinValue. If FAutoRescale is false we raise  }
  {an exception.}
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
    begin
      SValue:=(FData as T2DSingleArray)[I,J];
      if not EquivSingle(SValue,NullValue) then
      begin
        if SValue>MaxValue then
          if FAutoRescale then MaxValue:=SValue else raise ESynopticGlobal.Create(
            'the prescribed maximim value was too restrictive.');
        if SValue<MinValue then
          if FAutoRescale then MinValue:=SValue else raise ESynopticGlobal.Create(
            'the prescribed minimum value was too restrictive.');
      end;
    end;

  {Now we can happily go through the array knowing that we won't be out of range.}
  FScaling:=(MaxValue-MinValue)/65534.0;
  FOffset:=MinValue-FScaling;
  Hold:=T2DWordArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
    begin
      SValue:=(FData as T2DSingleArray)[I,J];
      if EquivSingle(SValue,NullValue) then Hold[I,J]:=0 else Hold[I,J]:=round((SValue-FOffset)/FScaling);
    end;
  (FData as T2DSingleArray).Free;
  FData:=Hold;
end; {End of TSynopticGlobal.CompressionOn.}

function TSynopticGlobal.GetCornerValues(Latitude,Longitude:double):TCornerRec;
{------------------------------------------------------------------------------}
{ This function returns the array of 4 values at the corners of the grid which }
{ span the passed latitude and longitude.                                      }
{------------------------------------------------------------------------------}
begin
  with Result do
  begin
    {It is still possible that the longitude is less than Header.StartLong.}
    if Longitude<Header.StartLong then
    begin
      X1:=Header.NumLong;
      X2:=1;
      DLong:=Longitude-Header.StartLong+Header.DeltaLong;
    end
    else
    begin
      if Header.DeltaLong=0.0 then {This means there is only one longitude.}
      begin
        X1:=1;
        X2:=1;
      end
      else
      begin
        X1:=trunc((Longitude-Header.StartLong)/Header.DeltaLong)+1;
        X2:=X1+1;
      end;
      if X2>Header.NumLong then
      begin
        if DeltaLong*NumLong>=360.0 then X2:=1 else X2:=X1;
      end;
      DLong:=Longitude-((X1-1)*Header.DeltaLong+Header.StartLong);
    end;
    if FNonUniformLats then
    begin
      Y1:=Header.NumLat;
      while FLatitudes[Y1-1]>Latitude do dec(Y1);
    end
    else Y1:=trunc((Latitude-Header.StartLat)/Header.DeltaLat)+1;
    if cardinal(Y1)=Header.NumLat then Y2:=Y1 else Y2:=Y1+1;
    DLat:=Latitude-FLatitudes[Y1-1];
    V[1]:=GetValue(X1,Y1);
    V[2]:=GetValue(X1,Y2);
    V[3]:=GetValue(X2,Y1);
    V[4]:=GetValue(X2,Y2);
  end;
end; {End of TSynopticGlobal.GetCornerValues.}

function TSynopticGlobal.AllNull:boolean;
var
  I,J:cardinal;
begin
  Result:=true;
  for I:=1 to Header.NumLat do
    for J:=1 to Header.NumLong do
      if IsNotNull(J,I) then Result:=false;
end;

function TSynopticGlobal.BiLinear(Latitude,Longitude:double):single;
{------------------------------------------------------------------------------}
{ This function uses bilinear interpolation to extract the value at the        }
{ specified latitude and longitude. If any of the four values required for the }
{ bilinear interpolation are not available, the function either raises an      }
{ an exception or returns NullValue depending on the setting of                }
{ FInterpolationFailPlan. The default setting is to return the null value.     }
{ Exceptions may also be raised if the specified latitude and longitude are    }
{ outside the geographical coverage of the gridded data.                       }
{------------------------------------------------------------------------------}
var
  D1,D2:double;
  CornerRec:TCornerRec;
  I:byte;
begin
  {First check that the latitude and longitude fall}
  {within the geographical region spanned by the data.}
  TransformLongitude(Longitude);
  TransformLatitude(Latitude);
  Result:=NullValue;
  CornerRec:=GetCornerValues(Latitude,Longitude);
  with CornerRec do
  begin
    if ((DLong=0.0) and (DLat=0.0)) then {No spatial interpolation required.}
    begin
      Result:=V[1];
      if (EquivSingle(Result,NullValue) and (FInterpolationFailPlan=raiseexception)) then
        raise ESynopticGlobal.Create('there was a null data value in the call to BiLinear');
      exit;
    end;
    for I:=1 to 4 do if EquivSingle(V[I],NullValue) then
    begin
      case FInterpolationFailPlan of
        raiseexception:raise ESynopticGlobal.Create('there was a null data value in the call to BiLinear');
        returnnull:exit;
      end;
    end;
    D1:=(DLong/Header.DeltaLong)*(V[3]-V[1])+V[1];
    D2:=(DLong/Header.DeltaLong)*(V[4]-V[2])+V[2];
    if FNonUniformLats then
    begin
      if DLat=0 then Result:=D1 else Result:=DLat/(FLatitudes[Y2-1]-FLatitudes[Y1-1])*(D2-D1)+D1;
    end
    else Result:=frac((Latitude-Header.StartLat)/Header.DeltaLat)*(D2-D1)+D1;
  end;
end; {End of TSynopticGlobal.BiLinear.}

procedure TSynopticGlobal.Clear;
{------------------------------------------------------------------------------}
{ This fills the data array with NullValue. Note that for a compressed data    }
{ data array where the values are held as 2 byte words, the value 0 is         }
{ reserved to denote the null data value.                                      }
{------------------------------------------------------------------------------}
var
  I,J:cardinal;
begin
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
    if (FData is T2DSingleArray) then (FData as T2DSingleArray)[I,J]:=NullValue
    else (FData as T2DWordArray)[I,J]:=0;
end; {End of TSynopticGlobal.Clear.}

procedure TSynopticGlobal.FillWithZero;
var
  I,J:cardinal;
begin
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
      PutValue(I,J,0.0);
end; {End of TSynopticGlobal.FillWithZero.}

procedure TSynopticGlobal.DeepCopy(Source:TSynopticGlobal);
{------------------------------------------------------------------------------}
{ This routine copies all data and setting from 'Source' to the current object.}
{------------------------------------------------------------------------------}
var
  I,J:cardinal;
begin
  Header:=Source.GetHeader;
  FNonUniformLats:=Source.NonUniformLats;
  FLatitudes:=Copy(Source.FLatitudes);
  MaxMemory:=Source.MaxMemory;
  FScaling:=Source.FScaling;
  FOffset:=Source.FOffset;
  FDataLevel:=Source.FDataLevel;
  FAutoRescale:=Source.FAutoRescale;
  FInterpolationFailPlan:=Source.FInterpolationFailPlan;
  FMaxLongInterpol:=Source.FMaxLongInterpol;
  FFracMissing:=Source.FFracMissing;
  FExtendToPole:=Source.FExtendToPole;
  FNonUniformLats:=Source.FNonUniformLats;
  FDateTimeStamp:=Source.FDateTimeStamp;
  FFunctionToUse:=Source.FFunctionToUse;
  if FData<>nil then FreeAndNil(FData);
  if Source.FData is T2DWordArray then
    FData:=T2DWordArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
  if Source.FData is T2DSingleArray then
    FData:=T2DSingleArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
    begin
      if FData is T2DWordArray then (FData as T2DWordArray)[I,J]:=
        (Source.FData as T2DWordArray)[I,J];
      if FData is T2DSingleArray then (FData as T2DSingleArray)[I,J]:=
        (Source.FData as T2DSingleArray)[I,J];
    end;
end; {End of TSynopticGlobal.DeepCopy.}

function TSynopticGlobal.DistWeighting(Latitude,Longitude:double;Order:byte):single;
var
  CornerRec:TCornerRec;
  Total,Weight:double;
  Dist:array[1..4] of double;
  I:byte;
  Done:boolean;
begin
  {First check that the latitude and longitude fall}
  {within the geographical region spanned by the data.}
  TransformLongitude(Longitude);
  TransformLatitude(Latitude);
  Result:=NullValue;
  CornerRec:=GetCornerValues(Latitude,Longitude);
  with CornerRec,Header do
  begin
    for I:=1 to 4 do if abs(V[I])>=1E25 then
    begin
      case FInterpolationFailPlan of
        raiseexception:raise ESynopticGlobal.Create('there was a null data value in the call to BiLinear');
        returnnull:exit;
      end;
    end;
    Dist[1]:=CalcDist(FLatitudes[Y1-1],(X1-1)*DeltaLong+StartLong,Latitude,Longitude)/1000.0;
    Dist[2]:=CalcDist(FLatitudes[Y2-1],(X1-1)*DeltaLong+StartLong,Latitude,Longitude)/1000.0;
    Dist[3]:=CalcDist(FLatitudes[Y1-1],(X2-1)*DeltaLong+StartLong,Latitude,Longitude)/1000.0;
    Dist[4]:=CalcDist(FLatitudes[Y2-1],(X2-1)*DeltaLong+StartLong,Latitude,Longitude)/1000.0;
    Done:=false;
    for I:=1 to 4 do if Dist[I]=0.0 then
    begin
      Result:=V[I];
      Done:=true;
    end;
    if not Done then
    begin
      for I:=1 to 4 do Dist[I]:=Power(Dist[I],Order);
      Weight:=0.0;
      for I:=1 to 4 do Weight:=Weight+1/Dist[I];
      Total:=0.0;
      for I:=1 to 4 do Total:=Total+(V[I]/Dist[I]);
      Result:=Total/Weight;
    end;
  end;
end; {End of TSynopticGlobal.DistWeighting.}

function TSynopticGlobal.GetHeader:TSynopticGlobalHeader;
begin
  Result:=Header;
end; {End of TSynopticGlobal.GetHeader.}

function TSynopticGlobal.GetMaximumValue:single;
var
  I,J:cardinal;
begin
  if AllNull then raise ESynopticGlobal.Create('there were no non-null data in the object.');
  Result:=-1E30;
  for I:=1 to Header.NumLat do
    for J:=1 to Header.NumLong do
      if (IsNotNull(J,I) and (GetValue(J,I)>Result)) then Result:=GetValue(J,I);
end;

procedure TSynopticGlobal.GetMemoryStats(var TDisk:boolean;var TNumPages:cardinal);
begin
  FData.GetMemoryStats(TDisk,TNumPages);
end; {End of TSynopticGlobal.GetMemoryStats.}

function TSynopticGlobal.GetMemoryUsed:cardinal;
begin
  Result:=FData.GetMemoryUsed;
end; {End of TSynopticGlobal.GetMemoryUsed.}

function TSynopticGlobal.GetMinimumValue:single;
var
  I,J:cardinal;
begin
  if AllNull then raise ESynopticGlobal.Create('there were no non-null data in the object.');
  Result:=1E30;
  for I:=1 to Header.NumLat do
    for J:=1 to Header.NumLong do
      if (IsNotNull(J,I) and (GetValue(J,I)<Result)) then Result:=GetValue(J,I);
end;

function TSynopticGlobal.GetValue(LongIndex,LatIndex:cardinal):single;
begin
  if FData is T2DWordArray then
  begin
    if (FData as T2DWordArray)[LongIndex,LatIndex]=0 then Result:=NullValue
    else Result:=(FData as T2DWordArray)[LongIndex,LatIndex]*FScaling+FOffset;
  end
  else Result:=(FData as T2DSingleArray)[LongIndex,LatIndex];
end; {End of TSynopticGlobal.GetValue.}

function TSynopticGlobal.GetValueByCoords(ALongitude,ALatitude:double;Interpolate:boolean):single;
var
  LongIndex,
  LatIndex:cardinal;
begin
  if Interpolate then Result:=Bilinear(ALatitude,ALongitude) else
  begin
    {First check that the latitude and longitude fall}
    {within the geographical region spanned by the data.}
    TransformLongitude(ALongitude);
    TransformLatitude(ALatitude);
    LongIndex:=trunc((ALongitude-Header.StartLong)/Header.DeltaLong)+1;
    if FNonUniformLats then
    begin
      LatIndex:=Header.NumLat;
      while FLatitudes[LatIndex-1]>ALatitude do dec(LatIndex);
    end
    else LatIndex:=trunc((ALatitude-Header.StartLat)/Header.DeltaLat)+1;
    Result:=GetValue(LongIndex,LatIndex);
  end;
end; {End of TSynopticGlobal.GetValue.}

function TSynopticGlobal.GetZonalMean(LatIndex:cardinal):single;
var
  I,Num:cardinal;
begin
  if LatIndex<1 then raise ESynopticGlobal.Create('the index passed to ZonalMean was less than 1.');
  if LatIndex>Header.NumLat then raise ESynopticGlobal.Create('the index passed to ZonalMean was greater than 1.');
  Result:=0.0;
  Num:=0;
  for I:=1 to Header.NumLong do
  if IsNotNull(I,LatIndex) then
  begin
    inc(Num);
    Result:=Result+GetValue(I,LatIndex);
  end;
  if Num=0 then Result:=NullValue else Result:=Result/Num;
end; {End of TSynopticGlobal.GetZonalMean.}

function TSynopticGlobal.IsNull(LongIndex,LatIndex:cardinal):boolean;
begin
  if FData is T2DWordArray then
  begin
    if (FData as T2DWordArray)[LongIndex,LatIndex]=0 then Result:=true else Result:=false;
  end
  else Result:=EquivSingle((FData as T2DSingleArray)[LongIndex,LatIndex],NullValue);
end; {End of TSynopticGlobal.IsNull.}

function TSynopticGlobal.IsNullByCoords(ALongitude,ALatitude:double):boolean;
var
  LongIndex,
  LatIndex:cardinal;
begin
  TransformLongitude(ALongitude);
  TransformLatitude(ALatitude);
  LongIndex:=trunc((ALongitude-Header.StartLong)/Header.DeltaLong)+1;
  if FNonUniformLats then
  begin
    LatIndex:=Header.NumLat;
    while FLatitudes[LatIndex-1]>ALatitude do dec(LatIndex);
  end
  else LatIndex:=trunc((ALatitude-Header.StartLat)/Header.DeltaLat)+1;
  Result:=IsNull(LongIndex,LatIndex);
end; {End of TSynopticGlobal.IsNull.}

function TSynopticGlobal.IsNotNull(LongIndex,LatIndex:cardinal):boolean;
begin
  Result:=not IsNull(LongIndex,LatIndex);
end; {End of TSynopticGlobal.IsNotNull.}

function TSynopticGlobal.GetCellArea(LatIndex:cardinal):single;
var
  Lat:array[1..2] of double;
  I,J:word;
const
  EarthRadius=6.371E6; {Mean radius of the Earth in meters.}
begin
  J:=High(FLatitudes);
  for I:=1 to 2 do
  begin
    if ((LatIndex=1) and (I=1)) then Lat[I]:=FLatitudes[0]-(FLatitudes[1]-FLatitudes[0])/2.0 else
      if ((LatIndex=(J+1)) and (I=2)) then Lat[I]:=FLatitudes[J]+(FLatitudes[J]-FLatitudes[J-1])/2.0 else
        Lat[I]:=(FLatitudes[LatIndex+I-2]+FLatitudes[LatIndex+I-3])/2.0;
    if Lat[I]>90 then Lat[I]:=90.0;
    if Lat[I]<-90 then Lat[I]:=-90.0;
  end;
  Result:=abs(DegToRad(Header.DeltaLong)*sqr(EarthRadius)*(sin(DegToRad(Lat[1]))-sin(DegToRad(Lat[2]))));
end; {End of TMercator.GetCellArea.}

function TSynopticGlobal.GetLatitude(Index:cardinal):double;
begin
  Result:=FLatitudes[Index-1];
end; {End of TSynopticGlobal.GetLatitude.}

function TSynopticGlobal.GetLongitude(Index:cardinal):double;
begin
  Result:=Header.StartLong+(Index-1)*Header.DeltaLong;
end; {End of TSynopticGlobal.GetLongitude.}

function TSynopticGlobal.EndLatitude:double;
begin
  Result:=FLatitudes[NumLat-1];
end; {End of TSynopticGlobal.EndLatitude.}

function TSynopticGlobal.EndLongitude:double;
begin
  Result:=StartLong+DeltaLong*(NumLong-1);
end; {End of TSynopticGlobal.EndLongitude.}

procedure TSynopticGlobal.FillWithDataFrom(Source:TSynopticGlobal);
var
  I,J:cardinal;
  HoldPlan:TInterpolationFailPlan;
begin
  HoldPlan:=Source.InterpolationFailPlan;
  Source.InterpolationFailPlan:=raiseexception;
  for I:=1 to Header.NumLat do
    for J:=1 to Header.NumLong do
    try
      Data[J,I]:=Source.BiLinear(GetLatitude(I),GetLongitude(J));
    except
      on ESynopticGlobal do Data[J,I]:=NullValue;
    end;
  Source.InterpolationFailPlan:=HoldPlan;
end; {End of TSynopticGlobal.FillWithDataFrom.}

procedure TSynopticGlobal.FillWithLongitudeInterpolation;
var
  NumMissing,
  LongInd1,LongInd2,
  LatIndex,LongIndex,ScanLongIndex:cardinal;
  Longitude,Long1,Long2:double;
begin
  for LatIndex:=1 to Header.NumLat do
  begin
    NumMissing:=0;
    for LongIndex:=1 to Header.NumLong do if IsNull(LongIndex,LatIndex) then inc(NumMissing);
    if ((NumMissing>0) and (NumMissing<Header.NumLong)) then
      for LongIndex:=1 to Header.NumLong do
      if IsNull(LongIndex,LatIndex) then
      begin
        {First look west.}
        Long1:=GetLongitude(LongIndex);
        LongInd1:=LongIndex;
        repeat
          dec(LongInd1);
          if LongInd1=0 then LongInd1:=Header.NumLong;
          Long1:=Long1-Header.DeltaLong;
        until IsNotNull(LongInd1,LatIndex);

        {Now look east.}
        Long2:=GetLongitude(LongIndex);
        LongInd2:=LongIndex;
        repeat
          inc(LongInd2);
          if LongInd2>Header.NumLong then LongInd2:=1;
          Long2:=Long2+Header.DeltaLong;
        until IsNotNull(LongInd2,LatIndex);

        if Long2-Long1<FMaxLongInterpol then
        begin
          ScanLongIndex:=LongInd1;
          Longitude:=Long1;
          repeat
            inc(ScanLongIndex);
            if ScanLongIndex>Header.NumLong then ScanLongIndex:=1;
            Longitude:=Longitude+Header.DeltaLong;
            Data[ScanLongIndex,LatIndex]:=(Longitude-Long1)*(Data[LongInd2,LatIndex]-Data[LongInd1,LatIndex])/
              (Long2-Long1)+Data[LongInd1,LatIndex];
          until ScanLongIndex>=(LongInd2-1);
        end;
      end;
  end;
end; {End of TSynopticGlobal.FillWithLongitudeInterpolation.}

function TSynopticGlobal.GetLatitudes:TDoubleArray;
begin
  Result:=Copy(FLatitudes);
end; {End of TSynopticGlobal.GetLatitudes.}

function TSynopticGlobal.GetTotalTimesArea(Scaling:double):single;
var
  I,J:cardinal;
begin
  Result:=0.0;
  for I:=1 to Header.NumLat do
    for J:=1 to Header.NumLong do
      if not EquivSingle(Data[J,I],NullValue) then
        Result:=Result+Data[J,I]*GetCellArea(I)*Scaling;
end; {End of TSynopticGlobal.GetTotalTimesArea.}

function TSynopticGlobal.GetAreaWeightedMean(LatSouth,LatNorth,LongWest,LongEast:double):double;
var
  LatIndStart,
  LatIndEnd,
  LongIndStart,
  LongIndEnd,
  TotalData,
  I,J:cardinal;
  Total,
  TotalArea,
  Lat1,
  Lat2,
  Long1,
  Long2,
  CellArea:double;
begin
  {This code still needs to be thoroughly validated.}
  LatIndStart:=1;
  while FLatitudes[LatIndStart-1]<LatSouth do inc(LatIndStart);
  LatIndEnd:=Header.NumLat;
  while FLatitudes[LatIndEnd-1]>LatNorth do dec(LatIndEnd);
  if Header.DeltaLong=0.0 then
  begin
    LongIndStart:=1;
    LongIndEnd:=1;
  end
  else
  begin
    LongIndStart:=1;
    while GetLongitude(LongIndStart)<LongWest do inc(LongIndStart);
    LongIndEnd:=Header.NumLong;
    while GetLongitude(LongIndEnd)>LongEast do dec(LongIndEnd);
  end;
  Total:=0.0;
  TotalArea:=0.0;
  TotalData:=0;
  for I:=LongIndStart to LongIndEnd do
  begin
    if I=LongIndStart then Long1:=LongWest else Long1:=GetLongitude(I)-(Header.DeltaLong/2);
    if I=LongIndEnd then Long2:=LongEast else Long2:=GetLongitude(I)+(Header.DeltaLong/2);
    for J:=LatIndStart to LatIndEnd do
    begin
      if J=LatIndStart then Lat1:=LatSouth else Lat1:=(FLatitudes[J-1]+FLatitudes[J-2])/2.0;
      if J=LatIndEnd then Lat2:=LatNorth else Lat2:=(FLatitudes[J]+FLatitudes[J-1])/2.0;
      if IsNotNull(I,J) then
      begin
        inc(TotalData);
        if Header.DeltaLong=0 then CellArea:=abs(sin(DegToRad(Lat1))-sin(DegToRad(Lat2)))
        else CellArea:=abs(sin(DegToRad(Lat1))-sin(DegToRad(Lat2)))*(Long2-Long1);
        Total:=Total+GetValue(I,J)*CellArea;
        TotalArea:=TotalArea+CellArea;
      end;
    end;
  end;
  if TotalArea>0.0 then Result:=Total/TotalArea else Result:=NullValue;
  FFracMissing:=1.0-TotalData/((LatIndEnd-LatIndStart+1)*(LongIndEnd-LongIndStart+1));
end; {End of TSynopticGlobal.GetAreaWeightedMean.}

function TSynopticGlobal.GetMeanAtLatIndex(LatIndex:cardinal):double;
var
  Total:double;
  Num,
  I:cardinal;
begin
  Total:=0.0;
  Num:=0;
  for I:=1 to Header.NumLong do
    if IsNotNull(I,LatIndex) then
    begin
      Total:=Total+GetValue(I,LatIndex);
      inc(Num);
    end;
  if Num=0 then raise ESynopticGlobal.Create('there were not enough non-null data to calculate the mean.')
  else Result:=Total/Num;
end;

function TSynopticGlobal.Missing:double;
begin
  Result:=Missing(1,Header.NumLat);
end; {End of TSynopticGlobal.Missing.}

function TSynopticGlobal.Missing(StartLatInd,EndLatInd:cardinal):double;
var
  NumNull,
  I,J:cardinal;
begin
  NumNull:=(EndLatInd-StartLatInd+1)*Header.NumLong;
  for I:=StartLatInd to EndLatInd do
    for J:=1 to Header.NumLong do
      if not EquivSingle(Data[J,I],NullValue) then dec(NumNull);
  Result:=NumNull/((EndLatInd-StartLatInd+1)*Header.NumLong);
end; {End of TSynopticGlobal.Missing.}

procedure TSynopticGlobal.PutValue(LongIndex,LatIndex:cardinal;Value:single);
var
  TestVal:double;
begin
  if FData is T2DWordArray then
  begin
    TestVal:=(Value-FOffset)/FScaling;
    if ((round(TestVal)<1) or (round(TestVal)>65535)) then
      if FAutoRescale then Rescale(Value) else raise ESynopticGlobal.Create(
        'AutoRescale was switched off and the value passed was out of range.');
    {Note that calling Rescale will have changed}
    {FOffset and FScaling before the next line is called.}
    (FData as T2DWordArray)[LongIndex,LatIndex]:=round((Value-FOffset)/FScaling);
  end
  else (FData as T2DSingleArray)[LongIndex,LatIndex]:=Value;
end; {End of TSynopticGlobal.PutValue.}

procedure TSynopticGlobal.PutValueByCoords(ALongitude,ALatitude:double;Value:single);
var
  LongIndex,
  LatIndex:cardinal;
begin
  {First check that the latitude and longitude fall}
  {within the geographical region spanned by the data.}
  TransformLongitude(ALongitude);
  TransformLatitude(ALatitude);
  LongIndex:=trunc((ALongitude-Header.StartLong)/Header.DeltaLong)+1;
  if FNonUniformLats then
  begin
    LatIndex:=Header.NumLat;
    while FLatitudes[LatIndex-1]>ALatitude do dec(LatIndex);
  end
  else LatIndex:=trunc((ALatitude-Header.StartLat)/Header.DeltaLat)+1;
  PutValue(LongIndex,LatIndex,Value);
end; {End of TSynopticGlobal.PutValue.}

function TSynopticGlobal.ReadCompressed:boolean;
begin
  if FData is T2DWordArray then Result:=true else Result:=false;
end; {End of TSynopticGlobal.ReadCompressed.}

procedure TSynopticGlobal.ReadDataGridFromFile(var Source:file);
begin
  if FData.LoadData(Source)<>0 then raise ESynopticGlobal.Create('LoadData failed');
end; {End of TSynopticGlobal.WriteDataGridToFile.}

procedure TSynopticGlobal.ReadDataGridFromFile(Filename:string);
var
  Source:file;
begin
  if not FileExists(Filename) then raise ESynopticGlobal.Create('the file named '+
    Filename+' couldn''t be found.');
  assignfile(Source,Filename);
  reset(Source,1);
  ReadDataGridFromFile(Source);
  closefile(Source);
end; {End of TSynopticGlobal.WriteDataGridToFile.}

procedure TSynopticGlobal.ReadFromFile(var Source:file);
var
  ESize:byte;
  Ch:char;
  I:integer;
  TestStr:string;
  StartPos:cardinal;
  Version:array[1..4] of byte;
begin
  StartPos:=filepos(Source);

  {Read in the characters that define the ID string. This is not written as a}
  {string so we read them in character by character and build up the ID string.}
  Teststr:='';
  for I:=1 to length(StrID) do
  begin
    blockread(Source,Ch,sizeof(Ch));
    TestStr:=TestStr+Ch;
  end;

  if TestStr=StrID then blockread(Source,Version,sizeof(Version)) {So far we do nothing with the version array.}
  else seek(Source,StartPos);
  blockread(Source,ESize,sizeof(byte));
  blockread(Source,Header,sizeof(Header));
  if TestStr=StrID then
  begin
    blockread(Source,FDataLevel,sizeof(FDataLevel));
    blockread(Source,FDateTimeStamp,sizeof(FDateTimeStamp));
    blockread(Source,FExtendToPole,sizeof(FExtendToPole));
    blockread(Source,FAutoRescale,sizeof(FAutoRescale));
    blockread(Source,FInterpolationFailPlan,sizeof(FInterpolationFailPlan));
  end
  else {This is version of the file before the format change on 29 April 2007.}
  begin
    {In older versions of the file the data level was stored in the Header.TagValue}
    {field and so we must pass it back to the DataLevel field here.}
    FDataLevel:=Header.TagValue;
  end;
  if ESize in [2,4] then
  begin
    FNonUniformLats:=false;
    setlength(FLatitudes,Header.NumLat);
    for I:=0 to Header.NumLat-1 do FLatitudes[I]:=Header.StartLat+I*Header.DeltaLat;
  end
  else
  begin
    FNonUniformLats:=true;
    ESize:=ESize-4;
    setlength(FLatitudes,Header.NumLat);
    for I:=1 to Header.NumLat do blockread(Source,FLatitudes[I-1],sizeof(double));
  end;
  if FData<>nil then FreeAndNil(FData);
  if ESize=2 then
  begin
    FData:=T2DWordArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
    blockread(Source,FScaling,sizeof(FScaling));
    blockread(Source,FOffset,sizeof(FScaling));
  end
  else FData:=T2DSingleArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
  FData.LoadData(Source);
end; {End of TSynopticGlobal.ReadFromFile.}

procedure TSynopticGlobal.Rescale(Value:single);
var
  MaxValue,
  MinValue,
  NewScaling,
  NewOffset,
  OldValue:double;
  I,J:cardinal;
begin
  if FData is T2DWordArray then
  begin
    MaxValue:=65535*FScaling+FOffset;
    if Value>MaxValue then MaxValue:=Value;
    MinValue:=FScaling+FOffset;
    if Value<MinValue then MinValue:=Value;
    NewScaling:=(MaxValue-MinValue)/65534.0;
    NewOffset:=MinValue-NewScaling;
    for I:=1 to Header.NumLong do
      for J:=1 to Header.NumLat do
      if (FData as T2DWordArray)[I,J]<>0 then
      begin
        OldValue:=(FData as T2DWordArray)[I,J]*FScaling+FOffset;
        (FData as T2DWordArray)[I,J]:=round((OldValue-NewOffset)/NewScaling);
      end;
    FScaling:=NewScaling;
    FOffset:=NewOffset;
  end;
end; {End of TSynopticGlobal.Rescale.}

function TSynopticGlobal.SameGridStructure(Source:TSynopticGlobal):boolean;
var
  I:integer;
begin
  Result:=true;
  if NumLong<>Source.NumLong then Result:=false;
  if StartLong<>Source.StartLong then Result:=false;
  if DeltaLong<>Source.DeltaLong then Result:=false;
  if NumLat<>Source.NumLat then Result:=false else
  begin
    for I:=1 to NumLat do
      if not EquivDouble(FLatitudes[I-1],Source.GetLatitude(I)) then Result:=false;
  end;
end; {End of TSynopticGlobal.SameGridStructure.}

procedure TSynopticGlobal.Scale(X:single);
var
  I,J:cardinal;
begin
  for I:=1 to Header.NumLong do
    for J:=1 to Header.NumLat do
    if IsNotNull(I,J) then
      Data[I,J]:=Data[I,J]*X;
end; {End of TSynopticGlobal.Scale.}

procedure TSynopticGlobal.SetScalingAndOffset(Scaling,Offset:double);
var
  MaxValue,
  MinValue:single;
begin
  MinValue:=Offset+Scaling;
  MaxValue:=Scaling*65535.0+Offset;
  if FData is T2DSingleArray then CompressionOn(MaxValue,MinValue);
end; {End of TSynopticGlobal.SetScalingAndOffset.}

procedure TSynopticGlobal.SpatialSmooth(Weights:T2DDoubleArray);
var
  OriginalData:T2DSingleArray;
  I,J,
  SubI,
  SubJ,
  IndI,
  IndJ,
  ContractI,
  ContractJ:integer;
  TotalWeight,
  Total:double;
  FullLongCoverage:boolean;
begin
  if (Weights.GetNumX mod 2)<>1 then raise ESynopticGlobal.Create('the width of the filter array must be odd.');
  if (Weights.GetNumY mod 2)<>1 then raise ESynopticGlobal.Create('the height of the filter array must be odd.');
  OriginalData:=T2DSingleArray.Create(Header.NumLong,Header.NumLat);
  try
    for I:=1 to Header.NumLong do
      for J:=1 to Header.NumLat do
        OriginalData[I,J]:=GetValue(I,J);
    if DeltaLong*Header.NumLong>=360.0 then FullLongCoverage:=true else FullLongCoverage:=false;
    for I:=1 to Header.NumLong do
      for J:=1 to Header.NumLat do
      if IsNotNull(I,J) then
      begin
        Total:=0.0;
        TotalWeight:=0.0;

        {Check if we bump into edges in an east-west direction}
        ContractI:=0;
        IndI:=I-(Weights.GetNumX div 2);
        if ((IndI<1) and (not FullLongCoverage)) then repeat inc(ContractI) until IndI+ContractI=1;
        IndI:=I+(Weights.GetNumX div 2);
        if ((IndI>Header.NumLong) and (not FullLongCoverage)) then repeat inc(ContractI) until IndI-ContractI=Header.NumLong;

        {Check if we bump into edge in a north-south direction}
        ContractJ:=0;
        IndJ:=J-(Weights.GetNumY div 2);
        if (IndJ<1) then repeat inc(ContractJ) until IndJ+ContractJ=1;
        IndJ:=J+(Weights.GetNumY div 2);
        if (IndJ>Header.NumLat) then repeat inc(ContractJ) until IndJ-ContractJ=Header.NumLat;

        for SubI:=ContractI-(Weights.GetNumX div 2) to (Weights.GetNumX div 2)-ContractI  do
          for SubJ:=ContractJ-(Weights.GetNumY div 2) to (Weights.GetNumY div 2)-ContractJ  do
          begin
            IndI:=I+SubI;
            if IndI<1 then IndI:=IndI+Header.NumLong;
            if IndI>Header.NumLong then IndI:=IndI-Header.NumLong;
            IndJ:=J+SubJ;
            if IsNotNull(IndI,IndJ) then
            begin
              Total:=Total+OriginalData[IndI,IndJ]*Weights[(Weights.GetNumX div 2)+1+SubI,(Weights.GetNumY div 2)+1+SubJ];
              TotalWeight:=TotalWeight+Weights[(Weights.GetNumX div 2)+1+SubI,(Weights.GetNumY div 2)+1+SubJ];
            end;
          end;
        if TotalWeight>0.0 then PutValue(I,J,Total/TotalWeight);
      end; {End of loop over all cells.}
  finally
    OriginalData.Free;
  end;
end; {End of TSynopticGlobal.SpatialSmooth.}

procedure TSynopticGlobal.SpecifyNonuniformLatitudes(Latitudes:TDoubleArray);
var
  I:integer;
begin
  setlength(FLatitudes,length(Latitudes));
  HeapSort(Latitudes);
  for I:=Low(Latitudes) to High(Latitudes) do FLatitudes[I]:=Latitudes[I];
  Header.StartLat:=Latitudes[0];
  Header.NumLat:=length(Latitudes);
  Header.DeltaLat:=0.0;
  FNonUniformLats:=true;
end; {End of TSynopticGlobal.SpecifyNonuniformLatitudes.}

procedure TSynopticGlobal.SwitchCompressionOff;
var
  Hold:T2DSingleArray;
  I,J:cardinal;
begin
  if FData is T2DWordArray then
  begin
    Hold:=T2DSingleArray.Create(Header.NumLong,Header.NumLat,MaxMemory);
    for I:=1 to Header.NumLong do
      for J:=1 to Header.NumLat do
      if (FData as T2DWordArray)[I,J]=0 then Hold[I,J]:=NullValue
      else Hold[I,J]:=(FData as T2DWordArray)[I,J]*FScaling+FOffset;
    (FData as T2DWordArray).Free;
    FData:=Hold;
  end;
end; {End of TSynopticGlobal.SwitchCompressionOff.}

procedure TSynopticGlobal.SwitchCompressionOn;
var
  I,J:cardinal;
  MaxValue,
  MinValue:double;
begin
  if FData is T2DSingleArray then
  begin
    MaxValue:=-1E30;
    MinValue:=1E30;
    for I:=1 to Header.NumLong do
      for J:=1 to Header.NumLat do
      with FData as T2DSingleArray do
      if not EquivSingle(Data[I,J],NullValue) then
      begin
        if Data[I,J]>MaxValue then MaxValue:=Data[I,J];
        if Data[I,J]<MinValue then MinValue:=Data[I,J];
      end;
    CompressionOn(MaxValue,MinValue);
  end;
end; {End of TSynopticGlobal.SwitchCompressionOn.}

procedure TSynopticGlobal.SwitchCompressionOn(MaxValue,MinValue:double);
begin
  if FData is T2DSingleArray then CompressionOn(MaxValue,MinValue);
end; {End of TSynopticGlobal.SwitchCompressionOn.}

procedure TSynopticGlobal.WriteCompressed(Value:boolean);
begin
  if Value then SwitchCompressionOn else SwitchCompressionOff;
end; {End of TSynopticGlobal.WriteCompressed.}

procedure TSynopticGlobal.WriteDataGridToFile(var Dest:file);
begin
  FData.SaveData(Dest);
end; {End of TSynopticGlobal.WriteDataGridToFile.}

procedure TSynopticGlobal.WriteDataToTextFile(Filename:string);
var
  Dest:textfile;
  I,J:integer;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'''Column A) Longitude.');
  writeln(Dest,'''Column B) Latitude.');
  writeln(Dest,'''Column C) Value.');
  for I:=1 to NumLat do
    for J:=1 to NumLong do
      if IsNotNull(J,I) then
        writeln(Dest,Header.StartLong+(J-1)*Header.DeltaLong:8:3,' ',FLatitudes[I-1]:7:3,' ',Data[J,I]:10:5);
  closefile(Dest);
end; {End of TSynopticGlobal.WriteDataToTextFile.}

procedure TSynopticGlobal.WriteToFile(var Dest:file);
const
  Version:array[1..4] of byte=(0,0,0,1);
var
  ESize:byte;
  I:integer;
begin
  {Write out the version information.}
  for I:=1 to length(StrID) do blockwrite(Dest,StrID[I],sizeof(char));
  blockwrite(Dest,Version,sizeof(Version));

  {ESize is used to denote whether the data are stored as a word}
  {array and whether the latitude spacing is non-uniform.}
  if FData is T2DWordArray then ESize:=2 else ESize:=4;
  if FNonUniformLats then ESize:=ESize+4;

  blockwrite(Dest,ESize,sizeof(byte));
  blockwrite(Dest,Header,sizeof(Header));
  blockwrite(Dest,FDataLevel,sizeof(FDataLevel));
  blockwrite(Dest,FDateTimeStamp,sizeof(FDateTimeStamp));
  blockwrite(Dest,FExtendToPole,sizeof(FExtendToPole));
  blockwrite(Dest,FAutoRescale,sizeof(FAutoRescale));
  blockwrite(Dest,FInterpolationFailPlan,sizeof(FInterpolationFailPlan));
  if FNonUniformLats then
    for I:=1 to Header.NumLat do blockwrite(Dest,FLatitudes[I-1],sizeof(double));
  if FData is T2DWordArray then
  begin
    blockwrite(Dest,FScaling,sizeof(FScaling));
    blockwrite(Dest,FOffset,sizeof(FScaling));
  end;
  FData.SaveData(Dest);
end; {End of TSynopticGlobal.WriteToFile.}

procedure TSynopticGlobal.WriteToFile(Filename:string);
var
  Dest:file;
begin
  assignfile(Dest,Filename);
  rewrite(Dest,1);
  WriteToFile(Dest);
  closefile(Dest);
end; {End of TSynopticGlobal.WriteToFile.}

destructor TSynopticGlobal.Destroy;
begin
  if FData<>nil then FreeAndNil(FData);
  setlength(FLatitudes,0);
  inherited Destroy;
end; {End of TSynopticGlobal.Destroy.}

procedure CopySynopticGlobalHeader(Source:TSynopticGlobalHeader;out Dest:TSynopticGlobalHeader);
begin
  Dest.NumLong:=Source.NumLong;
  Dest.NumLat:=Source.NumLat;
  Dest.TagValue:=Source.TagValue;
  Dest.DeltaLong:=Source.DeltaLong;
  Dest.DeltaLat:=Source.DeltaLat;
  Dest.StartLong:=Source.StartLong;
  Dest.StartLat:=Source.StartLat;
end; {End of CopySynopticGlobalHeader.}

function LoadSynopticGlobalHeader(ANumLong,ANumLat:cardinal;ADeltaLong,
  ADeltaLat,AStartLong,AStartLat:double):TSynopticGlobalHeader;
begin
  with Result do
  begin
    NumLong:=ANumLong;
    NumLat:=ANumLat;
    TagValue:=0.0;
    DeltaLong:=ADeltaLong;
    DeltaLat:=ADeltaLat;
    StartLong:=AStartLong;
    StartLat:=AStartLat;
  end;
end; {End of LoadSynopticGlobalHeader.}

function SynopticGlobalMultiplyAAndB(A,B:TSynopticGlobal):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function multiplies the TSynopticGlobals specified by 'A' and 'B' and   }
{ returns a new TSynopticGlobal. The returned object is created within this    }
{ function and so it is the responsibility of the calling program to free the  }
{ object. The geographical coverage and grid spacing of the returned object is }
{ the same as that of 'A'. Valid products are calculated only if valid values  }
{ can be extracted from both 'A' and 'B'. If the grid spacing for 'A' and 'B'  }
{ are not the same, values are bilinearly interpolated within 'B' at the grid  }
{ locations of 'A'.                                                            }
{------------------------------------------------------------------------------}
var
  I,J:integer; {Loop control variables through the latitudes and longitudes.}
  SameGrid:boolean; {Set to true if both 'A' and 'B' have the same grid spacing.}
  HoldFP:TInterpolationFailPlan; {Holds the interpolation fail plan of 'B'.}
begin
  Result:=TSynopticGlobal.Create(A);
  SameGrid:=A.SameGridStructure(B);
  HoldFP:=B.InterpolationFailPlan;
  B.InterpolationFailPlan:=raiseexception;
  for I:=1 to Result.NumLong do
    for J:=1 to Result.NumLat do
    if A.IsNotNull(I,J) then
      if SameGrid then
      begin
        if B.IsNull(I,J) then Result[I,J]:=NullValue else Result[I,J]:=A[I,J]*B[I,J];
      end
      else
      begin
        try
          Result[I,J]:=A[I,J]*B.BiLinear(A.GetLatitude(J),A.GetLongitude(I));
        except
          on ESynopticGlobal do Result[I,J]:=NullValue;
        end;
      end;
  B.InterpolationFailPlan:=HoldFP;
end; {End of SynopticGlobalMultiplyAAndB.}

function SynopticGlobalAddAAndB(A,B:TSynopticGlobal):TSynopticGlobal;
var
  I,J:integer; {Loop control variables through the latitudes and longitudes.}
  SameGrid:boolean; {Set to true if both 'A' and 'B' have the same grid spacing.}
  HoldFP:TInterpolationFailPlan; {Holds the interpolation fail plan of 'B'.}
begin
  Result:=TSynopticGlobal.Create(A);
  SameGrid:=A.SameGridStructure(B);
  HoldFP:=B.InterpolationFailPlan;
  B.InterpolationFailPlan:=raiseexception;
  for I:=1 to Result.NumLong do
    for J:=1 to Result.NumLat do
    if A.IsNotNull(I,J) then
      if SameGrid then
      begin
        if B.IsNull(I,J) then Result[I,J]:=NullValue else Result[I,J]:=A[I,J]+B[I,J];
      end
      else
      begin
        try
          Result[I,J]:=A[I,J]+B.BiLinear(A.GetLatitude(J),A.GetLongitude(I));
        except
          on ESynopticGlobal do Result[I,J]:=NullValue;
        end;
      end;
  B.InterpolationFailPlan:=HoldFP;
end; {End of SynopticGlobalAddAAndB.}

function SynopticGlobalDivideAByB(A,B:TSynopticGlobal):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function takes the TSynopticGlobal specified by 'A' and divides it by   }
{ the TSynopticGlobal specified by 'B' and returns a new TSynopticGlobal. The  }
{ returned object is created within this function and so it is the respons-    }
{ ibility of the calling program to free the object. The geographical coverage }
{ and grid spacing of the returned object is the same as that of 'A'. Valid    }
{ ratios are calculated only if valid values can be extracted from both 'A'    }
{ and 'B'. If 'B' returns zero for any location, a null value will be inserted }
{ into the returned field. If the grid spacing for 'A' and 'B' are not the     }
{ same, values are bilinearly interpolated within 'B' at the grid locations    }
{ of 'A'.                                                                      }
{------------------------------------------------------------------------------}
var
  I,J:integer; {Loop control variables through the latitudes and longitudes.}
  Value:double; {The values extracted from bilinear interpolation within 'B'.}
  SameGrid:boolean; {Set to true if both 'A' and 'B' have the same grid spacing.}
  HoldFP:TInterpolationFailPlan; {Holds the interpolation fail plan of 'B'.}
begin
  Result:=TSynopticGlobal.Create(A);
  SameGrid:=A.SameGridStructure(B);
  HoldFP:=B.InterpolationFailPlan;
  B.InterpolationFailPlan:=raiseexception;
  for I:=1 to Result.NumLong do
    for J:=1 to Result.NumLat do
    if A.IsNotNull(I,J) then
      if SameGrid then
      begin
        if (B.IsNull(I,J) or (A[I,J]=0.0)) then Result[I,J]:=NullValue
        else Result[I,J]:=A[I,J]/B[I,J];
      end
      else
      begin
        try
          Value:=B.BiLinear(A.GetLatitude(J),A.GetLongitude(I));
          if Value=0.0 then Result[I,J]:=NullValue else Result[I,J]:=A[I,J]/Value;
        except
          on ESynopticGlobal do Result[I,J]:=NullValue;
        end;
      end;
  B.InterpolationFailPlan:=HoldFP;
end; {End of SynopticGlobalDivideAByB.}

function SynopticGlobalMean(ASynopticGlobalArray:TSynopticGlobalArray;NumVals:word=1):TSynopticGlobal;
{------------------------------------------------------------------------------}
{ This function averages the list of TSynopticGlobals passed by                }
{ 'ASynopticGlobalArray', returning a new TSynopticGlobal. An optional second  }
{ parameter, specifying the minimum number of values that must be available    }
{ for calculating the mean, can be passed (otherwise a default value of 1 is   }
{ assumed). The returned object is created within this function and so it is   }
{ the responsibility of the calling program to free the object. The geograph-  }
{ ical coverage and grid spacing of the returned object is the same as that of }
{ the first object on the list. If the grid spacing for subsequent items on    }
{ the list is not the same as the first item, values are bilinearly inter-     }
{ polated to the locations defined by the first item. If there were insuffic-  }
{ ient values to calculate a mean, then a null value is entered for that       }
{ location.                                                                    }
{------------------------------------------------------------------------------}
var
  Counter:TSynopticGlobal;
  I,J,K:integer; {Loop control variables through the latitudes and longitudes.}
  Value:double; {The values extracted from bilinear interpolation within 'B'.}
  SameGrid:boolean; {Set to true if both 'A' and 'B' have the same grid spacing.}
  HoldFP:TInterpolationFailPlan; {Holds the interpolation fail plan of 'B'.}
begin
  if length(ASynopticGlobalArray)=0 then raise ESynopticGlobal.Create(
    'the array of TSynopticGlobals was empty');
  Result:=TSynopticGlobal.Create(ASynopticGlobalArray[0]);
  Result.FillWithZero;
  Counter:=TSynopticGlobal.Create(ASynopticGlobalArray[0]);
  Counter.FillWithZero;
  try
    for I:=Low(ASynopticGlobalArray) to High(ASynopticGlobalArray) do
    begin
      SameGrid:=ASynopticGlobalArray[0].SameGridStructure(ASynopticGlobalArray[I]);
      HoldFP:=ASynopticGlobalArray[I].InterpolationFailPlan;
      ASynopticGlobalArray[I].InterpolationFailPlan:=returnnull;
      for J:=1 to Result.NumLong do
        for K:=1 to Result.NumLat do
        begin
          if SameGrid then Value:=ASynopticGlobalArray[I][J,K]
          else Value:=ASynopticGlobalArray[I].BiLinear(Result.GetLatitude(K),Result.GetLongitude(J));
          if not EquivSingle(Value,NullValue) then
          begin
            Result[J,K]:=Result[J,K]+Value;
            Counter[J,K]:=Counter[J,K]+1;
          end;
        end;
      ASynopticGlobalArray[I].InterpolationFailPlan:=HoldFP;
    end;
    for J:=1 to Result.NumLong do
      for K:=1 to Result.NumLat do
        if Counter[J,K]<NumVals then Result[J,K]:=NullValue else Result[J,K]:=Result[J,K]/Counter[J,K];
  finally
    Counter.Free;
  end;
end;


function SynopticGlobalStdv(ASynopticGlobalArray:TSynopticGlobalArray;Mean:TSynopticGlobal):TSynopticGlobal;
begin

end;

end.
