//@***h* BDBP/gb_TMeasurementSet.pas
// 
// NAME
//   gb_TMeasurementSet.pas
//
// PURPOSE
//  Provides definitions and routines that are necessary to read and write to the
//  BDBP database.
//
// DESCRIPTION
//  This unit Provides definitions and routines that are necessary to read and write 
//  to the BDBP database.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_TMeasurementSet;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  SysUtils;

type
  {Add to this list freely as required but first make sure that what you want is not already}
  {on the list. Also please add the relevant line to the MeasurementDescToString function.}
  TMeasurementDesc=(GeopotentialHeight,Temperature,Pressure,Ozone_LowRes,NO2,H2O,NO,CH4,HCl,HF,
    ExtAt352p3nm,ExtAt353p4nm,ExtAt385nm,ExtAt441p6nm,ExtAt442p3nm,ExtAt453nm,
    ExtAt525nm,ExtAt603p4nm,ExtAt601p4nm,ExtAt779p4nm,ExtAt781nm,ExtAt921nm,
    ExtAt922p4nm,ExtAt1018nm,ExtAt1020nm,ExtAt1060nm,AerosolSAD,AerosolEffRad,RiseOrSet,
    RelHumidtyInPercent,SondeCorrFactorNotApplied,ExtAt1000nm,ExtAt450nm,Ozone_HighRes, ExtAt386nm,ExtAt452nm,HNO3,
    N2O,CFC11,CFC12,N2O5,CLONO2,ExtAt7120nm,ExtAt8270nm,ExtAt10600nm,ExtAt11760nm,ExtAt780nm,ExtAt8290nm,
    ClO,CO,HCN,HO2,HOCl,OH,RHI,SO2,TropopauseHeight,TropopausePressure);

  TMeasurement=record
    Value:single; {The actual measurement value.}
    Error:word; {The measurement error in percent x 100.}
    Desc:TMeasurementDesc; {The measurement descriptor.}
  end;
  PMeasurement=^TMeasurement;

  st10=string[10];

  TMeasurementSet=class(TList)
  strict private
    FSource:st10; {Describes the origin of this measurement set.}
    FLatitude,        {The latitude at which the measurement was made.}
    FLongitude,       {The longitude at which the measurement was made.}
    FEquivLat:single; {The equivalent latitude at which the measurement was made.}
    FDateTime:TDateTime; {The date and time of the measurement.}
    FMemoryUsed:cardinal; {The memory used by the object.}
    procedure ValidateFieldNum(FieldNum:word);
    procedure ClearItems;
  public
    constructor Create; overload;
    constructor Create(ExistingObj:TMeasurementSet); overload;
    procedure AddMeasurement(AValue:single;AnError:word;AMeasurementDesc:TMeasurementDesc); overload;
    procedure AddMeasurement(AMeasurement:TMeasurement); overload;
    function  GetNumMeasurements:word;
    function  GetValue(FieldNum:word):single;
    procedure PutValue(FieldNum:word;Value:single);
    function  GetValueForDesc(Desc:TMeasurementDesc):single;
    function  GetError(FieldNum:word):word;
    function  GetErrorForDesc(Desc:TMeasurementDesc):single;
    function  GetMeasurementDesc(FieldNum:word):TMeasurementDesc;
    procedure WriteToFile(var Dest:file);
    procedure WriteToTextFile(Filename:string;LevelCount:word);overload;
    procedure WriteToTextFile(Filename:string);overload;
    procedure AppendToTextFile(var Dest:textfile);
    procedure ReadFromFile(var ASource:file);
    destructor Destroy; override;

    property Values[FieldNum:word]:single read GetValue write PutValue;default;
    property MemoryUsed:cardinal read FMemoryUsed;
    property Source:st10 read FSource write FSource;
    property Latitude:single read FLatitude write FLatitude;
    property Longitude:single read FLongitude write FLongitude;
    property EquivalentLatitude:single read FEquivLat write FEquivLat;
    property DateTime:TDateTime read FDateTime write FDateTime;
  end;

  EMeasurementSet=class(exception);

function CopyMeasurementSet(AMeasurementSet:TMeasurementSet):TMeasurementSet;
function MeasurementDescToString(Value:TMeasurementDesc):string;

implementation

{------------------------------------------------------------------------------}
{ Here are the TMeasurementSet methods.                                        }
{------------------------------------------------------------------------------}
constructor TMeasurementSet.Create;
begin
  inherited Create;
  FMemoryUsed:=sizeof(st10)+sizeof(TDateTime);
  Source:='';
  DateTime:=0.0;
  FLatitude:=-1E10;
  FLongitude:=-1E10;
  FEquivLat:=-1E10;
end;

constructor TMeasurementSet.Create(ExistingObj:TMeasurementSet);
var
  I:integer;
begin
  Create;
  FSource:=ExistingObj.FSource;
  FDateTime:=ExistingObj.FDateTime;
  FLatitude:=ExistingObj.FLatitude;
  FLongitude:=ExistingObj.FLongitude;
  FEquivLat:=ExistingObj.FEquivLat;
  for I:=0 to ExistingObj.Count-1 do AddMeasurement(TMeasurement(ExistingObj.Items[I]^));
end;

procedure TMeasurementSet.ValidateFieldNum(FieldNum:word);
begin
  if ((FieldNum<1) or (FieldNum>Count)) then
    raise EMeasurementSet.Create(' the field number passed was out of range.');
end;

procedure TMeasurementSet.ClearItems;
var
  I:integer;
  Hold:PMeasurement;
begin
  for I:=0 to Count-1 do
  begin
    Hold:=Items[I];
    {I am not sure which of the next two lines is best. On 25 Feb 2009 I changed back to freemem.}
    freemem(Hold,sizeof(TMeasurement));
//    dispose(Hold);
    FMemoryUsed:=FMemoryUsed-sizeof(TMeasurement);
  end;
  Clear;
end;

procedure TMeasurementSet.AddMeasurement(AValue:single;AnError:word;AMeasurementDesc:TMeasurementDesc);
var
  Data:PMeasurement;
begin
  {I am not sure which of the next two lines is best. On 25 Feb 2009 I changed back to getmem.}
  getmem(Data,sizeof(TMeasurement));
//  new(Data);
  Data^.Value:=AValue;
  Data^.Error:=AnError;
  Data^.Desc:=AMeasurementDesc;
  Add(Data);
  FMemoryUsed:=FMemoryUsed+sizeof(TMeasurement);
end;

procedure TMeasurementSet.AddMeasurement(AMeasurement:TMeasurement);
var
  Data:PMeasurement;
begin
  {I am not sure which of the next two lines is best. On 25 Feb 2009 I changed back to getmem.}
  getmem(Data,sizeof(TMeasurement));
//  new(Data);
  Data^:=AMeasurement;
  Add(Data);
  FMemoryUsed:=FMemoryUsed+sizeof(TMeasurement);
end;

function TMeasurementSet.GetNumMeasurements:word;
begin
  Result:=Count;
end;

function TMeasurementSet.GetValue(FieldNum:word):single;
begin
  ValidateFieldNum(FieldNum);
  Result:=PMeasurement(Items[FieldNum-1])^.Value;
end;

procedure TMeasurementSet.PutValue(FieldNum:word;Value:single);
begin
  ValidateFieldNum(FieldNum);
  PMeasurement(Items[FieldNum-1])^.Value:=Value;
end;

function TMeasurementSet.GetValueForDesc(Desc:TMeasurementDesc):single;
var
  I:integer;
begin
  Result:=-999.99;
  for I:=0 to Count-1 do
    if PMeasurement(Items[I])^.Desc=Desc then Result:=PMeasurement(Items[I])^.Value;
end;

function TMeasurementSet.GetError(FieldNum:word):word;
begin
  ValidateFieldNum(FieldNum);
  Result:=PMeasurement(Items[FieldNum-1])^.Error;
end;

function TMeasurementSet.GetErrorForDesc(Desc:TMeasurementDesc):single;
var
  I:integer;
begin
  Result:=-999.99;
  for I:=0 to Count-1 do
    if PMeasurement(Items[I])^.Desc=Desc then Result:=PMeasurement(Items[I])^.Error;
end;

function TMeasurementSet.GetMeasurementDesc(FieldNum:word):TMeasurementDesc;
begin
  ValidateFieldNum(FieldNum);
  Result:=PMeasurement(Items[FieldNum-1])^.Desc;
end;

procedure TMeasurementSet.WriteToFile(var Dest:file);
var
  I:word;
begin
  if Count>0 then
  begin
    blockwrite(Dest,FSource,sizeof(st10));
    blockwrite(Dest,FDateTime,sizeof(TDateTime));
    blockwrite(Dest,FLatitude,sizeof(single));
    blockwrite(Dest,FLongitude,sizeof(single));
    blockwrite(Dest,FEquivLat,sizeof(single));
    blockwrite(Dest,Count,sizeof(word));
    for I:=0 to Count-1 do blockwrite(Dest,PMeasurement(Items[I])^,sizeof(TMeasurement));
  end;
end;

procedure TMeasurementSet.WriteToTextFile(Filename:string;LevelCount:word);
var
  Dest:textfile;
  I:word;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'This record is at level ',LevelCount);
  writeln(Dest,'Source: ',FSource);
  writeln(Dest,'Date and time: ',DateTimeToStr(FDateTime));
  writeln(Dest,'Latitide: ',FLatitude:8:4);
  writeln(Dest,'Longitude: ',FLongitude:8:4);
  writeln(Dest,'Equivalent latitude: ',FEquivLat:8:4);
  writeln(Dest,'There are ',Count,' measurements available in this measurement set. They are:');
  for I:=0 to Count-1 do
  with PMeasurement(Items[I])^ do
    writeln(Dest,I,') ',MeasurementDescToString(Desc),' Value=',Value:10:4,' Error=',Error/100:8:4,'%');
  closefile(Dest);
end;

procedure TMeasurementSet.WriteToTextFile(Filename:string);
var
  Dest:textfile;
  I:word;
begin
  assignfile(Dest,Filename);
  rewrite(Dest);
  writeln(Dest,'Source: ',FSource);
  writeln(Dest,'Date and time: ',DateTimeToStr(FDateTime));
  writeln(Dest,'Latitide: ',FLatitude:8:4);
  writeln(Dest,'Longitude: ',FLongitude:8:4);
  writeln(Dest,'Equivalent latitude: ',FEquivLat:8:4);
  writeln(Dest,'There are ',Count,' measurements available in this measurement set. They are:');
  for I:=0 to Count-1 do
  with PMeasurement(Items[I])^ do
    writeln(Dest,I,') ',MeasurementDescToString(Desc),' Value=',Value:10:4,' Error=',Error/100:8:4,'%');
  closefile(Dest);
end;

procedure TMeasurementSet.AppendToTextFile(var Dest:textfile);
var
  I:word;
begin
  writeln(Dest,'Source: ',FSource);
  writeln(Dest,'Date and time: ',DateTimeToStr(FDateTime));
  writeln(Dest,'Latitide: ',FLatitude:8:4);
  writeln(Dest,'Longitude: ',FLongitude:8:4);
  writeln(Dest,'Equivalent latitude: ',FEquivLat:8:4);
  writeln(Dest,'There are ',Count,' measurements available in this measurement set. They are:');
  for I:=0 to Count-1 do
  with PMeasurement(Items[I])^ do
    writeln(Dest,I,') ',MeasurementDescToString(Desc),' Value=',Value:10:4,' Error=',Error/100:8:4,'%');
end;

procedure TMeasurementSet.ReadFromFile(var ASource:file);
var
  NumMeasurements,
  I:word;
  Ch:AnsiChar;
  Test:byte;
  Data:PMeasurement;
begin
  blockread(ASource,FSource,sizeof(st10));
  blockread(ASource,FDateTime,sizeof(DateTime));
  blockread(ASource,FLatitude,sizeof(single));
  blockread(ASource,FLongitude,sizeof(single));
  blockread(ASource,FEquivLat,sizeof(single));
  blockread(ASource,NumMeasurements,sizeof(word));
  ClearItems;
  for I:=1 to NumMeasurements do
  begin
    getmem(Data,sizeof(TMeasurement));
    blockread(ASource,Data^,sizeof(TMeasurement));
    Add(Data);
    FMemoryUsed:=FMemoryUsed+sizeof(TMeasurement);
  end;
end;

destructor TMeasurementSet.Destroy;
begin
  ClearItems;
  inherited Destroy;
end;

function CopyMeasurementSet(AMeasurementSet:TMeasurementSet):TMeasurementSet;
var
  I:word;
begin
  with AMeasurementSet do
  begin
    Result:=TMeasurementSet.Create;
    Result.Source:=Source;
    Result.DateTime:=DateTime;
    Result.Latitude:=Latitude;
    Result.Longitude:=Longitude;
    Result.EquivalentLatitude:=EquivalentLatitude;
    for I:=1 to GetNumMeasurements do
      Result.AddMeasurement(PMeasurement(Items[I-1])^);
  end;
end;

function MeasurementDescToString(Value:TMeasurementDesc):string;
begin
  case Value of
    GeopotentialHeight:Result:='Geopotential height in km';
    Temperature:Result:='Temperature in kelvin';
    Pressure:Result:='Pressure in hPa';
    Ozone_LowRes:Result:='Ozone concentration in 10^18 molec.m^-3';
    NO2:Result:='Nitrogen dioxide concentration in 10^18 molec.m^-3';
    H2O:Result:='Water vapour concentration in 10^18 molec.m^-3';
    NO:Result:='Nitric oxide concentration in 10^18 molec.m^-3';
    CH4:Result:='Methane concentration in 10^18 molec.m^-3';
    HCl:Result:='Hydrogen chloride concentration in 10^18 molec.m^-3';
    HF:Result:='Hydrogen flouride concentration in 10^18 molec.m^-3';
    ExtAt352p3nm:Result:='Aerosol extinction at 352.3 nm in km^-1';
    ExtAt353p4nm:Result:='Aerosol extinction at 353.4 nm in km^-1';
    ExtAt385nm:Result:='Aerosol extinction at 385 nm in km^-1';
    ExtAt441p6nm:Result:='Aerosol extinction at 441.6 nm in km^-1';
    ExtAt442p3nm:Result:='Aerosol extinction at 442.3 nm in km^-1';
    ExtAt453nm:Result:='Aerosol extinction at 453 nm in km^-1';
    ExtAt525nm:Result:='Aerosol extinction at 525 nm in km^-1';
    ExtAt603p4nm:Result:='Aerosol extinction at 603.4 nm in km^-1';
    ExtAt601p4nm:Result:='Aerosol extinction at 601.4 nm in km^-1';
    ExtAt779p4nm:Result:='Aerosol extinction at 779.4 nm in km^-1';
    ExtAt781nm:Result:='Aerosol extinction at 781 nm in km^-1';
    ExtAt921nm:Result:='Aerosol extinction at 921 nm in km^-1';
    ExtAt922p4nm:Result:='Aerosol extinction at 922.4 nm in km^-1';
    ExtAt1018nm:Result:='Aerosol extinction at 1018 nm in km^-1';
    ExtAt1020nm:Result:='Aerosol extinction at 1020 nm in km^-1';
    ExtAt1060nm:Result:='Aerosol extinction at 1060 nm in km^-1';
    AerosolSAD:Result:='Aerosol surface aera density in micrometer^2/cm^3';
    AerosolEffRad:Result:='Aerosol effective radius in micrometers';
    RiseOrSet:Result:='Sun rise or sun set event';
    RelHumidtyInPercent:Result:='Relative humidity in percent';
    SondeCorrFactorNotApplied:Result:='Ozonesonde correction factor that has NOT been applied';
    ExtAt1000nm:Result:='Aerosol extinction at 1000 nm in km^-1';
    ExtAt450nm:Result:='Aerosol extinction at 450 nm in km^-1';
    Ozone_HighRes:Result:='Ozone concentration in 10^18 molec.m^-3';
    ExtAt386nm:Result:='Aerosol extinction at 386 nm in km^-1';
    ExtAt452nm:Result:='Aerosol extinction at 452 nm in km^-1';
    HNO3:Result:='Nitric acid concentration in 10^18 molec.m^-3';
    N2O:Result:='Nitrous oxide concentration in 10^18 molec.m^-3';
    CFC11:Result:='ChloroFluoroCarbon-11 concentration in 10^18 molec.m^-3';
    CFC12:Result:='ChloroFluoroCarbon-12 concentration in 10^18 molec.m^-3';
    N2O5:Result:='Dinitrogen pentoxide concentration in 10^18 molec.m^-3';
    CLONO2:Result:='Chlorine Nitrate concentration in 10^18 molec.m^-3';
    ExtAt7120nm:Result:='Aerosol extinction at 7120 nm in km^-1';
    ExtAt8270nm:Result:='Aerosol extinction at 8270 nm in km^-1';
    ExtAt10600nm:Result:='Aerosol extinction at 10600 nm in km^-1';
    ExtAt11760nm:Result:='Aerosol extinction at 11760 nm in km^-1';
    ExtAt780nm:Result:='Aerosol extinction at 780 nm in km^-1';
    ExtAt8290nm:Result:='Aerosol extinction at 8290 nm in km^-1';
    ClO:Result:='Chlorine monoxide concentration in 10^18 molec.m^-3';
    CO:Result:='Carbon monoxide concentration in 10^18 molec.m^-3';
    HCN:Result:='Hydrogen cyanide concentration in 10^18 molec.m^-3';
    HO2:Result:='Hydroperoxyl concentration in 10^18 molec.m^-3';
    HOCl:Result:='Hypochlorus acid concentration in 10^18 molec.m^-3';
    OH:Result:='Hydroxyl radical concentration in 10^18 molec.m^-3';
    RHI:Result:='Relative humidity with respect to ice';
    SO2:Result:='Sulfur dioxide concentration in 10^18 molec.m^-3';


  end;
end;

end.
