//@***h* BDBP/gb_TripletArray.pas
// 
// NAME
//   gb_TripletArray.pas
//
// PURPOSE
//  Defines a class which stores data values with their uncertainty, a time stamp 
//  and some information about the data.  
//
// DESCRIPTION
//  This unit defines a class which stores data values with their uncertainty, a 
//  time stamp and some information about the data.  
//  
//
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_TripletArray;

{$I UNITDIREC.INC}

interface

uses
  Classes,
  gb_array;

type
  TInfoObject=class
  private
    FNumFields:word;
  public
    procedure WriteToFile(var Dest:textfile);virtual;abstract;
    function  GetFieldDescriptor(FieldNum:word):string;virtual;abstract;
    function  GetColumnsDescriptor:string;virtual;abstract;
    function  SpawnCopy:TInfoObject;virtual;abstract;
    property NumFields:word read FNumFields write FNumFields;
  end;

  TRegressionTriplet=record
    DateTime:TDateTime;
    Measurement,
    MeasurementError:double;
    InfoObject:TInfoObject;
  end;

  TTripletArray=class(TList)
  private
    {This field is true if the array has been sorted into increasing time order.}
    SortedByTime:boolean;
  protected
    procedure ClearAllEntries;
    procedure TestN(N:integer);
  public
    constructor Create;
    procedure AddEntry(ARegressionTriplet:TRegressionTriplet);overload;
    procedure AddEntry(ADateTime:TDateTime;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);overload;
    procedure DeepCopy(Source:TTripletArray);
    function  GetEntry(N:integer):TRegressionTriplet;
    function  GetDateTime(N:integer):TDateTime;
    function  GetDateTimeList:TDoubleArray;
    function  GetMeasurement(N:integer):double;
    function  GetMeasurementError(N:integer):double;
    function  GetInfoObject(N:integer):TInfoObject;
    function  Get1DMeasurementArray:T1DDoubleArray;
    function  GetMeasurementArray:TDoubleArray;
    function  Get1DMeasurementErrorArray:T1DDoubleArray;
    procedure PutMeasurement(N:integer;Value:double);
    procedure SortByTime;
    destructor Destroy;override;
    property Data[I:integer]:TRegressionTriplet read GetEntry;default;
  end;

implementation

uses
  SysUtils,
  gb_math;

constructor TTripletArray.Create;
begin
  SortedByTime:=false;
  inherited Create;
end;

procedure TTripletArray.ClearAllEntries;
var
  I:integer;
  Hold:^TRegressionTriplet;
begin
  for I:=0 to Count-1 do
  begin
    Hold:=Items[I];
    FreeAndNil(Hold.InfoObject);
    freemem(Hold,sizeof(TRegressionTriplet));
  end;
  Clear;
end;

procedure TTripletArray.TestN(N:integer);
begin
  if ((N<1) or (N>Count)) then raise EListError.Create('the index value passed to a TTripletArray method was out of bounds.');
end;

procedure TTripletArray.AddEntry(ARegressionTriplet:TRegressionTriplet);
var
  Hold:^TRegressionTriplet;
begin
  getmem(Hold,sizeof(TRegressionTriplet));
  Hold^:=ARegressionTriplet;
  Add(Hold);
  SortedByTime:=false;
end;

procedure TTripletArray.AddEntry(ADateTime:TDateTime;AMeasurement,AMeasurementError:double;AInfoObj:TInfoObject=nil);
var
  Hold:^TRegressionTriplet;
begin
  getmem(Hold,sizeof(TRegressionTriplet));
  Hold^.DateTime:=ADateTime;
  Hold^.Measurement:=AMeasurement;
  Hold^.MeasurementError:=AMeasurementError;
  Hold^.InfoObject:=AInfoObj;
  Add(Hold);
  SortedByTime:=false;
end;

procedure TTripletArray.DeepCopy(Source:TTripletArray);
var
  I:integer;
  Hold:TRegressionTriplet;
begin
  ClearAllEntries;
  for I:=1 to Source.Count do
  begin
    Hold:=Source.GetEntry(I);
    if Hold.InfoObject=nil then AddEntry(Hold.DateTime,Hold.Measurement,Hold.MeasurementError,nil)
    else AddEntry(Hold.DateTime,Hold.Measurement,Hold.MeasurementError,Hold.InfoObject.SpawnCopy);
  end;
  SortedByTime:=Source.SortedByTime;
end;

function TTripletArray.GetEntry(N:integer):TRegressionTriplet;
begin
  TestN(N);
  Result:=TRegressionTriplet(Items[N-1]^);
end;

function TTripletArray.GetDateTime(N:integer):TDateTime;
begin
  TestN(N);
  Result:=TRegressionTriplet(Items[N-1]^).DateTime;
end;

function TTripletArray.GetDateTimeList:TDoubleArray;
var
  I:integer;
  Hold:TDoubleArray;
begin
  setlength(Hold,Count);
  for I:=0 to Count-1 do Hold[I]:=TRegressionTriplet(Items[I]^).DateTime;
  HeapSort(Hold);

  {There may be duplicate times in the list e.g. if measurements are different-}
  {iated by latitude. We should avoid putting duplicate on the returned array.}
  {So now transfer to the Result array and remove any duplicates.}
  setlength(Result,1);
  Result[0]:=Hold[0];
  for I:=Low(Hold) to High(Hold) do
  if Hold[I]<>Result[High(Result)] then
  begin
    setlength(Result,length(Result)+1);
    Result[High(Result)]:=Hold[I];
  end;
end; {End of TTripletArray.GetDateTimeList.}

function TTripletArray.GetMeasurement(N:integer):double;
begin
  TestN(N);
  Result:=TRegressionTriplet(Items[N-1]^).Measurement;
end;

function TTripletArray.GetMeasurementError(N:integer):double;
begin
  TestN(N);
  Result:=TRegressionTriplet(Items[N-1]^).MeasurementError;
end;

function TTripletArray.GetInfoObject(N:integer):TInfoObject;
begin
  TestN(N);
  Result:=TRegressionTriplet(Items[N-1]^).InfoObject;
end;

function TTripletArray.Get1DMeasurementArray:T1DDoubleArray;
var
  I:integer;
begin
  Result:=T1DDoubleArray.Create(Count);
  for I:=1 to Count do Result[I]:=GetMeasurement(I);
end;

function TTripletArray.GetMeasurementArray:TDoubleArray;
var
  I:integer;
begin
  setlength(Result,Count);
  for I:=1 to Count do Result[I-1]:=GetMeasurement(I);
end;

function  TTripletArray.Get1DMeasurementErrorArray:T1DDoubleArray;
var
  I:integer;
begin
  Result:=T1DDoubleArray.Create(Count);
  for I:=1 to Count do Result[I]:=GetMeasurementError(I);
end;

procedure TTripletArray.PutMeasurement(N:integer;Value:double);
begin
  TestN(N);
  TRegressionTriplet(Items[N-1]^).Measurement:=Value;
end;

function TTripletArraySort(Item1,Item2:pointer):integer;
begin
  Result:=0;
  if TRegressionTriplet(Item1^).DateTime<TRegressionTriplet(Item2^).DateTime then Result:=-1;
  if TRegressionTriplet(Item1^).DateTime>TRegressionTriplet(Item2^).DateTime then Result:=1;
end;

procedure TTripletArray.SortByTime;
begin
  if not SortedByTime then Sort(TTripletArraySort);
  SortedByTime:=true;
end;

destructor TTripletArray.Destroy;
begin
  ClearAllEntries;
  inherited Destroy;
end;

end.
