//@***h* BDBP/gb_astro.pas
// 
// NAME
//   gb_astro.pas
//
// PURPOSE
//  Holds routines for the calculation of various astronomical values 
//  which are not provided by any delphi units. 
//
// DESCRIPTION
//  This unit hold routines for the calculation of various astronomical values 
//  which are not provided by any delphi units. Some of these routines were  
//  obtained from: Iqbal M., An introduction to Solar Radiation, Academic   
//  Press, 1983.      
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_astro;

{$I UNITDIREC.INC}

interface

uses
  {$IFNDEF DELPHI5}
  DateUtils,
  {$ENDIF}
  sysutils;

type
  EAstro=class(Exception);

{Here follows an alphabetical list of the routines available in this unit.}
function CalcDist(Lat1,Long1,Lat2,Long2:double):double;
function CalcSZA(Year,Day,TZone:integer;Lat,Long,Hour:double):double; overload;
function CalcSZA(DateTime:TDateTime;TZone:integer;Lat,Long:double):double; overload;
procedure CalcSZA(Year,Day,TZone:integer;Lat,Long,Hour:double;var SZA,SAA:double); overload;
function ForceRange(R1,R2,Value:double):double;
function GetHour(Day,TimeZone:integer;LocalAppTime,Long:double):double;
function GetLocalAppTime(Day,TimeZone:integer;Hour,Long:double):double;
function GetTimeForSZA(Day,TimeZone:integer;Lat,Long,SZA:double;AM:boolean):double;
function GetZenithAngle(Day,TZone:integer;Lat,Long,Hour:double;STime:boolean):double;
function GetRiseTime(Day,TZone:integer;Lat,Long:double):double;
function GetSetTime(Day,TZone:integer;Lat,Long:double):double;
procedure NewThetaPhi(Lat1,Long1,Theta1,Phi1,Lat2,Long2:double;var Theta2:double;var Phi2:double);

implementation

uses
  math, {The delphi math unit.}
  gb_datetime;

const
  {These are the constants for calculating the equation of time.}
  T1:double=0.000075;
  T2:double=0.001868;
  T3:double=0.032077;
  T4:double=0.014615;
  T5:double=0.04089;

  {These are the constants for calculating the declination.}
  D1:double=0.006918;
  D2:double=0.399912;
  D3:double=0.07257;
  D4:double=0.006758;
  D5:double=0.000907;
  D6:double=0.002697;
  D7:double=0.00148;

function CalcDist(Lat1,Long1,Lat2,Long2:double):double;
{------------------------------------------------------------------------------}
{ This equation calculates the distance, in meters, between the 2 locations    }
{ specified by (Lat1,Long1) and (Lat2,Long2) on the Earth's surface. The lat-  }
{ itude and longitude must be specified in degrees and not radians. The funct- }
{ ion does not take into account that fact that the Earth is not spherical.    }
{------------------------------------------------------------------------------}
var
  DeltaLong,       {The difference in longitude.}
  DeltaLat:double; {The difference in latitude.}
begin
  DeltaLat:=DegToRad(Lat1-Lat2);
  DeltaLong:=DegToRad(Long1-Long2);
  if DeltaLong>Pi then DeltaLong:=DeltaLong-(2*Pi);
  if DeltaLong<-Pi then DeltaLong:=DeltaLong+(2*Pi);
  CalcDist:=6367650*abs(arccos(cos(DeltaLat)*cos(cos(DegToRad(Lat2))*DeltaLong)));
end; {End of CalcDist.}

function GetHour(Day,TimeZone:integer;LocalAppTime,Long:double):double;
{------------------------------------------------------------------------------}
{ This routine calculates the standard time given the day number in the year   }
{ (`Day'), the time zone of the location (`TimeZone'), the local apparent      }
{ time (`LocalAppTime') in decimal hours and the longitude of the location     }
{ (`Long') in degrees. The standard time is returned in decimal hours.         }
{------------------------------------------------------------------------------}
var
  EquationOfTime, {The value for the equation of time.}
  RetHour,        {The value returned by the function.}
  DA:double;      {The day angle.}
  I:word;         {Loop control variable.}
begin
  RetHour:=12.0;
  for I:=1 to 5 do
  begin
    DA:=2*pi*(Day-1+(RetHour/24.0))/365.25;
    EquationOfTime:=(T1+T2*cos(DA)-T3*sin(DA)-T4*cos(2*DA)-T5*sin(2*DA))*229.18/60.0;
    RetHour:=LocalAppTime-EquationOfTime-Long/15.0+TimeZone;
  end;
  GetHour:=RetHour;
end; {End of GetHour.}

function ForceRange(R1,R2,Value:double):double;
var
  Diff:double;
begin
  Diff:=R2-R1;
  while Value>R2 do Value:=Value-Diff;
  while Value<R1 do Value:=Value+Diff;
  Result:=Value;
end;

function CalcSZA(Year,Day,TZone:integer;Lat,Long,Hour:double):double;
{------------------------------------------------------------------------------}
{ This routine makes use of the algorithm published in Michalsky, 1988, to     }
{ calculate the solar zenith angle. It is much more accurate than the          }
{ GetZenithAngle routine. It has been compared with sunzen3, and agrees within }
{ 0.01 degrees.                                                                }
{------------------------------------------------------------------------------}
var
  JulianDay,     {The Julian Day.}
  MeanLong,      {The mean longitude.}
  MeanAnomaly,   {The mean anomaly.}
  EclipticLong,  {The ecliptic longitude.}
  Obliquity,     {The obliquity.}
  RightAscension,{The right ascension.}
  Num,           {Storage variable for numerator.}
  Den,           {Storage variable for denominator.}
  Declination,   {The Declination.}
  GMST,          {Greenwhich mean sidereal time.}
  LMST,          {Local mean sidereal time.}
  HourAngle,     {The hour angle.}
  Time:double;   {The time in decimal days.}
begin
  Hour:=Hour-TZone;
  if Hour<0.0 then
  begin
    Hour:=Hour+24.0;
    Dec(Day);
    if Day<1 then
    begin
      dec(Year);
      Day:=DaysInAYear(Year);
    end;
  end;
  JulianDay:=32916.5+(Year-1949)*365+trunc((Year-1949)/4.0)+Day+(Hour/24.0);
  Time:=JulianDay-51545.0;
  MeanLong:=ForceRange(0.0,360.0,280.460+0.9856474*Time);
  MeanAnomaly:=DegToRad(ForceRange(0.0,360.0,357.528+0.9856003*Time));
  EclipticLong:=DegToRad(ForceRange(0.0,360.0,MeanLong+1.915*sin(MeanAnomaly)+
    0.020*sin(2*MeanAnomaly)));
  Obliquity:=DegToRad(23.439-0.00000004*Time);
  Num:=cos(Obliquity)*sin(EclipticLong);
  Den:=cos(EclipticLong);
  RightAscension:=arctan(Num/Den);
  if Den<0.0 then RightAscension:=RightAscension+Pi else
    if Num<0.0 then RightAscension:=RightAscension+2*Pi;
  Declination:=arcsin(sin(Obliquity)*sin(EclipticLong));
  GMST:=ForceRange(0.0,24.0,6.697375+0.0657098242*Time+Hour);
  LMST:=ForceRange(0.0,24.0,GMST+(Long/15.0));
  LMST:=DegToRad(LMST*15.0);
  HourAngle:=ForceRange(-Pi,Pi,LMST-RightAscension);
  Result:=arccos(sin(Declination)*sin(DegToRad(Lat))+cos(Declination)*cos(DegToRad(Lat))*cos(HourAngle));
end; {End of CalcSZA.}

function CalcSZA(DateTime:TDateTime;TZone:integer;Lat,Long:double):double;
var
  DecHour:double;
  {$IFDEF DELPHI5}
  Year,Month,Date:word;
  {$ENDIF}
begin
  DecHour:=frac(DateTime)*24.0;
  {$IFDEF DELPHI5}
  DecodeDate(DateTime,Year,Month,Date);
  Result:=CalcSZA(Year,CalcDay(Year,Month,Date),TZone,Lat,Long,DecHour);
  {$ELSE}
  Result:=CalcSZA(YearOf(DateTime),DayOfTheYear(DateTime),TZone,Lat,Long,DecHour);
  {$ENDIF}
end;

procedure CalcSZA(Year,Day,TZone:integer;Lat,Long,Hour:double;var SZA,SAA:double);
var
  JulianDay,     {The Julian Day.}
  MeanLong,      {The mean longitude.}
  MeanAnomaly,   {The mean anomaly.}
  EclipticLong,  {The ecliptic longitude.}
  Obliquity,     {The obliquity.}
  RightAscension,{The right ascension.}
  Num,           {Storage variable for numerator.}
  Den,           {Storage variable for denominator.}
  Declination,   {The Declination.}
  GMST,          {Greenwhich mean sidereal time.}
  LMST,          {Local mean sidereal time.}
  Time:double;   {The time in decimal days.}
begin
  Hour:=Hour-TZone;
  if Hour<0.0 then
  begin
    Hour:=Hour+24.0;
    Dec(Day);
    if Day<1 then
    begin
      dec(Year);
      Day:=DaysInAYear(Year);
    end;
  end;
  JulianDay:=32916.5+(Year-1949)*365+trunc((Year-1949)/4.0)+Day+(Hour/24.0);
  Time:=JulianDay-51545.0;
  MeanLong:=ForceRange(0.0,360.0,280.460+0.9856474*Time);
  MeanAnomaly:=DegToRad(ForceRange(0.0,360.0,357.528+0.9856003*Time));
  EclipticLong:=DegToRad(ForceRange(0.0,360.0,MeanLong+1.915*sin(MeanAnomaly)+
    0.020*sin(2*MeanAnomaly)));
  Obliquity:=DegToRad(23.439-0.00000004*Time);
  Num:=cos(Obliquity)*sin(EclipticLong);
  Den:=cos(EclipticLong);
  RightAscension:=arctan(Num/Den);
  if Den<0.0 then RightAscension:=RightAscension+Pi else
    if Num<0.0 then RightAscension:=RightAscension+2*Pi;
  Declination:=arcsin(sin(Obliquity)*sin(EclipticLong));
  GMST:=ForceRange(0.0,24.0,6.697375+0.0657098242*Time+Hour);
  LMST:=ForceRange(0.0,24.0,GMST+(Long/15.0));
  LMST:=DegToRad(LMST*15.0);
  SAA:=ForceRange(-Pi,Pi,LMST-RightAscension);
  SZA:=arccos(sin(Declination)*sin(DegToRad(Lat))+cos(Declination)*cos(DegToRad(Lat))*cos(SAA));
end; {End of CalcSZA.}

function GetLocalAppTime(Day,TimeZone:integer;Hour,Long:double):double;
{------------------------------------------------------------------------------}
{ This routine calculates the local apparent time (the sun time) given the day }
{ number in the year (`Day'), the time zone (`TimeZone'), the longitude        }
{ (`Long') of the location in degrees, and the time of the day in decimal      }
{ hours (`Hour'). The function also makes the correction for the equation of   }
{ time. The routine returns the local apparent time in decimal hours.          }
{------------------------------------------------------------------------------}
var
  EquationOfTime, {The value for the equation of time.}
  DA:double;      {The day angle.}
begin
  DA:=2*pi*(Day-1+(Hour/24.0))/365.25;
  EquationOfTime:=(T1+T2*cos(DA)-T3*sin(DA)-T4*cos(2*DA)-T5*sin(2*DA))*229.18/60.0;
  GetLocalAppTime:=Hour+Long/15.0-TimeZone+EquationOfTime;
end; {End of GetLocalAppTime.}

function GetTimeForSZA(Day,TimeZone:integer;Lat,Long,SZA:double;AM:boolean):double;
{------------------------------------------------------------------------------}
{ This program calculates the time corresponding to a certain solar zenith     }
{ angle. The inputs required by the routine are:                               }
{ 1) The day number of the year, `Day'.                                        }
{ 2) The time zone for the location, `TimeZone'.                               }
{ 3) The latitude and longitude of the location in degrees, `Lat' and `Long'.  }
{ 4) The solar zenith angle, `SZA' in degrees.                                 }
{ For each solar zenith angle there will be two times in the day. If `AM' is   }
{ set to true the routine will return the time before local noon, otherwise it }
{ will return the time after local noon. The returned times are in decimal     }
{ hours. If no times are available for the SZA, the routine returns -999.      }
{------------------------------------------------------------------------------}
var
  RetHour,      {The value returned by the function.}
  DA,           {The day angle.}
  Declination,  {The solar declination.}
  HourAngle,    {The hour angle.}
  LocalAppTime, {The local apparent time.}
  Term1,        {General calculation variable.}
  Term2,        {General calculation variable.}
  Term3,        {General calculation variable.}
  Term4:double; {A function of terms 1 to 3.}
  I:word;       {Loop control variable.}
begin
  RetHour:=12.0;
  I:=0;
  repeat
    inc(I);
    DA:=2*pi*(Day-1+(RetHour/24.0))/365.25;
    Declination:=D1-D2*cos(DA)+D3*sin(DA)-D4*cos(2*DA)+D5*sin(2*DA)-
                 D6*cos(3*DA)+D7*sin(3*DA);
    Term1:=cos(SZA*pi/180.0);
    Term2:=sin(Declination)*sin(Lat*pi/180.0);
    Term3:=cos(Declination)*cos(Lat*pi/180.0);
    Term4:=(Term1-Term2)/Term3;
    if Term4>1.0 then Term4:=1.0;
    if ((Term4<-1.0) or (Term4>1.0)) then RetHour:=-999.0
    else
    begin
      if AM then HourAngle:=arccos(Term4) else HourAngle:=-1.0*arccos(Term4);
      LocalAppTime:=12.0-(HourAngle*12.0/pi);
      RetHour:=GetHour(Day,TimeZone,LocalAppTime,Long);
    end;
  until ((I=5) or (RetHour<0.0));
  GetTimeForSZA:=RetHour;
end; {End of GetTimeForSZA.}

function GetZenithAngle(Day,TZone:integer;Lat,Long,Hour:double;STime:boolean):double;
{------------------------------------------------------------------------------}
{ This function calculates the solar zenith angle as a function of the passed  }
{ parameters. The `Day' is the day number of the year, `TZone' is the time     }
{ zone eg. +12 for New Zealand, `Lat' and `Long' are the latitude and longit-  }
{ ude. The convention is positive for North and East. These values must be     }
{ decimal. The time is specified by `Hour' which is the decimal hour in the    }
{ day. If the `STime' parameter is true then the Hour parameter is taken as    }
{ being solar time. The routine returns the solar zenith angle in radians.     }
{------------------------------------------------------------------------------}
var
  HourAngle,     {The hour angle.}
  Declination,   {The solar declination.}
  LocalAppTime,  {The local apparent time.}
  Term1,         {The sine of the declination.}
  Term2,         {The sine of the latitude in radians.}
  Term3,         {Cosine of the declination in radians.}
  Term4,         {The cosine of the latitude in radians.}
  Term5,         {The cosine of the hour angle.}
  Term6,         {Term1*Term2+Term3*Term4*Term5.}
  FinalResult,   {The final result.}
  Lambda:double; {The day angle.}
begin
  Lambda:=2*pi*(Day-1+(Hour/24.0))/365.25;
  Declination:=D1-D2*cos(Lambda)+D3*sin(Lambda)       {This is from Iqbal }
                 -D4*cos(2*Lambda)+D5*sin(2*Lambda)   {equation 1.3.1 on  }
                 -D6*cos(3*Lambda)+D7*sin(3*Lambda);  {page 7 of the book.}
  if STime then LocalAppTime:=Hour
  else LocalAppTime:=GetLocalAppTime(Day,TZone,Hour,Long);
  HourAngle:=(12.0-LocalAppTime)*15.0*pi/180.0;
  Term1:=sin(Declination);
  Term2:=sin(Lat*pi/180.0);
  Term3:=cos(Declination);
  Term4:=cos(Lat*pi/180.0);
  Term5:=cos(HourAngle);
  Term6:=Term1*Term2+Term3*Term4*Term5;
  FinalResult:=arccos(Term6);
  GetZenithAngle:=FinalResult;
end; {End of GetZenithAngle routine.}

function GetRiseTime(Day,TZone:integer;Lat,Long:double):double;
{-----------------------------------------------------------------------------}
{ Given the day number of the year (`Day'), the time zone of the location     }
{ ('TZone'), the latitude and longitude of the location (`Lat' and `Long') in }
{ degrees, this routine returns the time (in decimal hours) for sun rise. If  }
{ there was no sun rise for this day at this location, the routine returns    }
{ -999.                                                                       }
{-----------------------------------------------------------------------------}
begin
  GetRiseTime:=GetTimeForSZA(Day,TZone,Lat,Long,90.0,true);
end; {End of GetRiseTime.}

function GetSetTime(Day,TZone:integer;Lat,Long:double):double;
{-----------------------------------------------------------------------------}
{ Given the day number of the year (`Day'), the time zone of the location     }
{ ('TZone'), the latitude and longitude of the location (`Lat' and `Long') in }
{ degrees, this routine returns the time (in decimal hours) for sun set. If   }
{ there was no sun set for this day at this location, the routine returns     }
{ -999.                                                                       }
{-----------------------------------------------------------------------------}
begin
  GetSetTime:=GetTimeForSZA(Day,TZone,Lat,Long,90.0,false);
end; {End of GetSetTime.}

procedure NewThetaPhi(Lat1,Long1,Theta1,Phi1,Lat2,Long2:double;var Theta2:double;var Phi2:double);
{-----------------------------------------------------------------------------}
{ Given the directions of a vector, (Theta1,Phi1), at certain position,       }
{ (Lat1,Long1), this routine will return the new value of theta and phi when  }
{ the vector is parallel translated to a new position (Lat2,Long2).           }
{-----------------------------------------------------------------------------}
var
  LongDiff,     {The longitude difference.}
  Hold1,        {Storage variable.}
  Hold2,        {Storage variable.}
  Hold3,        {Storage variable.}
  Hold4:double; {Storage variable.}
begin
  if ((Long1=Long2) and (Lat1=Lat2)) then
  begin
    Theta2:=Theta1;
    Phi2:=Phi1;
  end
  else
  begin
    Lat1:=DegToRad(Lat1);
    Long1:=DegToRad(Long1);
    Lat2:=DegToRad(Lat2);
    Long2:=DegToRad(Long2);
    LongDiff:=Long1-Long2;
    Hold4:=sin(Lat1)*sin(Theta1)*cos(Phi1)-cos(Lat1)*cos(Theta1);
    Hold1:=cos(Lat2)*(cos(Lat1)*sin(Theta1)*cos(Phi1)+sin(Lat1)*cos(Theta1))-
           sin(Lat2)*(sin(LongDiff)*sin(Theta1)*sin(Phi1)-cos(LongDiff)*Hold4);
    Hold2:=cos(LongDiff)*sin(Theta1)*sin(Phi1)+sin(LongDiff)*Hold4;
    Hold3:=sin(Lat2)*(cos(Lat1)*sin(Theta1)*cos(Phi1)+sin(Lat1)*cos(Theta1))+
           cos(Lat2)*(sin(LongDiff)*sin(Theta1)*sin(Phi1)-cos(LongDiff)*Hold4);
    if ((Hold3>1.0) or (Hold3<-1.0)) then raise EAstro.Create('The Hold3 value '+
      'passed to arccos was out of range.');
    Theta2:=arccos(Hold3);
    if sin(Theta2)=0.0 then Hold2:=1.0 else Hold2:=Hold2/sin(Theta2);
    if Hold2>1.0 then Hold2:=1.0;
    if Hold2<-1.0 then Hold2:=-1.0;
    if Hold1>=0 then Phi2:=arcsin(Hold2);
    if Hold1<0 then Phi2:=Pi-arcsin(Hold2);
    if Phi2<0 then Phi2:=Phi2+2*Pi;
  end;
end; {End of NewThetaPhi.}

{$WARNINGS OFF}
end. {End of gb_astro unit.}
{$WARNINGS ON}
