//@***h* BDBP/gb_datetime.pas
// 
// NAME
//   gb_datetime.pas
//
// PURPOSE
//  Defines a number of routines for manipulating dates and times.
//
// DESCRIPTION
//  This unit defines a number of routines for manipulating dates and times.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_datetime;

{$I UNITDIREC.INC}

interface

uses
  sysutils;

const
  month:array[1..12] of string[3]=('Jan','Feb','Mar','Apr','May','Jun',
                                   'Jul','Aug','Sep','Oct','Nov','Dec');
  FullMonth:array[1..12] of string[9]=('January','February','March','April',
    'May','June','July','August','September','October','November','December');
  weekday:array[1..7] of string[9]=('Monday','Tuesday','Wednesday','Thursday',
                                    'Friday','Saturday','Sunday');

type
  TTimeRec=record {This is the standard time record.}
    Hour,      {The Hour.}
    Min,       {The Minute.}
    Sec,       {The Second.}
    Hund:word; {The hundredths of a second.}
  end;
  PTimeRec=^TTimeRec;
  TDateRec=record {This is the standard date record.}
    Year,      {The Year.}
    Month,     {The Month.}
    Date:word; {The Date in the Month.}
  end;
  PDateRec=^TDateRec;

  EDateTime=class(Exception);

var
  NullDate:TDateRec;
  NullTime:TTimeRec;

{Here follows an alphabetical list of the routines available in this unit.}
procedure AddMinutes(var Date:TDateRec;var Time:TTimeRec;AddMin:double);
procedure AddNDays(NumDays:longint;var Date:TDateRec);
procedure AddTime(var Time:TTimeRec;HundSec:double);
procedure AssignDate(var Date:TDateRec;AYear,AMonth,ADate:word);
function  CalcDate(Year,Day:word):word;
procedure CalcDateFromNumDays(StartYear,StartDay:word;NumDays:longint;
  var Year:word;var Day:word);
function  CalcDay(Year,Mon,Date:word):word;
function  CalcMonth(Year,Day:word):word;
function  CalcNumDays(Year1,Day1,Year2,Day2:word):longint;
function  CalcTimeDiff(Time1,Time2:TTimeRec):double;
function  CompareDates(Date1,Date2:TDateRec):boolean;
procedure ConvertToUT(var Date:TDateRec;var Time:TTimeRec;TZone:integer);
function  ConvertDateTimeToDecYear(DateTime:TDateTime):double;
function  ConvertDecYearToDateTime(DecYear:double):TDateTime;
function  DateAfter(ChkDate,TheDate:TDateRec):boolean;
function  DateBetween(ChkDate,StartDate,EndDate:TDateRec):boolean;
function  DateTimeBetween(Test,FirstDT,LastDT:TDateTime):boolean;
function  DateRecToDateStr(Date:TDateRec;FormatStr:string):string;
function  DateStrToDateRec(DateStr,FormatStr:string):TDateRec;
function  DateStrToDateTime(DateStr,FormatStr:string):TDateTime;
function  DayInMonth(Year,Mon:word):byte;
procedure DecDate(var Date:TDateRec);
procedure IncDate(var Date:TDateRec);
function  LeapYear(Year:word):boolean;
procedure SubtractNDays(NumDays:longint;var Date:TDateRec);
function  TimeBetween(ChkDate,StartDate,EndDate:TDateRec;ChkTime,StartTime,
  EndTime:TTimeRec):boolean;
function  TimeStr(Time:TTimeRec):string;
function  TimeStr2(Time:TTimeRec):string;
procedure WaitNMinutes(NumMin:word);
procedure WaitNSeconds(NumSec:word);

{$IFDEF DELPHI5}
function DayInYear(Year:word):word;
function DaysInAYear(Year:word):word;
{$ENDIF}

{The following procedures only work with delphi 6.0 and 7.0 professional.}
{$IFNDEF DELPHI5}
function  DateTimeToStr(DateTime:TDateTime;FormatStr:string):string;
{$ENDIF}

{Old code formatting}
{DateStr yyyy-mm-dd}
{DateStr2 yy/mm/dd}
{DateStr3 yymmdd}
{DateStr4 dd mmm yyyy}
{DateStr5 yyyymmdd}

implementation

uses
  {$IFNDEF DELPHI5}
  DateUtils,
  {$ENDIF}

  gb_math,
  gb_strings;

procedure AddMinutes(var Date:TDateRec;var Time:TTimeRec;AddMin:double);
{-----------------------------------------------------------------------------}
{ This routine adds `Min' minutes to the time specified by `Time', updating   }
{ the date passed via `Date' if the resultant time passes midnight.           }
{-----------------------------------------------------------------------------}
var
  HundSec, {The 100ths of a second to be added.}
  Hold:double; {Storage variable.}
  Hour,Min,Sec,Hund:word; {The time to be added.}
begin
  if AddMin<0.0 then halt;
  HundSec:=AddMin*6000.0;
  Hour:=trunc(HundSec/360000);
  Hold:=Hour*1.0;
  HundSec:=HundSec-Hold*360000;
  Min:=trunc(HundSec/6000);
  Hold:=Min*1.0;
  HundSec:=HundSec-Hold*6000;
  Sec:=trunc(HundSec/100);
  Hold:=Sec*1.0;
  HundSec:=HundSec-Hold*100;
  Hund:=trunc(HundSec);
  Time.Hour:=Time.Hour+Hour;
  Time.Min:=Time.Min+Min;
  Time.Sec:=Time.Sec+Sec;
  Time.Hund:=Time.Hund+Hund;
  if Time.Hund>99 then
  begin
    Time.Hund:=Time.Hund-100;
    inc(Time.Sec);
  end;
  if Time.Sec>59 then
  begin
    Time.Sec:=Time.Sec-60;
    inc(Time.Min);
  end;
  if Time.Min>59 then
  begin
    Time.Min:=Time.Min-60;
    inc(Time.Hour);
  end;
  while Time.Hour>23 do
  begin
    Time.Hour:=Time.Hour-24;
    inc(Date.Date);
  end;
  while Date.Date>DayInMonth(Date.Year,Date.Month) do
  begin
    Date.Date:=Date.Date-DayInMonth(Date.Year,Date.Month);
    inc(Date.Month);
  end;
  while Date.Month>12 do
  begin
    Date.Month:=Date.Month-12;
    inc(Date.Year);
  end;
end; {End of AddMinutes.}

procedure AddNDays(NumDays:longint;var Date:TDateRec);
{-----------------------------------------------------------------------------}
{ This procedure adds `NumDays' to the date passed via `Date'.                }
{-----------------------------------------------------------------------------}
var
  TempDate:longint; {Temporary storage variable for the date.}
begin
  TempDate:=Date.Date;
  TempDate:=TempDate+NumDays;
  while TempDate>DayInMonth(Date.Year,Date.Month) do
  begin
    TempDate:=TempDate-DayInMonth(Date.Year,Date.Month);
    inc(Date.Month);
    if Date.Month>12 then
    begin
      inc(Date.Year);
      Date.Month:=1;
    end;
  end;
  while TempDate<1 do
  begin
    dec(Date.Month);
    if Date.Month<1 then
    begin
      dec(Date.Year);
      Date.Month:=12;
    end;
    TempDate:=TempDate+DayInMonth(Date.Year,Date.Month);
  end;
  Date.Date:=TempDate;
end; {End of AddNDays.}

procedure AddTime(var Time:TTimeRec;HundSec:double);
{-----------------------------------------------------------------------------}
{ This routine adds `HundSec' hundredths of a second to Time.                 }
{-----------------------------------------------------------------------------}
var
  Hold:double; {Temporary storage variable.}
  Hour,Min,Sec,Hund:word; {The time to be added.}
begin
  Hour:=trunc(HundSec/360000);
  Hold:=Hour*1.0;
  HundSec:=HundSec-Hold*360000;
  Min:=trunc(HundSec/6000);
  Hold:=Min*1.0;
  HundSec:=HundSec-Hold*6000;
  Sec:=trunc(HundSec/100);
  Hold:=Sec*1.0;
  HundSec:=HundSec-Hold*100;
  Hund:=trunc(HundSec);
  Time.Hour:=Time.Hour+Hour;
  Time.Min:=Time.Min+Min;
  Time.Sec:=Time.Sec+Sec;
  Time.Hund:=Time.Hund+Hund;
  if Time.Hund>99 then
  begin
    Time.Hund:=Time.Hund-100;
    inc(Time.Sec);
  end;
  if Time.Sec>59 then
  begin
    Time.Sec:=Time.Sec-60;
    inc(Time.Min);
  end;
  if Time.Min>59 then
  begin
    Time.Min:=Time.Min-60;
    inc(Time.Hour);
  end;
  if Time.Hour>23 then Time.Hour:=Time.Hour-24;
end; {End of AddTime.}

procedure AssignDate(var Date:TDateRec;AYear,AMonth,ADate:word);
{-----------------------------------------------------------------------------}
{ Assigns year, month and date to Date.                                       }
{-----------------------------------------------------------------------------}
begin
  with Date do
  begin
    Year:=AYear;
    Month:=AMonth;
    Date:=ADate;
  end;
end; {End of AssignDate.}

function CalcDate(Year,Day:word):word;
{-----------------------------------------------------------------------------}
{ Given the year (`Year') and the day number in the year (`Day'), this rout-  }
{ ine calculates the date in the month.                                       }
{-----------------------------------------------------------------------------}
begin
  Result:=65535;
  if (LeapYear(Year) and (Day=60)) then Result:=29 else
  begin
    if (LeapYear(Year) and (Day>=60)) then dec(Day);
    case Day of
      1..31:Result:=Day;
      32..59:Result:=Day-31;
      60..90:Result:=Day-59;
      91..120:Result:=Day-90;
      121..151:Result:=Day-120;
      152..181:Result:=Day-151;
      182..212:Result:=Day-181;
      213..243:Result:=Day-212;
      244..273:Result:=Day-243;
      274..304:Result:=Day-273;
      305..334:Result:=Day-304;
      335..365:Result:=Day-334;
    end;
  end;
end; {End of CalcDate.}

procedure CalcDateFromNumDays(StartYear,StartDay:word;NumDays:longint;
  var Year:word;var Day:word);
{-----------------------------------------------------------------------------}
{ This routine returns the Year and Day which is NumDays after StartYear and  }
{ StartDay.                                                                   }
{-----------------------------------------------------------------------------}
var
  HoldDay:longint; {Temporary storage variable.}
begin
  Year:=StartYear;
  HoldDay:=StartDay;
  HoldDay:=HoldDay+NumDays;
  while HoldDay>DaysInAYear(Year) do
  begin
    HoldDay:=HoldDay-DaysInAYear(Year);
    inc(Year);
  end;
  while HoldDay<1 do
  begin
    dec(Year);
    HoldDay:=HoldDay+DaysInAYear(Year);
  end;
  Day:=HoldDay;
end; {End of CalcDateFromNumDays.}

function CalcDay(Year,Mon,Date:word):word;
{-----------------------------------------------------------------------------}
{ Given the year, month and date, this routine calculates the day number in   }
{ the year.                                                                   }
{-----------------------------------------------------------------------------}
begin
  Result:=65535;
  case Mon of
    1:Result:=0;
    2:Result:=31;
    3:Result:=59;
    4:Result:=90;
    5:Result:=120;
    6:Result:=151;
    7:Result:=181;
    8:Result:=212;
    9:Result:=243;
    10:Result:=273;
    11:Result:=304;
    12:Result:=334;
  end;
  if (LeapYear(Year) and (Mon>2)) then inc(Result);
  Result:=Result+Date;
end; {End of CalcDay.}

function CalcMonth(Year,Day:word):word;
{-----------------------------------------------------------------------------}
{ Given the year and day number in the year, this routine calculates the      }
{ month number.                                                               }
{-----------------------------------------------------------------------------}
begin
  Result:=65535;
  if (LeapYear(Year) and (Day>=60)) then Day:=Day-1;
  case Day of
    1..31:Result:=1;
    32..59:Result:=2;
    60..90:Result:=3;
    91..120:Result:=4;
    121..151:Result:=5;
    152..181:Result:=6;
    182..212:Result:=7;
    213..243:Result:=8;
    244..273:Result:=9;
    274..304:Result:=10;
    305..334:Result:=11;
    335..365:Result:=12;
  end;
end; {End of CalcMonth.}

function CalcNumDays(Year1,Day1,Year2,Day2:word):longint;
{-----------------------------------------------------------------------------}
{ This function returns the number of days between year1,day1 and year2,day2. }
{ It takes into account leap years and will also return a negative answer if  }
{ the end date occurs before the start date.                                  }
{-----------------------------------------------------------------------------}
var
  I:integer; {Loop control variable.}
  Total:longint; {The current total number of days.}
  NegAns:boolean; {Set to true if the answer must be negative.}
begin
  if (Year1>Year2) then
  begin
    NegAns:=true;
    SwapWord(Year1,Year2);
    SwapWord(Day1,Day2);
  end
  else NegAns:=false;
  if (Year1=Year2) then CalcNumDays:=Day2-Day1
  else
  begin
    if (Year1 mod 4=0) then Total:=366-Day1 else Total:=365-Day1;
    for I:=Year1+1 to Year2-1 do
      if LeapYear(I) then Total:=Total+366 else Total:=Total+365;
    Total:=Total+Day2;
    if NegAns then CalcNumDays:=-1*Total else CalcNumDays:=Total;
  end;             
end; {End of CalcNumDays.}

function CalcTimeDiff(Time1,Time2:TTimeRec):double;
{------------------------------------------------------------------------------}
{ This routine returns the time (in hundreths of a second) between Time1 and   }
{ Time2.                                                                       }
{------------------------------------------------------------------------------}
var
  HourDiff, {The difference in hours.}
  MinDiff,  {The difference in minutes.}
  SecDiff,  {The difference in seconds.}
  HundDiff:double; {The difference in 100th of seconds.}
begin
  HourDiff:=Time2.Hour;
  HourDiff:=HourDiff-Time1.Hour;
  MinDiff:=Time2.Min;
  MinDiff:=MinDiff-Time1.Min;
  SecDiff:=Time2.Sec;
  SecDiff:=SecDiff-Time1.Sec;
  HundDiff:=Time2.Hund;
  HundDiff:=HundDiff-Time1.Hund;
  CalcTimeDiff:=HourDiff*360000+MinDiff*6000+SecDiff*100+HundDiff;
end; {End of CalcTimeDiff.}

function CompareDates(Date1,Date2:TDateRec):boolean;
{------------------------------------------------------------------------------}
{ If Date1 is the same as Date 2 then the function returns true, otherwise it  }
{ returns false.                                                               }
{------------------------------------------------------------------------------}
begin
  if ((Date1.Year=Date2.Year) and (Date1.Month=Date2.Month) and
      (Date1.Date=Date2.Date)) then CompareDates:=true
  else CompareDates:=false;
end; {End of CompareDates.}

procedure ConvertToUT(var Date:TDateRec;var Time:TTimeRec;TZone:integer);
{------------------------------------------------------------------------------}
{ This routine converts the date and time specified by `Date' and `Time' from  }
{ local apparent time to UT.                                                   }
{------------------------------------------------------------------------------}
var
  Hold:integer;
begin
  Hold:=Time.Hour;
  Hold:=Hold-TZone;
  if Hold<0 then
  begin
    Time.Hour:=Hold+24;
    dec(Date.Date);
    if Date.Date<1 then
    begin
      dec(Date.Month);
      if Date.Month<1 then
      begin
        dec(Date.Year);
        Date.Month:=12;
      end;
      Date.Date:=DayInMonth(Date.Year,Date.Month);
    end;
  end
  else Time.Hour:=Hold;
  if Time.Hour>23 then
  begin
    Time.Hour:=Time.Hour-24;
    inc(Date.Date);
    if Date.Date>DayInMonth(Date.Year,Date.Month) then
    begin
      Date.Date:=1;
      inc(Date.Month);
      if Date.Month>12 then
      begin
        Date.Month:=1;
        inc(Date.Year);
      end;
    end;
  end;
end; {End of ConvertToUT.}

function ConvertDateTimeToDecYear(DateTime:TDateTime):double;
var
  StartDateTime:TDateTime;
  Year:word;
begin
  Year:=YearOf(DateTime);
  StartDateTime:=StartOfTheYear(DateTime);
  Result:=Year+(DateTime-StartDateTime)/DaysInAYear(Year);
end;

function ConvertDecYearToDateTime(DecYear:double):TDateTime;
var
  Year:word;
begin
  Year:=trunc(DecYear);
  Result:=EncodeDate(Year,1,1)+frac(DecYear)*DaysInAYear(Year);
end;

function DateAfter(ChkDate,TheDate:TDateRec):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if `TheDate' is after `ChkDate', otherwise it     }
{ returns false. If `ChkDate' and `TheDate' are on the same day then the       }
{ function returns false.                                                      }
{------------------------------------------------------------------------------}
begin
  Result:=false;
  if TheDate.Year>ChkDate.Year then Result:=true;
  if TheDate.Year<ChkDate.Year then Result:=false;
  if TheDate.Year=ChkDate.Year then
  begin
    if TheDate.Month>ChkDate.Month then Result:=true;
    if TheDate.Month<ChkDate.Month then Result:=false;
    if TheDate.Month=ChkDate.Month then
    begin
      if TheDate.Date>ChkDate.Date then Result:=true;
      if TheDate.Date<ChkDate.Date then Result:=false;
      if TheDate.Date=ChkDate.Date then Result:=false;
    end;
  end;
end; {End of DateAfter.}

function DateBetween(ChkDate,StartDate,EndDate:TDateRec):boolean;
{-----------------------------------------------------------------------------}
{ This function returns true if the date specified by `ChkDate' falls between }
{ the start date specified by `StartDate' and the end date specified by       }
{ `EndDate'.                                                                  }
{-----------------------------------------------------------------------------}
var
  Day1,Day2:word;    {Temporary storage variables to hold day number in the year.}
begin
  Result:=true;
  Day1:=CalcDay(StartDate.Year,StartDate.Month,StartDate.Date);
  Day2:=CalcDay(ChkDate.Year,ChkDate.Month,ChkDate.Date);
  if CalcNumDays(StartDate.Year,Day1,ChkDate.Year,Day2)<0 then Result:=false;
  Day1:=CalcDay(EndDate.Year,EndDate.Month,EndDate.Date);
  if CalcNumDays(ChkDate.Year,Day2,EndDate.Year,Day1)<0 then Result:=false;
end; {End of DateBetween.}

function DateTimeBetween(Test,FirstDT,LastDT:TDateTime):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if test falls between the dates and times         }
{ specified by FirstDT and LastDT.                                             }
{------------------------------------------------------------------------------}
begin
  if ((Test>=FirstDT) and (Test<=LastDT)) then Result:=true else Result:=false;
end; {End of DateTimeBetween.}

function DateRecToDateStr(Date:TDateRec;FormatStr:string):string;
{-----------------------------------------------------------------------------}
{ Given the date as a record, and the format of the string into which it      }
{ must be converted, this routine creates the appropriate date string e.g.    }
{ passing 'Date' as (2002,3,24) and 'FormatStr' as 'YYYY-MM-DD' would         }
{ produce '2002-03-24'. The substrings recognised in FormatStr are:           }
{   'DD' is replaced by the date.                                             }
{   'DDD' is replaced by the day number in the year.                          }
{   'MM' is replaced by the two digit month number.                           }
{   'MMM' is replaced by the 3 letter abbreviated month name.                 }
{   'MMMM' is replaced by the full month name.                                }
{   'YY' is replaced by the two digit year number.                            }
{   'YYYY' is replaced by the four digit year number.                         }
{-----------------------------------------------------------------------------}
var
  I:word; {The index in FormatStr of the indicator string.}
  WorkStr:string; {General purpose working string.}
begin
  while pos('DDD',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(CalcDay(Date.Year,Date.Month,Date.Date),3,true);
    I:=pos('DDD',FormatStr);
    Delete(FormatStr,I,3);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('DD',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(Date.Date,2,true);
    WorkStr:=Replace(WorkStr,' ','0');
    I:=pos('DD',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MMMM',FormatStr)<>0 do
  begin
    WorkStr:=string(FullMonth[Date.Month]);
    I:=pos('MMMM',FormatStr);
    Delete(FormatStr,I,4);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MMM',FormatStr)<>0 do
  begin
    WorkStr:=string(month[Date.Month]);
    I:=pos('MMM',FormatStr);
    Delete(FormatStr,I,3);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MM',FormatStr)<>0 do
  begin
    str(Date.Month:2,WorkStr);
    WorkStr:=Replace(WorkStr,' ','0');
    I:=pos('MM',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('YYYY',FormatStr)<>0 do
  begin
    WorkStr:=inttostr(Date.Year);
    I:=pos('YYYY',FormatStr);
    Delete(FormatStr,I,4);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('YY',FormatStr)<>0 do
  begin
    str(Date.Year mod 100:2,WorkStr);
    WorkStr:=Replace(WorkStr,' ','0');
    I:=pos('YY',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  Result:=FormatStr;
end; {End of DateRecToDateStr.}

{$IFNDEF DELPHI5}
function DateTimeToStr(DateTime:TDateTime;FormatStr:string):string;
{------------------------------------------------------------------------------}
{ Given the date and time as a TDateTime variable, and the format of the       }
{ string into which it must be converted, this routine creates the appropriate }
{ date and time string. The substrings recognised in DateTimeToStr are:        }
{   'DD' is replaced by the date.                                              }
{   'DDD' is replaced by the day number in the year.                           }
{   'MM' is replaced by the two digit month number.                            }
{   'MMM' is replaced by the 3 letter abbreviated month name.                  }
{   'MMMM' is replaced by the full month name.                                 }
{   'YY' is replaced by the two digit year number.                             }
{   'YYYY' is replaced by the four digit year number.                          }
{   'HH' is replaced by the 2 digit hour.                                      }
{   'TT' is replaced by the 2 digit minute.                                    }
{   'SS' is replaced by the 2 digit second.                                    }
{------------------------------------------------------------------------------}
var
  I:word; {The index in FormatStr of the indicator string.}
  WorkStr:string; {General purpose working string.}
begin
  while pos('DDD',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(DayOfTheYear(DateTime),3,true);
    I:=pos('DDD',FormatStr);
    Delete(FormatStr,I,3);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('DD',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(DayOf(DateTime),2,true);
    I:=pos('DD',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MMMM',FormatStr)<>0 do
  begin
    WorkStr:=FullMonth[MonthOf(DateTime)];
    I:=pos('MMMM',FormatStr);
    Delete(FormatStr,I,4);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MMM',FormatStr)<>0 do
  begin
    WorkStr:=month[MonthOf(DateTime)];
    I:=pos('MMM',FormatStr);
    Delete(FormatStr,I,3);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('MM',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(MonthOf(DateTime),2,true);
    I:=pos('MM',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('YYYY',FormatStr)<>0 do
  begin
    WorkStr:=inttostr(YearOf(DateTime));
    I:=pos('YYYY',FormatStr);
    Delete(FormatStr,I,4);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('YY',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(YearOf(DateTime) mod 100,2,true);
    I:=pos('YY',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('HH',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(HourOf(DateTime),2,true);
    I:=pos('HH',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('TT',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(MinuteOf(DateTime),2,true);
    I:=pos('TT',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  while pos('SS',FormatStr)<>0 do
  begin
    WorkStr:=inttostrf(SecondOf(DateTime),2,true);
    I:=pos('SS',FormatStr);
    Delete(FormatStr,I,2);
    Insert(WorkStr,FormatStr,I);
  end;
  Result:=FormatStr;
end; {End of DateTimeToStr.}
{$ENDIF}

function DateStrToDateRec(DateStr,FormatStr:string):TDateRec;
var
  Day,    {The day number in the year.}
  I:word; {General purpose loop control variable.}
  WorkStr:string; {General purpose working string.}
begin
  Result:=NullDate;
  if pos('MMMM',FormatStr)>0 then
  begin
    for I:=1 to 12 do
    begin
      WorkStr:=copy(DateStr,pos('MMMM',FormatStr),length(FullMonth[I]));
      if Uppercase(WorkStr)=Uppercase(FullMonth[I]) then Result.Month:=I;
    end;
    if Result.Month<>65535 then
    begin
      I:=pos('MMMM',FormatStr);
      Delete(DateStr,I,length(FullMonth[Result.Month]));
      Delete(FormatStr,I,4);
    end;
  end;
  if ((Result.Month=65535) and (pos('MMM',FormatStr)>0)) then
    for I:=1 to 12 do
    begin
      WorkStr:=copy(DateStr,pos('MMM',FormatStr),3);
      if Uppercase(WorkStr)=Uppercase(month[I]) then Result.Month:=I;
    end;
  if ((Result.Month=65535) and (pos('MM',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('MM',FormatStr),2);
    try Result.Month:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid month number.') end;
  end;
  if pos('YYYY',FormatStr)>0 then
  begin
    WorkStr:=copy(DateStr,pos('YYYY',FormatStr),4);
    try Result.Year:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid value for the year.') end;
  end;
  if ((Result.Year=65535) and (pos('YY',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('YY',FormatStr),2);
    try Result.Year:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid value for the year.') end;
    Result.Year:=Result.Year+2000;
  end;
  if pos('DDD',FormatStr)>0 then
  begin
    WorkStr:=copy(DateStr,pos('DDD',FormatStr),3);
    try Day:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid day number of the year.') end;
    if Result.Year<>65535 then
    begin
      Result.Month:=CalcMonth(Result.Year,Day);
      Result.Date:=CalcDate(Result.Year,Day);
    end;
  end;
  if ((Result.Date=65535) and (pos('DD',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('DD',FormatStr),2);
    try Result.Date:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid date.') end;
  end;
  if ((Result.Year=NullDate.Year) or (Result.Month=NullDate.Month)
    or (Result.Date=NullDate.Date)) then raise EDateTime.Create(
      'the date string couldn''t be decoded properly.');
end; {End of DateStrToDateRec.}

function DateStrToDateTime(DateStr,FormatStr:string):TDateTime;
var
  Year,
  Mon,
  Date,
  Day,    {The day number in the year.}
  I:word; {General purpose loop control variable.}
  WorkStr:string; {General purpose working string.}
begin
  Year:=65535;
  Mon:=65535;
  Date:=65535;
  if pos('MMMM',FormatStr)>0 then
  begin
    for I:=1 to 12 do
    begin
      WorkStr:=copy(DateStr,pos('MMMM',FormatStr),length(FullMonth[I]));
      if Uppercase(WorkStr)=Uppercase(FullMonth[I]) then Mon:=I;
    end;
    if Mon<>65535 then
    begin
      I:=pos('MMMM',FormatStr);
      Delete(DateStr,I,length(FullMonth[Mon]));
      Delete(FormatStr,I,4);
    end;
  end;
  if ((Mon=65535) and (pos('MMM',FormatStr)>0)) then
    for I:=1 to 12 do
    begin
      WorkStr:=copy(DateStr,pos('MMM',FormatStr),3);
      if Uppercase(WorkStr)=Uppercase(month[I]) then Mon:=I;
    end;
  if ((Mon=65535) and (pos('MM',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('MM',FormatStr),2);
    try Mon:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid month number.') end;
  end;
  if pos('YYYY',FormatStr)>0 then
  begin
    WorkStr:=copy(DateStr,pos('YYYY',FormatStr),4);
    try Year:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid value for the year.') end;
  end;
  if ((Year=65535) and (pos('YY',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('YY',FormatStr),2);
    try Year:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid value for the year.') end;
    Year:=Year+2000;
  end;
  if pos('DDD',FormatStr)>0 then
  begin
    WorkStr:=copy(DateStr,pos('DDD',FormatStr),3);
    try Day:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid day number of the year.') end;
    if Year<>65535 then
    begin
      Mon:=CalcMonth(Year,Day);
      Date:=CalcDate(Year,Day);
    end;
  end;
  if ((Date=65535) and (pos('DD',FormatStr)>0)) then
  begin
    WorkStr:=copy(DateStr,pos('DD',FormatStr),2);
    try Date:=strtoint(WorkStr) except on EConvertError do
      raise EDateTime.Create(WorkStr+' is not a valid date.') end;
  end;
  if ((Year=65535) or (Mon=65535) or (Date=65535)) then raise EDateTime.Create(
      'the date string couldn''t be decoded properly.');
  Result:=EncodeDate(Year,Mon,Date);
end; {End of DateStrToDateRec.}

function DayInMonth(Year,Mon:word):byte;
{-----------------------------------------------------------------------------}
{ This routine returns the number of days in the month number `Mon'. The year }
{ is also passed as a parameter to check for leap years.                      }
{-----------------------------------------------------------------------------}
begin
  DayInMonth:=255;
  case Mon of
    1,3,5,7,8,10,12:DayInMonth:=31;
    2:begin
        if LeapYear(Year) then DayInMonth:=29
        else DayInMonth:=28;
      end;
    4,6,9,11:DayInMonth:=30;
  end;
end; {End of DayInMonth.}

{$IFDEF DELPHI5}
function DayInYear(Year:word):word;
{-----------------------------------------------------------------------------}
{ This routine returns the number of days in the year specified by the passed }
{ parameter `Year'.                                                           }
{-----------------------------------------------------------------------------}
begin
  if LeapYear(Year) then Result:=366 else Result:=365;
end; {End of DayInYear.}

function DaysInAYear(Year:word):word;
{-----------------------------------------------------------------------------}
{ This routine returns the number of days in the year specified by the passed }
{ parameter `Year'.                                                           }
{-----------------------------------------------------------------------------}
begin
  if LeapYear(Year) then Result:=366 else Result:=365;
end; {End of DaysInAYear.}
{$ENDIF}

procedure DecDate(var Date:TDateRec);
{------------------------------------------------------------------------------}
{ This routine decrements the date specified by 'Date' by one day.             }
{------------------------------------------------------------------------------}
begin
  with Date do
  begin
    dec(Date);
    if Date<1 then
    begin
      dec(Month);
      if Month<1 then
      begin
        Month:=12;
        dec(Year);
      end;
      Date:=DayInMonth(Year,Month);
    end;
  end;
end; {End of DecDate.}

procedure GregsDecodeDate(YearStr,MonthStr,DateStr:string;var Date:TDateRec);
{------------------------------------------------------------------------------}
{ Given the string equivalent of the year (e.g. '1999'), the string equivalent }
{ of the month (e.g. 'Jan') and the string equivalent of the date (e.g. '31'), }
{ this routine decodes the string values into the fields of 'Date'. If the     }
{ conversion of string values to word values is not successful and EDateTime   }
{ exception is raised.                                                         }
{------------------------------------------------------------------------------}
var
  I:word; {Loop control variable through months.}
begin
  try Date.Year:=strtoint(YearStr) except on EConvertError do
    raise EDateTime.Create(YearStr+' is not a valid value for the year.') end;
  try Date.Date:=strtoint(DateStr) except on EConvertError do
    raise EDateTime.Create(DateStr+' is not a valid value for the date.') end;
  Date.Month:=0;
  for I:=1 to 12 do if UpperCase(MonthStr)=UpperCase(Month[I]) then Date.Month:=I;
  if Date.Month=0 then raise EDateTime.Create(MonthStr+' is not a valid value for the month.');
end; {End of GregsDecodeDate.}

procedure IncDate(var Date:TDateRec);
{------------------------------------------------------------------------------}
{ This routine increments the date specified by 'Date' by one day.             }
{------------------------------------------------------------------------------}
begin
  with Date do
  begin
    inc(Date);
    if Date>DayInMonth(Year,Month) then
    begin
      Date:=1;
      inc(Month);
      if Month>12 then
      begin
        Month:=1;
        inc(Year);
      end;
    end;
  end;
end; {End of IncDate.}

function LeapYear(Year:word):boolean;
begin
  if (Year mod 4<>0) then
    LeapYear:=false
  else if (Year mod 100<>0) then
    LeapYear:=true
  else if (Year mod 400<>0) then
    LeapYear:=false
  else LeapYear:=true;
end; {End of LeapYear.}

procedure SubtractNDays(NumDays:longint;var Date:TDateRec);
{-----------------------------------------------------------------------------}
{ This procedure subtracts `NumDays' from the date passed via `Date' and      }
{ returns the resultant date.                                                 }
{-----------------------------------------------------------------------------}
var
  TempDate:longint; {Temporary storage variable for date.}
begin
  TempDate:=Date.Date;
  TempDate:=TempDate-NumDays;
  while TempDate<1 do
  begin
    dec(Date.Month);
    if Date.Month<1 then
    begin
      dec(Date.Year);
      Date.Month:=12;
    end;
    TempDate:=TempDate+DayInMonth(Date.Year,Date.Month);
  end;
  Date.Date:=TempDate;
end; {End of SubtractNDays.}

function TimeBetween(ChkDate,StartDate,EndDate:TDateRec;ChkTime,StartTime,
  EndTime:TTimeRec):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if the date and time specified by `ChkDate' and   }
{ `ChkTime' fall between the start and end dates and times specified by `Start-}
{ Date', `StartTime', `EndDate' and `EndTime'. Otherwise it returns false.     }
{------------------------------------------------------------------------------}
var
  ReturnVal:boolean; {The value returned by the function.}
  Day1,Day2:word;    {Temporary storage variables to hold day number in the year.}
begin
  ReturnVal:=true;
  Day1:=CalcDay(StartDate.Year,StartDate.Month,StartDate.Date);
  Day2:=CalcDay(ChkDate.Year,ChkDate.Month,ChkDate.Date);
  if CalcNumDays(StartDate.Year,Day1,ChkDate.Year,Day2)<0 then ReturnVal:=false;
  Day1:=CalcDay(EndDate.Year,EndDate.Month,EndDate.Date);
  if CalcNumDays(ChkDate.Year,Day2,EndDate.Year,Day1)<0 then ReturnVal:=false;
  if ReturnVal then {either at start date or end date or in between}
  begin
    if ((CalcNumDays(ChkDate.Year,Day2,EndDate.Year,Day1)=0) and
        (CalcTimeDiff(ChkTime,EndTime)<0)) then ReturnVal:=false;
    Day1:=CalcDay(StartDate.Year,StartDate.Month,StartDate.Date);
    if ((CalcNumDays(StartDate.Year,Day1,ChkDate.Year,Day2)=0) and
        (CalcTimeDiff(StartTime,ChkTime)<0)) then ReturnVal:=false;
  end;
  TimeBetween:=ReturnVal;
end; {End of TimeBetween.}

function TimeStr(Time:TTimeRec):string;
{-----------------------------------------------------------------------------}
{ Given the time as a record, this routine creates a time string in the form  }
{ hh:mm:ss.hh                                                                 }
{-----------------------------------------------------------------------------}
var
  HourStr, {The string equivalent of the hour.}
  MinStr,  {The string equivalent of the minute.}
  SecStr,  {The string equivalent of the second.}
  HundStr:string; {The string equivalent of the hundreths of asecond.}
begin
  str(Time.Hour:2,HourStr);
  str(Time.Min:2,MinStr);
  str(Time.Sec:2,SecStr);
  str(Time.Hund:2,HundStr);
  TimeStr:=Replace(HourStr,' ','0')+':'+Replace(MinStr,' ','0')+':'+
           Replace(SecStr,' ','0')+'.'+Replace(HundStr,' ','0');
end; {End of TimeStr.}

function TimeStr2(Time:TTimeRec):string;
{-----------------------------------------------------------------------------}
{ Given the time as a record, this routine creates a time string in the form  }
{ hh:mm:ss                                                                    }
{-----------------------------------------------------------------------------}
var
  HourStr,       {The string equivalent of the hour.}
  MinStr,        {The string equivalent of the minute.}
  SecStr:string; {The string equivalent of the second.}
begin
  str(Time.Hour:2,HourStr);
  str(Time.Min:2,MinStr);
  str(Time.Sec:2,SecStr);
  Result:=Replace(HourStr,' ','0')+':'+Replace(MinStr,' ','0')+':'+Replace(SecStr,' ','0');
end; {End of TimeStr.}

procedure WaitNMinutes(NumMin:word);
{-----------------------------------------------------------------------------}
{ This routine waits for `NumMin' minutes.                                    }
{-----------------------------------------------------------------------------}
begin
  Sleep(NumMin*60000);
end; {End of WaitNMinutes.}

procedure WaitNSeconds(NumSec:word);
{-----------------------------------------------------------------------------}
{ This routine waits for `NumSec' seconds.                                    }
{-----------------------------------------------------------------------------}
begin
  Sleep(NumSec*1000);
end; {End of WaitNSeconds.}

{$WARNINGS OFF}
begin
  NullDate.Year:=65535;
  NullDate.Month:=65535;
  NullDate.Date:=65535;
  NullTime.Hour:=65535;
  NullTime.Min:=65535;
  NullTime.Sec:=65535;
  NullTime.Hund:=65535;
end. {End of datetime unit.}
{$WARNINGS ON}







