//@***h* BDBP/gb_files.pas
// 
// NAME
//   gb_files.pas
//
// PURPOSE
//  Holds a number of functions and procedures that are used in file manipulation.
//
// DESCRIPTION
//  This unit holds a number of functions and procedures that are used in file 
//  manipulation.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_files;

{$I UNITDIREC.INC}

interface

uses
  SysUtils,
  ComCtrls,
  Classes,
  DateUtils,
  Forms;

{Here is an alphabetical list of the routines available in this unit.}
function ConfirmDataLine(Linein:string):boolean;
function CopyFile(Source,Dest:string;ProgBar:TProgressBar):boolean;
function CompareFiles(Source1,Source2:string):boolean;
function CountDataLines(Filename:string):longint;
function CountFiles(FileMask,Directory:string):longint;
function CountHeaderLines(const FileName:string;const Mask:string=''''):Integer;
function CountLines(Filename:string):longint;
function DeleteAllFilesInDirectory(MaskStr:string):longint;
function DeleteHeader(var Source:textfile):word;
function FindJunkFile:string;
function FixFileEOL(const FileSource,FileDest:string;App:TApplication):byte;
function GenerateDirTree(Dir:char):string;overload;
function GenerateDirTree(Dir:string):TStringList;overload;
function GetFileDate(Filename:string):TDateTime;
function GetFileList(MaskStr:string;IncludeSubDirs:boolean=false):TStringList;
function GetFileSize(Filename:string):longint;
function GetFirstLine(TextFilename:string):string;
function GetPrompt(var Source:textfile;Prompt:string):string;
function GetShortFilename(Filename:string):string;
function GetYNPrompt(var Source:textfile;Prompt:string;var Value:boolean):byte;
function NeedsFixFileEOL(Filename:string; Encoding: TEncoding = nil):boolean;
function NumAvailFileHandles:byte;
function ParseDirectoryList(DirecStr:string):TStringList;
function ReadStringFromBinFile(var Source:file):string;
function SortTextFile(Filename:string;Column:byte):byte;
procedure WaitForFile(Filename:string);
procedure WriteStringToBinFile(Str:string;var Dest:file);

implementation

uses
  {These units are part of Borland Delphi.}
  Windows,
  {These are other general purpose units.}
  gb_Array,
  gb_DateTime,
  gb_strings;

{$HINTS OFF}
function ConfirmDataLine(Linein:string):boolean;
{------------------------------------------------------------------------------}
{ This function returns true if all words (strings separated by spaces) in     }
{ 'Linein' are numerical (return a code of zero from the val command) other-   }
{ wise it returns false. The line must contain at least 1 digit.               }
{------------------------------------------------------------------------------}
var
  ValCode,   {The error code from the val command.}
  I:integer; {Loop control variable through words.}
  Hold:double; {Temporary storage variable.}
  TestStr:string; {The word to be tested.}
begin
  if Linein='' then Result:=false else
  begin
    Result:=false;
    for I:=1 to length(Linein) do if CharInSet(Linein[I],['0'..'9']) then Result:=true;
    if Result=true then
    begin
      I:=0;
      repeat
        inc(I);
        TestStr:=GetWord(Linein,' ',I);
        if TestStr<>'' then
        begin
          val(TestStr,Hold,ValCode);
          if ValCode<>0 then Result:=false;
        end;
      until ((TestStr='') or (Result=false));
    end;
  end;
end; {End of ConfirmDataLine.}
{$HINTS ON}

function CopyFile(Source,Dest:string;ProgBar:TProgressBar):boolean;
{------------------------------------------------------------------------------}
{ This routine copies the file specified by name 'Source' to the file specif-  }
{ ied by name 'Dest'. It copies the file byte for byte and makes no assumpt-   }
{ ions about its format. It returns the value true if the copy was successful  }
{ or false if there were problems.                                             }
{------------------------------------------------------------------------------}
const
  UseWinAPI=false;
var
  Buffer:array[1..4096] of byte; {The data buffer.}
  File1,File2:file;              {The two file handles.}
  Numread,Numwrite:integer;      {The number of values transferred.}
  TotalBytesRead,                {The total number of bytes read so far.}
  SizeOfSource:longint;          {The size of the source file in bytes.}
begin
  if not FileExists(Source) then raise Exception.Create('the file named '+Source+' did not exist.');
  if UseWinAPI then
  begin
    Result:=Windows.CopyFile(PChar(Source),PChar(Dest),false);
  end
  else
  begin
    FileMode:=0;
    assignfile(File1,Source);
    reset(File1,1);
    SizeOfSource:=filesize(File1);
    assignfile(File2,Dest);
    rewrite(File2,1);
    TotalBytesRead:=0;
    if IOresult<>0 then Result:=false
    else
    begin
      repeat
        blockread(File1,Buffer,sizeof(Buffer),Numread);
        blockwrite(File2,Buffer,Numread,Numwrite);
        TotalBytesRead:=TotalBytesRead+Numread;
        if ProgBar<>nil then
        begin
          ProgBar.Position:=round(TotalBytesRead*100/SizeOfSource);
          Application.ProcessMessages;
        end;
      until (Numread=0) or (Numwrite<>Numread);
      closefile(File1);
      closefile(File2);
      if (Numwrite<>Numread) then CopyFile:=false else CopyFile:=true;
    end;
    if TotalBytesRead<>SizeOfSource then
    begin
      erase(File2);
      Result:=false;
    end;
    FileMode:=2;
  end;
end; {End of CopyFile routine.}

function CompareFiles(Source1,Source2:string):boolean;
var
  Buffer1,
  Buffer2:array[1..4096] of byte; {The data buffer.}
  File1,File2:file;               {The two file handles.}
  Numread1,
  Numread2,
  I:integer;
begin
  Result:=true;
  if not FileExists(Source1) then halt;
  if not FileExists(Source2) then halt;
  FileMode:=0;
  assignfile(File1,Source1);
  reset(File1,1);
  assignfile(File2,Source2);
  reset(File2,1);
  if IOresult<>0 then Result:=false
  else
  begin
    repeat
      blockread(File1,Buffer1,sizeof(Buffer1),Numread1);
      blockread(File2,Buffer2,sizeof(Buffer2),Numread2);
      if Numread1=Numread2 then
      begin
        for I:=1 to Numread1 do if Buffer1[I]<>Buffer2[I] then
          Result:=false;
      end
      else Result:=false;
    until ((Numread1=0) or (Numread2=0) or (Result=false));
    closefile(File1);
    closefile(File2);
  end;
  FileMode:=2;
end;

function CountDataLines(Filename:string):longint;
{------------------------------------------------------------------------------}
{ This routine returns the number of data lines in the file specified by       }
{ `Filename'. Header lines, preceded by ' characters, are automatically ig-    }
{ nored. The routine calls the ConfirmDataLine function, defined in ths unit,  }
{ to determine whether any non-header lines are data lines. The routine does   }
{ not check if the file is a text file. If the file does not exist, the rou-   }
{ tine returns a value of -1.                                                  }
{------------------------------------------------------------------------------}
var
  LineIn:string; {The line of text read in from the file.}
  Source:textfile; {The file handle for the input file.}
begin
  if not FileExists(Filename) then Result:=-1 else
  begin
    assignfile(Source,Filename);
    reset(Source);
    Result:=0;
    repeat
      readln(Source,LineIn);
      Linein:=Replace(Linein,#9,' ');
      if ((length(LineIn)<>0) and (LineIn[1]<>chr(39)) and ConfirmDataLine(Linein)) then inc(Result);
    until eof(Source);
    closefile(Source);
  end;
end; {End of CountDataLines.}

function CountFiles(FileMask,Directory:string):longint;
{------------------------------------------------------------------------------}
{ This routine counts the number of files matching the 'FileMask' mask in the  }
{ directory specified by 'Directory'.                                          }
{------------------------------------------------------------------------------}
var
  DirInfo:TSearchRec; {The search record used by findfirst and findnext.}
begin
  Result:=0;
  Directory:=EraseEndChar(Directory,'\');
  if findfirst(Directory+'\'+FileMask,faAnyFile,DirInfo)=0 then
  repeat inc(Result) until findnext(DirInfo)<>0;
  SysUtils.findclose(DirInfo);
end; {End of CountFiles.}

function CountLines(Filename:string):longint;
{------------------------------------------------------------------------------}
{ This routine returns the number of non-header lines or blank lines in the    }
{ file specified by `Filename'. A header line is a line that starts with the ' }
{ character, while a blank line is empty. The routine does not check if the    }
{ file is a text file. If the file does not exist, the routine returns a       }
{ value of -1.                                                                 }
{------------------------------------------------------------------------------}
var
  LineIn:string; {The line of text read in from the file.}
  Source:textfile; {The file handle for the input file.}
begin
  if not FileExists(Filename) then Result:=-1 else
  begin
    assignfile(Source,Filename);
    reset(Source);
    Result:=0;    repeat
      readln(Source,LineIn);
      if ((length(LineIn)<>0) and (LineIn[1]<>chr(39))) then inc(Result);
    until eof(Source);
    closefile(Source);
  end;
end; {End of CountLines.}

function CountHeaderLines(const FileName:string;const Mask:string=''''):Integer;
{------------------------------------------------------------------------------}
{ This routine returns the number of header lines or blank lines in the file   }
{ specified by `Filename'. A header line is a line that starts with the `Mask'-}
{ string. A header is assumed to start at the first no-blank line of the file  }
{ and consist of consecutive header-lines. The routine does not check if the   }
{ file is a text file. If the file does not exist, the routine returns a       }
{ value of -1.                                                                 }
{------------------------------------------------------------------------------}
var
  F: TextFile;
  Temp: string;
begin
  Result:=-1;
  AssignFile(F,FileName);
  try
    FileMode:=fmOpenRead;
    Reset(F);
    Result:=0;
    while not Eof(F) do
    begin
      ReadLn(F,Temp);
      if Length(Temp)>0 then
        if CompareStr(Copy(Temp,1,Length(Mask)),Mask)=0 then Inc(Result)
        else Break;
    end;
  finally
    CloseFile(F);
  end;
end; {End of CountHeaderLines.}

{$HINTS OFF}
function DeleteAllFilesInDirectory(MaskStr:string):longint;
{------------------------------------------------------------------------------}
{ This routine deletes all the files in the directory specified by MaskStr,    }
{ which can also be used to control which files are deleted. The fucntion      }
{ returns the number of files deleted.                                         }
{------------------------------------------------------------------------------}
var
  FileList:TStringList;
  I:longint;
begin
  FileList:=GetFileList(MaskStr);
  Result:=FileList.Count;
  try
    for I:=0 to FileList.Count-1 do SysUtils.DeleteFile(FileList[I]);
  finally
    FileList.Free;
  end;
end;
{$HINTS ON}

function DeleteHeader(var Source:textfile):word;
{------------------------------------------------------------------------------}
{ This routine is used to kill header lines from the top of a text file. The   }
{ header lines will always start with the character ' to identify the lines to }
{ be killed. The file handle for the text data file is passed as a parameter.  }
{ Note that this program only skips over header lines at the top of the file.  }
{------------------------------------------------------------------------------}
var
  I:integer; {Loop control variable.}
  Linein:string; {The line of text read in from the file.}
begin
  Result:=0;
  reset(Source);
  repeat
    inc(Result);
    readln(Source,Linein)
  until ((Linein[1]<>chr(39)) or (eof(Source)));
  if not eof(Source) then dec(Result);
  reset(Source);
  for I:=1 to Result do readln(Source);
  DeleteHeader:=Result;
end; {End of DeleteHeader.}

function FindJunkFile:string;
{------------------------------------------------------------------------------}
{ This routine searches the current directory to find a file named             }
{ junk****.txt that does not already exist. If there are no unused files       }
{ available the routine returns '' otherwise it returns the file name.         }
{------------------------------------------------------------------------------}
var
  I:word; {Loop control variable.}
  WorkStr:string; {The junk file name.}
begin
  I:=0;
  repeat
    inc(I);
    WorkStr:='junk'+inttostrf(I,4,true)+'.txt';
  until (not FileExists(WorkStr) or (I=9999));
  if ((I=9999) and FileExists(WorkStr)) then FindJunkFile:=''
  else FindJunkFile:=WorkStr;
end; {End of FindJunkFile.}

function FixFileEOL(const FileSource,FileDest:string;App:TApplication):byte;
{-----------------------------------------------------------------------------}
{ This routine fixes any incorrect end of lines characters in the file whose  }
{ name is passed via filename.                                                }
{-----------------------------------------------------------------------------}
var
  List: TStringList;
begin
  Result := 255;
  List := TStringList.Create;
  try
    List.LoadFromFile(FileSource);
    List.SaveToFile(FileDest);
  finally
    List.Free;
  end;
  Result := 0;
end; {End of FixFileEOL.}

function GenerateDirTree(Dir:char):string;
{------------------------------------------------------------------------------}
{ This function accepts a directory letter and generates a file listing all    }
{ subdirectories in this drive. The name of the file listing the subdirector-  }
{ ies is returned by the function.                                             }
{------------------------------------------------------------------------------}
type
  TDirRec=record              {The data type written to the storage file.}
    Name:shortstring;
    Searched:boolean;
  end;
var
  Done:boolean; {Set to true when all searches have been done.}
  Direc:TDirRec; {The data written to the storage file.}
  Dest:textfile; {File handle for the output file.}
  CurrentDir:string; {The current directory.}
  DirInfo:TSearchRec; {Standard BP7 record for file searches.}
  Source:file of TDirRec; {File handle for the storage file.}
begin
  assignfile(Source,FindJunkFile);
  rewrite(Source);
  Direc.Name:=ShortString(Dir+':');
  Direc.Searched:=false;
  write(Source,Direc);
  repeat
    reset(Source);
    Done:=true;
    repeat read(Source,Direc) until (eof(Source) or (Direc.Searched=false));
    if Direc.Searched=false then
    begin
      seek(Source,filepos(Source)-1);
      Direc.Searched:=true;
      write(Source,Direc);
      Done:=false;
      seek(Source,filesize(Source));
      CurrentDir:=string(Direc.Name);
      if findfirst(CurrentDir+'\*.*',faDirectory,DirInfo)=0 then
      repeat
        with DirInfo do
        if ((Name<>'.') and (Name<>'..') and (Attr in [16,17,48])) then
        begin
          if length(CurrentDir+'\'+Name)>255 then raise Exception.Create('a directory name in GenerateDirTree exceeded 255 characters.');
          Direc.Name:=ShortString(CurrentDir+'\'+Name);
          Direc.Searched:=false;
          write(Source,Direc);
        end;
      until findnext(DirInfo)<>0;
      SysUtils.findclose(DirInfo);
    end;
  until Done;
  Result:=FindJunkFile;
  assignfile(Dest,Result);
  rewrite(Dest);
  reset(Source);
  repeat
    read(Source,Direc);
    writeln(Dest,Direc.Name);
  until eof(Source);
  closefile(Source);
  erase(Source);
  closefile(Dest);
end; {End of GenerateDirTree.}

function GenerateDirTree(Dir:string):TStringList;
{------------------------------------------------------------------------------}
{ This function accepts a directory string and generates a file listing all    }
{ subdirectories off this directory. The function returns a TStringList and it }
{ is the responsibility of the calling program to ensure that the destructor   }
{ of the returned string list is called.                                       }
{------------------------------------------------------------------------------}
type
  TDirRec=record              {The data type written to the storage file.}
    Name:shortstring;
    Searched:boolean;
  end;
var
  Done:boolean; {Set to true when all searches have been done.}
  Direc:TDirRec; {The data written to the storage file.}
  CurrentDir:string; {The current directory.}
  DirInfo:TSearchRec; {Standard BP7 record for file searches.}
  Source:file of TDirRec; {File handle for the storage file.}
begin
  if length(Dir)>255 then raise Exception.Create('the directory name passed to GenerateDirTree exceeded 255 characters.');

  {Prepare the temporary binary storage file.}
  assignfile(Source,FindJunkFile);
  rewrite(Source);
  Direc.Name:=ShortString(EraseEndChar(Dir,'\'));
  Direc.Searched:=false;
  write(Source,Direc);
  repeat
    filemode:=fmopenreadwrite;
    reset(Source);
    Done:=true;
    repeat read(Source,Direc) until (eof(Source) or (Direc.Searched=false));
    if Direc.Searched=false then
    begin
      seek(Source,filepos(Source)-1);
      Direc.Searched:=true;
      write(Source,Direc);
      Done:=false;
      seek(Source,filesize(Source));
      CurrentDir:=String(Direc.Name);
      if findfirst(CurrentDir+'\*.*',faDirectory,DirInfo)=0 then
      repeat
        with DirInfo do
        if ((Name<>'.') and (Name<>'..') and (Attr in [16,17,48])) then
        begin
          if length(CurrentDir+'\'+Name)>255 then raise Exception.Create('the directory name in GenerateDirTree exceeded 255 characters.');
          Direc.Name:=ShortString(CurrentDir+'\'+Name);
          Direc.Searched:=false;
          write(Source,Direc);
        end;
      until findnext(DirInfo)<>0;
      SysUtils.findclose(DirInfo);
    end;
  until Done;

  Result:=TStringList.Create;
  reset(Source);
  repeat
    read(Source,Direc);
    Result.Add(string(Direc.Name));
  until eof(Source);
  closefile(Source);
  erase(Source);
end; {End of GenerateDirTree.}

function GetFileDate(Filename:string):TDateTime;
begin
  if FileExists(Filename) then FileAge(Filename,Result) else Result:=0;
end;

function GetFileList(MaskStr:string;IncludeSubDirs:boolean):TStringList;
{------------------------------------------------------------------------------}
{ This function returns a string list of all files matching the mask specified }
{ by 'MaskStr'.                                                                }
{------------------------------------------------------------------------------}
var
  DirInfo:TSearchRec;
  DirList:TStringList;
  WorkStr:string;
  I:integer;
begin
  if IncludeSubDirs then
  begin
    WorkStr:=ExtractFilename(MaskStr);
    DirList:=GenerateDirTree(ExtractFilePath(MaskStr));
    for I:=0 to DirList.Count-1 do DirList[I]:=DirList[I]+'\'+WorkStr;
  end
  else
  begin
    DirList:=TStringList.Create;
    DirList.Add(MaskStr);
  end;
  try
    Result:=TStringList.Create;
    for I:=0 to DirList.Count-1 do
    begin
      if findfirst(DirList[I],faAnyFile,DirInfo)=0 then
      repeat Result.Add(ExtractFilePath(DirList[I])+DirInfo.Name) until findnext(DirInfo)<>0;
    end;
  finally
    FreeAndNil(DirList);
  end;
  SysUtils.findclose(DirInfo);
end; {End of GetFileList.}

function GetFileSize(Filename:string):longint;
var
  Source:file of byte;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    Result:=filesize(Source);
    closefile(Source);
  end
  else Result:=0;
end;

function GetFirstLine(TextFilename:string):string;
{------------------------------------------------------------------------------}
{ This function returns the first line in the file named 'TextFilename'.       }
{------------------------------------------------------------------------------}
var
  Source:textfile;
begin
  FileMode:=0;
  assignfile(Source,TextFilename);
  reset(Source);
  readln(Source,Result);
  closefile(Source);
  FileMode:=2;
end;

function GetPrompt(var Source:text;Prompt:string):string;
{------------------------------------------------------------------------------}
{ This routie is used for reading strings in from command, default or init-   }
{ ialization files. The prompt for the search string is passed via the         }
{ `prompt' string. The routine expects to see lines within the text file in    }
{ the form: prompt+' = '+datafield. The file handle for the text file is       }
{ passed via the `Source' parameter. The prompt string is case sensitive.      }
{------------------------------------------------------------------------------}
var
  PromptStr,     {The prompt string read in from the command file.}
  LineIn:string; {The line of text read in from the command file.}
  EPos:integer;  {The position of the = sign in the prompt string.}
begin
  Result:='';
  reset(Source);
  repeat
    readln(Source,LineIn);
    PromptStr:=copy(Linein,1,length(Prompt));
  until ((PromptStr=Prompt) or eof(Source));
  if PromptStr=Prompt then
  begin
    EPos:=pos('=',LineIn);
    if EPos<>0 then Result:=Trim(copy(LineIn,EPos+2,length(LineIn)-EPos+1));
  end;
end; {End of GetPrompt routine.}

{$WARNINGS OFF}
function GetShortFilename(Filename:string):string;
var
  DirInfo:TSearchRec;
begin
  if findfirst(Filename,faArchive,DirInfo)=0 then
  begin
    Result:=DirInfo.FindData.cAlternateFileName;
    if Result='' then Result:=DirInfo.FindData.cFileName;
    Result:=ExtractFilePath(Filename)+Result;
  end
  else Result:='';
end;
{$WARNINGS ON}

function GetYNPrompt(var Source:text;Prompt:string;var Value:boolean):byte;
{------------------------------------------------------------------------------}
{ This routine is used for reading and decoding Yes/No settings from command,  }
{ default and initialization files. It makes use of the GetPrompt routine de-  }
{ fined in this unit. If the prompt could not be found, the `Value' parameter  }
{ is left unchanged and the routine returns 1. If the prompt was found and     }
{ could be decoded (ie. was either YES or NO) the `Value' parameter is updated }
{ (set to true for YES and false for NO) and the routine returns 0. If the     }
{ prompt was found but could not be decoded, the `Value' parameter is left un- }
{ changed and the routine returns 2.                                           }
{------------------------------------------------------------------------------}
var
  Hold:string;  {Temporary storage variable.}
begin
  Result:=0;
  Hold:=UpperCase(GetPrompt(Source,Prompt));
  if (Hold='') then Result:=1 else
  begin
    if ((Hold='YES') or (Hold='NO')) then
      if Hold='YES' then Value:=true else Value:=false
    else Result:=2;
  end;
end; {End of GetYNPrompt.}

function NeedsFixFileEOL(Filename:string; Encoding: TEncoding = nil):boolean;
{------------------------------------------------------------------------------}
{ This file returns true if the file specified by 'Filename' has incorrect     }
{ line breaks, that is: #13 not followed by #10, or #10 not preceeded by #13.  }
{ If this routine returns true, you will probably want to call FixFileEOL.     }
{------------------------------------------------------------------------------}
var
  Stream: TFileStream;
  Size: Integer;
  Buffer: TBytes;
  Text: string;
  P: PChar;
begin
  Result := false;
  Encoding := nil;
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    Size := Stream.Size - Stream.Position;
    SetLength(Buffer, Size);
    Stream.Read(Buffer[0], Size);

    Size := TEncoding.GetBufferEncoding(Buffer, Encoding);
    Text := Encoding.GetString(Buffer, Size, Length(Buffer) - Size);

    P := PChar(Text);
    if P <> nil then
      while P^ <> #0 do
        begin
          if P^ = #13 then
          begin
            if (P+1)^ <> #10 then
            begin
              Result := true;
              Break;
            end;
          end;
          if P^ = #10 then
          begin
            if (P-1)^ <> #13 then
            begin
              Result := true;
              Break;
            end;
          end;
          Inc(P);
        end;
  finally
    Stream.Free;
  end;
end; {End of NeedsFixFileEOL.}

function NumAvailFileHandles:byte;
{-----------------------------------------------------------------------------}
{ This routine returns the number of file handles that are still available    }
{ before a `Too many open files' error is given.                              }
{-----------------------------------------------------------------------------}
var
  I,J:word; {Loop control variables.}
  Name:string; {The name of the temporary junk file.}
  Dest:array[1..50] of file; {The array of file handles.}
begin
  I:=0;
  repeat
    inc(I);
    Name:=FindJunkFile;
    assignfile(Dest[I],Name);
    {$I-} rewrite(Dest[I]); {$I+}
  until (not FileExists(Name) or (I=50));
  for J:=1 to I-1 do
  begin
    closefile(Dest[J]);
    erase(Dest[J]);
  end;
  NumAvailFileHandles:=I-1;
end; {End of NumAvailFileHandles.}

function ParseDirectoryList(DirecStr:string):TStringList;
{------------------------------------------------------------------------------}
{ This function parses the directory string passed via `DirecStr' and returns  }
{ a TStringList object of strings that hold the names of the input directories.}
{ Directories in the string are separated by `,' or `;' characters. All errors }
{ are dealt with inside this routine. The routine creates the returned         }
{ TStringList object and it is the responsibility of the calling code to later }
{ dispose of this object.                                                      }
{------------------------------------------------------------------------------}
var
  I:word; {Arbitrary loop control variable.}
begin
  Result:=TStringList.Create;
  if length(DirecStr)>1 then
  begin
    {$WARNINGS OFF}
    for I:=1 to length(DirecStr) do if DirecStr[I] in [';',','] then DirecStr[I]:=';';
    {$WARNINGS ON}
    DirecStr:=EraseFrontAndEndChar(DirecStr,';');
    I:=1;
    while GetWord(DirecStr,';',I)<>'' do
    begin
      DirecStr:=EraseEndChar(DirecStr,'\');
      Result.Add(GetWord(DirecStr,';',I));
      inc(I);
    end;
  end;
end; {End of ParseDirectoryList.}

function SortTextFile(Filename:string;Column:byte):byte;
{------------------------------------------------------------------------------}
{ This routine sorts the text file whose name is passed via `Filename' into    }
{ ascending order according to the values listed in column number `Column'.    }
{ The data in each line can be separated by commas or spaces. The function     }
{ returns error codes according to:                                            }
{ 0: The sorting was successful.                                               }
{------------------------------------------------------------------------------}
var
  Linein1,
  Linein2,
  Junkname:string;
  Sorted:boolean;
  Source,
  Dest:textfile;
  Alt1,
  Alt2:double;
  Code:integer;
begin
  Junkname:=FindJunkFile;
  repeat
    Sorted:=true;
    assignfile(Source,Filename);
    reset(Source);
    assignfile(Dest,Junkname);
    rewrite(Dest);
    repeat
      readln(Source,Linein2);
      if CharInSet(Linein2[1],['''','"']) then writeln(Dest,Linein2);
    until not CharInSet(Linein2[1],['''','"']);
    repeat
      Linein1:=Linein2;
      readln(Source,Linein2);
      if pos(',',Linein1)=0 then val(GetWord(Linein1,' ',Column),Alt1,Code)
        else val(GetWord(Linein1,',',Column),Alt1,Code);
      if Code<>0 then halt;
      if pos(',',Linein2)=0 then val(GetWord(Linein2,' ',Column),Alt2,Code)
        else val(GetWord(Linein2,',',Column),Alt2,Code);
      if Code<>0 then halt;
      if Alt1<=Alt2 then writeln(Dest,Linein1) else
      begin
        Sorted:=false;
        writeln(Dest,Linein2);
        Linein2:=Linein1;
      end;
    until eof(Source);
    writeln(Dest,Linein2);
    closefile(Source);
    closefile(Dest);
    erase(Source);
    rename(Dest,Filename);
  until Sorted;
  SortTextFile:=0;
end; {End of SortTextFile.}

function ReadStringFromBinFile(var Source:file):string;
var
  Hold:PByteArray;
  LenStr:cardinal;
  I:integer;
begin
  blockread(Source,LenStr,sizeof(cardinal));
  getmem(Hold,LenStr);
  blockread(Source,Hold^,LenStr);
  Result:='';
  for I:=1 to LenStr do Result:=Result+chr(Hold^[I-1]);
  freemem(Hold,LenStr);
end;

procedure WaitForFile(Filename:string);
begin
  while not FileExists(Filename) do
  begin
    writeln('Waiting for file '+Filename);
    WaitNMinutes(2);
  end;
end;

procedure WriteStringToBinFile(Str:string;var Dest:file);
var
  Hold:PByteArray;
  LenStr:cardinal;
  I:integer;
begin
  LenStr:=length(Str);
  getmem(Hold,LenStr);
  for I:=1 to LenStr do Hold^[I-1]:=ord(Str[I]);
  blockwrite(Dest,LenStr,sizeof(cardinal));
  blockwrite(Dest,Hold^,LenStr);
  freemem(Hold,LenStr);
end;

{$WARNINGS OFF}
end. {End of gb_files unit.}
{$WARNINGS ON}
