//@***h* BDBP/gb_g1binary.pas
// 
// NAME
//   gb_g1binary.pas
//
// PURPOSE
//  Defines an object which manages generic version 1 binary location data 
//  files. The files make provision for storing location data of up to 32767 
//  readings per day in compressed data files.
//
// DESCRIPTION
//  This unit defines an object which manages generic version 1 binary location data 
//  files. The files make provision for storing location data of up to 32767 
//  readings per day in compressed data files.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_g1binary;


{$I UNITDIREC.INC}

interface

uses
  classes,     {The delphi unit defining the TList Object.}
  sysutils,    {The delphi unit defining system utilities including excpetions.}
  gb_array,    {This unit defines the dynamic array objects include T2DWordArray.}
  gb_datetime; {This unit defines date and time manipulation routines.}

type
  TG1Header=record       {This record defines the header appearing }
    FileType:byte;       {at the beginning of the file. Read the   }
    FullName:string[40]; {document entitled `Generic version 1     }
    AbbrName:string[3];  {binary location data files' for more     }
    DataDesc:string[80]; {information on the fields in this record.}
    Units:string[30];
    Latitude,
    Longitude,
    NullData:double;
    Altitude,
    NumPerDay:word;
  end;

  TG1YearRec=record    {This record defines the records pointed to by the TList}
    Year:word;         {object. Read the document entitled `Generic version 1  }
    Source:string[40]; {binary location data files' for more information on the}
    Offset,            {fields in this record.}
    Scaling:double;
    FilePosition:cardinal;
    Data:T2DWordArray;
  end;
  PG1YearRec=^TG1YearRec;

  TFileDataRec=record
    Year:word;
    Source:string[40];
    Offset,
    Scaling:double;
  end;

  {Here is the location data file object definition.}
  TG1Binary=class(TList)
  private
    FChanged:boolean;
    CurrentIndex,
    CurrentYear:word;
    FTagValue,
    InterpolDist,
    DefaultScaling,
    DefaultOffset:double;
    SourceForIO,
    Filename:string;
    function  CalcEntry(DateTime:TDateTime):longint;
    function  GetDataIndex(YearIndex,Day:word;Entry:longint):word;
    procedure LoadForYearIndex(YearIndex:word);
    procedure PutDataIndex(YearIndex,Day:word;Entry:longint;Value:word);
  protected
    FileInfo:TG1Header;
  public
    constructor Create(FName:string;var FInfo:TG1Header);overload;
    constructor Create(FName:string);overload;
    function AddYear(AYear:word):integer;
    function CalcIndex(Time:double;PositionLabel:string):word;
    function CalcTime(TIndex:word;TimeLabel:string):double;
    procedure ChangeFilename(NewName:string);
    procedure EnsureAllDataLoaded;
    procedure Filter(FilterWidth:word);
    function GetAltitude:word;
    function GetDataDesc:string;
    function GetInterpolValue(Year,Day:word;Time:double):double; overload;
    function GetInterpolValue(DateTime:TDateTime):double; overload;
    function GetInterpolDist:double;
    function GetLatitude:double;
    function GetLocationAbbr:string;
    function GetLocationName:string;
    function GetLongitude:double;
    function GetFilename:string;
    function GetNullDataValue:double;
    function GetNumPerDay:word;
    function GetUnits:string;
    procedure PutValue(Year:integer;Day:word;Entry:longint;Value:double); overload;
    function GetValue(Year:integer;Day:word;Entry:longint):double; overload;
    procedure PutValue(DateTime:TDateTime;Value:double); overload;
    function GetValue(DateTime:TDateTime):double; overload;
    procedure PutWordValue(Year:integer;Day:word;Entry:longint;Value:word);
    function GetWordValue(Year:integer;Day:word;Entry:longint):word;overload;
    function GetYear(Index:integer):word;
    function GetYearIndex(AYear:word):integer;
    function GetSource(Index:integer):string;
    procedure OffsetBy(X:double);
    function GetMeanValue:double;
    function IsNull(Year:integer;ADay:word;Entry:longint):boolean; overload;
    function IsNull(DateTime:TDateTime):boolean; overload;
    function IsNotNull(Year:integer;ADay:word;Entry:longint):boolean; overload;
    function IsNotNull(DateTime:TDateTime):boolean; overload;
    procedure GetEndDate(var EDate:TDateRec);
    function  GetEndDateTime:TDateTime;
    procedure GetStartDate(var SDate:TDateRec);
    function  GetStartDateTime:TDateTime;
    function  NoDataAccessed:boolean;
    function  HasChanged:boolean;
    procedure RemoveYear(AYear:word;ASource:string);
    procedure ReScale(Year:integer;NewValue:double);
    procedure SaveData;
    procedure SetSource(ASource:string);
    procedure SetScalingAndOffset(Max,Min:double);
    procedure SortByYear; virtual;
    destructor Destroy;override;

    {Properties}
    property Data[Year:integer;Day:word;Entry:longint]:double read GetValue write PutValue;default;
    property Data[DateTime:TDateTime]:double read GetValue write PutValue;default;
    property WordData[Year:integer;Day:word;Entry:longint]:word read GetWordValue write PutWordValue;
    property TagValue:double read FTagValue write FTagValue;
  end;
  EG1BinaryError=class(Exception);

{This is an alphabetical list of other routines defined in this unit.}
function CountTotalNonNullData(Data:TG1Binary):longint;overload;
function CountTotalNonNullData(Data:TG1Binary;StartDate,EndDate:TDateTime):longint;overload;
function IsNZLocation(Data:TG1Binary):boolean;
procedure ThreePointMedianFilter(Data:TG1Binary);


implementation

uses
  {$IFNDEF DELPHI5}
  DateUtils,
  {$ENDIF}

  Controls,
  Dialogs,
  gb_math;

constructor TG1Binary.Create(FName:string;var FInfo:TG1Header);
var
  FileHandle:file; {The file handle for the associated data file.}
  DataPtr:PG1YearRec; {The pointer to each year of data.}
  FileData:TFileDataRec; {The data record from the file.}
begin
  FChanged:=false;
  FileMode:=0;
  inherited Create;
  Filename:=FName;
  CurrentYear:=0;
  if FileExists(Filename) then
  try {Start of FileHandle resource protection block.}
    assignfile(FileHandle,Filename);
    reset(FileHandle,1);
    if filesize(FileHandle)<sizeof(TG1Header) then raise EG1BinaryError.Create(
      'the file was smaller than the header and therefore invalid');
    blockread(FileHandle,FInfo,sizeof(TG1Header));
    if ((FInfo.FileType<1) or (FInfo.FileType>50) or (FInfo.Latitude>90.0) or
      (FInfo.Latitude<-90.0) or (FInfo.Longitude>180.0) or (FInfo.Longitude<-180.0) or
      not (FInfo.AbbrName[1] in ['a'..'z','A'..'Z','0'..'9'])) then
      raise EG1BinaryError.Create('the header read in suggests that the file format is not valid.');
    FileInfo:=FInfo;

    {Now read in all of the data.}
    while not eof(FileHandle) do
    begin
      getmem(DataPtr,sizeof(TG1YearRec));
      try {Start of memory resource protection.}
        blockread(FileHandle,FileData,sizeof(FileData));
        with DataPtr^ do
        begin
          Year:=FileData.Year;
          Source:=FileData.Source;
          Offset:=FileData.Offset;
          Scaling:=FileData.Scaling;
          FilePosition:=FilePos(FileHandle);
          Data:=nil;
        end;
        seek(FileHandle,FilePos(FileHandle)+366*FileInfo.NumPerDay*sizeof(word));
        Add(DataPtr);
        if Count=1 then CurrentYear:=DataPtr^.Year;
      except
        on E:Exception do {Note: If the constructor fails, the destructor is called }
        begin           {automatically and memory assigned already is freed there.}
          ShowMessage('The error inside TG1Binary.Create was '+E.Message);
          freemem(DataPtr,sizeof(TG1YearRec));
          raise; {re-raise the exception.}
        end;
      end;
    end;
  finally
    closefile(FileHandle);
  end
  else FileInfo:=FInfo;
  SourceForIO:='';
  DefaultScaling:=0.0;
  DefaultOffset:=0.0;
  CurrentIndex:=0;
  FileMode:=2;
end; {End of TG1Binary.Create.}

constructor TG1Binary.Create(FName:string);
begin
  if not FileExists(FName) then raise EG1BinaryError.Create(
    'this constructor can only be called for files that exist');
  Create(FName,FileInfo);
end;

function TG1Binary.AddYear(AYear:word):integer;
var
  DataPtr:PG1YearRec; {The pointer to each year of data.}
begin
  getmem(DataPtr,sizeof(TG1YearRec));
  DataPtr^.Year:=AYear;
  DataPtr^.Source:=copy(SourceForIO,1,40);
  DataPtr^.Scaling:=DefaultScaling;
  DataPtr^.Offset:=DefaultOffset;
  DataPtr^.FilePosition:=1; {Used to denote that it must be written to file.}
  DataPtr^.Data:=T2DWordArray.Create(366,FileInfo.NumPerDay);
  DataPtr^.Data.Clear;
  Result:=Add(DataPtr);
end;

function TG1Binary.CalcIndex(Time:double;PositionLabel:string):word;
{-----------------------------------------------------------------------------}
{ Give the time and a label specifying whether the index must be the one      }
{ above or below the time, this routine returns the appropriate index. NOTE:  }
{ This function may return a value of 0.                                      }
{-----------------------------------------------------------------------------}
begin
  if ((Time<0.0) or (Time>1.0)) then raise EG1BinaryError.Create(
    'the Time value passed to CalcIndex was <0.0 or >1.0');
  Result:=trunc(Time*FileInfo.NumPerDay+0.500000001);
  if PositionLabel='ABOVE' then inc(Result);
end; {End of TG1Binary.CalcIndex.}

function TG1Binary.CalcTime(TIndex:word;TimeLabel:string):double;
{-----------------------------------------------------------------------------}
{ Given the index and a label specifying where in the data cell the time      }
{ stamp is, this function returns the time of the day in decimal days.        }
{-----------------------------------------------------------------------------}
begin
  if ((TIndex<1) or (TIndex>FileInfo.NumPerDay)) then
    raise EG1BinaryError.Create('TIndex out of range in TG1Binary.CalcTime');
  Result:=0.0;
  if TimeLabel='START' then Result:=(TIndex-1)/FileInfo.NumPerDay;
  if TimeLabel='MIDDLE' then Result:=(TIndex-0.5)/FileInfo.NumPerDay;
  if TimeLabel='END' then Result:=TIndex/FileInfo.NumPerDay;
end; {End of TG1Binary.CalcTime.}

procedure TG1Binary.ChangeFilename(NewName:string);
{------------------------------------------------------------------------------}
{ Changes the name of the file associated with the object to 'NewName'.        }
{------------------------------------------------------------------------------}
begin
  EnsureAllDataLoaded;
  Filename:=NewName;
end; {End of TG1Binary.ChangeFilename.}

procedure TG1Binary.EnsureAllDataLoaded;
{------------------------------------------------------------------------------}
{ This routine ensures that all outstanding unloaded list records are loaded.  }
{------------------------------------------------------------------------------}
var
  YearIndex:integer;
begin
  try
    for YearIndex:=0 to Count-1 do
      if PG1YearRec(Items[YearIndex])^.Data=nil then LoadForYearIndex(YearIndex);
  except
    on E:EG1BinaryError do raise EG1BinaryError.Create('A problem occurred when trying '+
      'to load all of the data. The specific error message was "'+E.Message+
      '". The most likely reason is that the binary file is corrupt. Do what '+
      'you can to fix it, replace it with a backup, or delete it.');
  end;
end; {End of TG1Binary.EnsureAllDataLoaded.}

procedure TG1Binary.Filter(FilterWidth:word);
{------------------------------------------------------------------------------}
{ This function filters the data with a running mean filter of width speci-    }
{ fied by `FilterWidth'. All errors are handled by raising exceptions.         }
{-----------------------------------------------------------------------------}
var
  Data:PMathArray; {The current data string to be filtered.}
  CurrentDate,
  LastDate:TDateTime;
  StartYear,
  EndYear,
  Month,
  Date,
  Year:word;
  AvgWidth,
  Day,
  Num:integer;
  Total:double;
  Store:T3DDoubleArray; {Used to store filtered data.}
  Finished:boolean;
  HFW,       {The half width of the filter.}
  FW,        {Need this because FilterWidth is passed as word and sometimes need negative.}
  Entry,
  I:longint;
begin
  FW:=FilterWidth;
  if (FW mod 2=0) then inc(FW);
  HFW:=FW div 2;
  SortByYear;
  StartYear:=GetYear(0);
  EndYear:=GetYear(Count-1);
  getmem(Data,FW*sizeof(double));

  {Don't use more than 10Mb of memory for the storage array.}
  Store:=T3DDoubleArray.Create(EndYear-StartYear+1,366,FileInfo.NumPerDay,10000000);
  try
    for I:=StartYear to EndYear do
      for Day:=1 to 366 do
        for Entry:=1 to FileInfo.NumPerDay do
          Store[I-StartYear+1,Day,Entry]:=FileInfo.NullData;
    CurrentDate:=GetStartDateTime;
    LastDate:=GetEndDateTime;

    {Fill the data array for first time.}
    for I:=-HFW to HFW do
      Data^[I+HFW+1]:=GetValue(CurrentDate+I/FileInfo.NumPerDay);

    repeat
      {Calculate the year, day and entry corresponding to CurrentDate}
      DecodeDate(CurrentDate,Year,Month,Date);
      Day:=CalcDay(Year,Month,Date);
      Entry:=CalcEntry(CurrentDate);

      {Process the filter array}
      if Data^[HFW+1]<>FileInfo.NullData then
      begin
        AvgWidth:=0;
        Num:=1;
        Total:=Data^[HFW+1];
        Finished:=false;
        repeat
          inc(AvgWidth);
          if ((Data^[HFW+1-AvgWidth]<>FileInfo.NullData) and
              (Data^[HFW+1+AvgWidth]<>FileInfo.NullData)) then
          begin
            Num:=Num+2;
            Total:=Total+Data^[HFW+1-AvgWidth]+Data^[HFW+1+AvgWidth];
          end
          else Finished:=true;
          if AvgWidth=HFW then Finished:=true;
        until Finished;
        Store[Year-StartYear+1,Day,Entry]:=Total/Num;
      end
      else Store[Year-StartYear+1,Day,Entry]:=FileInfo.NullData;

      for I:=1 to FW-1 do Data^[I]:=Data^[I+1];
      CurrentDate:=CurrentDate+1/FileInfo.NumPerDay;
      Data^[FW]:=GetValue(CurrentDate+(HFW/FileInfo.NumPerDay));
    until CurrentDate>=LastDate;

    {Now copy the data in Store back to the object.}
    for I:=0 to Count-1 do
    begin
      Year:=GetYear(I);
      for Day:=1 to DaysInAYear(Year) do
        for Entry:=1 to FileInfo.NumPerDay do
        if not IsNull(Year,Day,Entry) then PutValue(-I,Day,Entry,
          Store[Year-StartYear+1,Day,Entry]);
    end;
  finally
    Store.Free;
    freemem(Data,FW*sizeof(double));
  end;
end; {End of TG1Binary.Filter.}

function TG1Binary.GetAltitude:word;
begin
  GetAltitude:=FileInfo.Altitude;
end;

function TG1Binary.GetDataDesc:string;
begin
  GetDataDesc:=FileInfo.DataDesc;
end;

function TG1Binary.GetInterpolValue(Year,Day:word;Time:double):double;
{------------------------------------------------------------------------------}
{ This routine uses temporal linear interpolation to extract the data value at }
{ the date and time specified by 'Year', 'Day' and 'Time'. The 'Time' para-    }
{ meter is in decimal hours. After calling this method, the GetInterpolDist    }
{ method can be called to get the distance in decimal hours from the nearest   }
{ measurement. If two values spanning the date and time cannot be found,       }
{ InterpolDist is set to -1.0 and the function returns the value of the near-  }
{ est measurement.                                                             }
{------------------------------------------------------------------------------}
var
  YearIndex:integer; {The array index for the particular year.}
  IndexCount,        {Used to search for the upper and lower limit.}
  UpperYear,         {The year for the upper limit.}
  UpperDay,          {The day for the upper limit.}
  LowerYear,         {The year for the lower limit.}
  LowerDay:word;     {The day for the lower limit.}
  UpperData,         {The data value at the upper limit.}
  LowerData,         {The data value at the lower limit.}
  UpperDist,         {The time period, in decimal days to the upper limit.}
  LowerDist,         {The time period, in decimal days to the lower limit.}
  UpperTime,         {The time, in decimal days, of the upper limit.}
  LowerTime:double;  {The time, in decimal days, of the lower limit.}
  EndOfRecord:boolean; {True when the start or end of the data record has been reached.}
begin
  SortByYear;
  Time:=Time/24.0;

  {Look for the upper bound.}
  UpperYear:=Year;
  UpperDay:=Day;
  UpperTime:=0.0;
  IndexCount:=CalcIndex(Time,'ABOVE');
  if IndexCount>FileInfo.NumPerDay then
  begin
    IndexCount:=1;
    inc(UpperDay);
    if UpperDay>DaysInAYear(UpperYear) then
    begin
      UpperDay:=1;
      inc(UpperYear);
    end;
  end;
  if UpperYear>PG1YearRec(Items[Count-1])^.Year then UpperData:=FileInfo.NullData else
  begin
    YearIndex:=GetYearIndex(UpperYear);
    while YearIndex<0 do
    begin
      inc(UpperYear);
      YearIndex:=GetYearIndex(UpperYear);
    end;
    EndOfRecord:=false;
    while ((GetDataIndex(YearIndex,UpperDay,IndexCount)=0) and (not EndOfRecord)) do
    begin
      inc(IndexCount);
      if IndexCount>FileInfo.NumPerDay then
      begin
        IndexCount:=1;
        inc(UpperDay);
        if UpperDay>DaysInAYear(GetYear(YearIndex)) then
        begin
          if YearIndex=(Count-1) then
          begin
            EndOfRecord:=true;
            UpperDay:=DaysInAYear(GetYear(YearIndex))
          end
          else
          begin
            inc(YearIndex);
            UpperDay:=1;
          end;
        end;
      end;
    end;
    UpperData:=GetValue(-YearIndex,UpperDay,IndexCount); {This value could be null which is fine.}
    if not IsNull(-YearIndex,UpperDay,IndexCount) then
    begin
      UpperYear:=GetYear(YearIndex);
      UpperTime:=CalcTime(IndexCount,'MIDDLE');
    end;
  end;

  {Look for the lower bound.}
  LowerYear:=Year;
  LowerDay:=Day;
  LowerTime:=0.0;
  IndexCount:=CalcIndex(Time,'BELOW');
  if IndexCount<1 then
  begin
    IndexCount:=FileInfo.NumPerDay;
    dec(LowerDay);
    if LowerDay<1 then
    begin
      dec(LowerYear);
      LowerDay:=DaysInAYear(LowerYear);
    end;
  end;
  if LowerYear<PG1YearRec(Items[0])^.Year then LowerData:=FileInfo.NullData else
  begin
    YearIndex:=GetYearIndex(LowerYear);
    while YearIndex<0 do
    begin
      dec(LowerYear);
      YearIndex:=GetYearIndex(LowerYear);
    end;
    EndOfRecord:=false;
    while ((GetDataIndex(YearIndex,LowerDay,IndexCount)=0) and (not EndOfRecord)) do
    begin
      dec(IndexCount);
      if IndexCount<1 then
      begin
        IndexCount:=FileInfo.NumPerDay;
        dec(LowerDay);
        if LowerDay<1 then
        begin
          if YearIndex=0 then
          begin
            EndOfRecord:=true;
            LowerDay:=1;
          end
          else
          begin
            dec(YearIndex);
            LowerDay:=DaysInAYear(GetYear(YearIndex));
          end;
        end;
      end;
    end;
    LowerData:=GetValue(-YearIndex,LowerDay,IndexCount); {This value could be null which is fine.}
    if not IsNull(-YearIndex,LowerDay,IndexCount) then
    begin
      LowerYear:=GetYear(YearIndex);
      LowerTime:=CalcTime(IndexCount,'MIDDLE');
    end;
  end;

  InterpolDist:=-1.0;
  Result:=FileInfo.NullData;
  if (EquivDouble(UpperData,FileInfo.NullData) and (LowerData<>FileInfo.NullData)) then Result:=LowerData;
  if ((UpperData<>FileInfo.NullData) and EquivDouble(LowerData,FileInfo.NullData)) then Result:=UpperData;
  if ((UpperData<>FileInfo.NullData) and (LowerData<>FileInfo.NullData)) then
  begin
    UpperDist:=CalcNumDays(Year,Day,UpperYear,UpperDay);
    UpperDist:=UpperDist+UpperTime-Time;
    LowerDist:=CalcNumDays(Year,Day,LowerYear,LowerDay);
    LowerDist:=-1.0*(LowerDist+LowerTime-Time);
    if LowerDist>UpperDist then InterpolDist:=UpperDist*24.0
    else InterpolDist:=LowerDist*24.0;
    Result:=LowerDist*(UpperData-LowerData)/(LowerDist+UpperDist)+LowerData;
  end;
end; {End of TG1Binary.GetInterpolValue.}

function TG1Binary.GetInterpolValue(DateTime:TDateTime):double;
var
  Year,
  Day:word;
  DecHour:double;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  DecHour:=frac(DateTime)*24.0;
  Result:=GetInterpolValue(Year,Day,DecHour);
end;

function TG1Binary.GetInterpolDist:double;
begin
  Result:=InterpolDist;
end;

function TG1Binary.IsNull(Year:integer;ADay:word;Entry:longint):boolean;
var
  YearIndex:integer;
begin
  Result:=false;
  if Year<=0 then YearIndex:=abs(Year) else YearIndex:=GetYearIndex(Year);
  if YearIndex<0 then Result:=true else
  begin
    if PG1YearRec(Items[YearIndex])^.Data=nil then LoadForYearIndex(YearIndex);
    if PG1YearRec(Items[YearIndex])^.Data[ADay,Entry]=0 then Result:=true;
  end;
end; {End of TG1Binary.IsNull.}

function TG1Binary.IsNull(DateTime:TDateTime):boolean;
var
  Year:word;
  Day,
  Entry:longint;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  Entry:=CalcEntry(DateTime);
  Result:=IsNull(Year,Day,Entry);
end; {End of TG1Binary.IsNull.}

function TG1Binary.IsNotNull(Year:integer;ADay:word;Entry:longint):boolean;
begin
  Result:=not IsNull(Year,ADay,Entry);
end; {End of TG1Binary.IsNotNull.}

function TG1Binary.IsNotNull(DateTime:TDateTime):boolean;
begin
  Result:=not IsNull(DateTime);
end; {End of TG1Binary.IsNotNull.}

procedure TG1Binary.GetEndDate(var EDate:TDateRec);
var
  YearIndex,
  Day,I:integer;
  Found:boolean;
begin
  SortByYear;
  Found:=false;
  if SourceForIO='' then YearIndex:=Count-1 else
  begin
    YearIndex:=Count;
    repeat dec(YearIndex)
    until ((PG1YearRec(Items[YearIndex])^.Source=copy(SourceForIO,1,40)) or (YearIndex=0));
    if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
    begin
      EDate:=NullDate;
      Exit;
    end;
  end;
  Day:=DaysInAYear(GetYear(YearIndex));
  I:=GetNumPerDay;
  repeat
    if GetDataIndex(YearIndex,Day,I)<>0 then Found:=true else
    begin
      dec(I);
      if I<1 then
      begin
        I:=GetNumPerDay;
        dec(Day);
        if Day<1 then
        begin
          if SourceForIO='' then
          begin
            dec(YearIndex);
            Day:=DaysInAYear(GetYear(YearIndex));
          end
          else
          begin
            repeat dec(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
              copy(SourceForIO,1,40)) or (YearIndex=0));
            if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
            begin
              Day:=1;
              I:=1;
            end
            else Day:=DaysInAYear(GetYear(YearIndex));
          end;
        end;
      end;
    end;
  until (Found or ((YearIndex=0) and (Day=1) and (I=1)));
  if Found then
  begin
    EDate.Year:=GetYear(YearIndex);
    EDate.Month:=CalcMonth(EDate.Year,Day);
    EDate.Date:=CalcDate(EDate.Year,Day);
  end
  else EDate:=NullDate;
end; {End of TG1Binary.GetEndDate.}

function TG1Binary.GetEndDateTime:TDateTime;
var
  Year:word;
  DecDay:double;
  YearIndex,
  Day,I:integer;
  Found:boolean;
begin
  SortByYear;
  Found:=false;
  if SourceForIO='' then YearIndex:=Count-1 else
  begin
    YearIndex:=Count;
    repeat dec(YearIndex)
    until ((PG1YearRec(Items[YearIndex])^.Source=copy(SourceForIO,1,40)) or (YearIndex=0));
    if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
    begin
      Result:=-999.99;
      Exit;
    end;
  end;
  Day:=DaysInAYear(GetYear(YearIndex));
  I:=GetNumPerDay;
  repeat
    if GetDataIndex(YearIndex,Day,I)<>0 then Found:=true else
    begin
      dec(I);
      if I<1 then
      begin
        I:=GetNumPerDay;
        dec(Day);
        if Day<1 then
        begin
          if SourceForIO='' then
          begin
            dec(YearIndex);
            Day:=DaysInAYear(GetYear(YearIndex));
          end
          else
          begin
            repeat dec(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
              copy(SourceForIO,1,40)) or (YearIndex=0));
            if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
            begin
              Day:=1;
              I:=1;
            end
            else Day:=DaysInAYear(GetYear(YearIndex));
          end;
        end;
      end;
    end;
  until (Found or ((YearIndex=0) and (Day=1) and (I=1)));
  if Found then
  begin
    Year:=GetYear(YearIndex);
    DecDay:=(I-0.5)/GetNumPerDay;
    Result:=EncodeDateTime(Year,CalcMonth(Year,Day),CalcDate(Year,Day),
      trunc(24*DecDay),trunc(frac(24*DecDay)*60),0,0);
  end
  else Result:=-999.99;
end; {End of TG1Binary.GetEndDateTime.}

procedure TG1Binary.GetStartDate(var SDate:TDateRec);
var
  YearIndex,
  Day,I:integer;
  Found:boolean;
begin
  SortByYear;
  Found:=false;
  if SourceForIO='' then YearIndex:=0 else
  begin
    YearIndex:=-1;
    repeat inc(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
      copy(SourceForIO,1,40)) or (YearIndex=Count-1));
    if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
    begin
      SDate:=NullDate;
      Exit;
    end;
  end;
  Day:=1;
  I:=1;
  repeat
    if GetDataIndex(YearIndex,Day,I)<>0 then Found:=true else
    begin
      inc(I);
      if I>GetNumPerDay then
      begin
        I:=1;
        inc(Day);
        if Day>DaysInAYear(GetYear(YearIndex)) then
        begin
          if SourceForIO='' then
          begin
            inc(YearIndex);
            Day:=1;
          end
          else
          begin
            repeat inc(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
              copy(SourceForIO,1,40)) or (YearIndex=Count-1));
            if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
            begin
              Day:=DaysInAYear(GetYear(YearIndex));
              I:=GetNumPerDay;
            end
            else Day:=1;
          end;
        end;
      end;
    end;
  until (Found or ((YearIndex=Count-1) and (Day=DaysInAYear(GetYear(YearIndex))) and (I=GetNumPerDay)));
  if Found then
  begin
    SDate.Year:=GetYear(YearIndex);
    SDate.Month:=CalcMonth(SDate.Year,Day);
    SDate.Date:=CalcDate(SDate.Year,Day);
  end
  else SDate:=NullDate;
end; {End of TG1Binary.GetStartDate.}

function TG1Binary.GetStartDateTime:TDateTime;
var
  Year:word;
  DecDay:double;
  YearIndex,
  Day,I:integer;
  Found:boolean;
begin
  SortByYear;
  Found:=false;
  if SourceForIO='' then YearIndex:=0 else
  begin
    YearIndex:=-1;
    repeat inc(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
      copy(SourceForIO,1,40)) or (YearIndex=Count-1));
    if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
    begin
      Result:=-999.99;
      Exit;
    end;
  end;
  Day:=1;
  I:=1;
  repeat
    if GetDataIndex(YearIndex,Day,I)<>0 then Found:=true else
    begin
      inc(I);
      if I>GetNumPerDay then
      begin
        I:=1;
        inc(Day);
        if Day>DaysInAYear(GetYear(YearIndex)) then
        begin
          if SourceForIO='' then
          begin
            inc(YearIndex);
            Day:=1;
          end
          else
          begin
            repeat inc(YearIndex) until ((PG1YearRec(Items[YearIndex])^.Source=
              copy(SourceForIO,1,40)) or (YearIndex=Count-1));
            if PG1YearRec(Items[YearIndex])^.Source<>copy(SourceForIO,1,40) then
            begin
              Day:=DaysInAYear(GetYear(YearIndex));
              I:=GetNumPerDay;
            end
            else Day:=1;
          end;
        end;
      end;
    end;
  until (Found or ((YearIndex=Count-1) and (Day=DaysInAYear(GetYear(YearIndex))) and (I=GetNumPerDay)));
  if Found then
  begin
    Year:=GetYear(YearIndex);
    DecDay:=(I-0.5)/GetNumPerDay;
    Result:=EncodeDateTime(Year,CalcMonth(Year,Day),CalcDate(Year,Day),
      trunc(24*DecDay),trunc(frac(24*DecDay)*60),0,0);
  end
  else Result:=-999.99;
end; {End of TG1Binary.GetStartDateTime.}

function TG1Binary.NoDataAccessed:boolean;
{------------------------------------------------------------------------------}
{ This function returns true if none of the 2DWordArrays have yet been loaded  }
{ into the object. Otherwise it returns false.                                 }
{------------------------------------------------------------------------------}
var
  I:word;
begin
  Result:=true;
  for I:=0 to Count-1 do
    if PG1YearRec(Items[I])^.Data<>nil then Result:=false;
end; {End of TG1Binary.NoDataAccessed.}

function TG1Binary.GetLatitude:double;
begin
  GetLatitude:=FileInfo.Latitude;
end;

function TG1Binary.GetLocationAbbr:string;
begin
  GetLocationAbbr:=FileInfo.AbbrName;
end; {End of TG1Binary.GetLocationAbbr.}

function TG1Binary.GetLocationName:string;
begin
  GetLocationName:=FileInfo.Fullname;
end; {End of TG1Binary.GetLocationName.}

function TG1Binary.GetLongitude:double;
begin
  GetLongitude:=FileInfo.Longitude;
end;

function TG1Binary.GetFilename:string;
begin
  Result:=Filename;
end;

function TG1Binary.GetNullDataValue:double;
begin
  GetNullDataValue:=FileInfo.NullData;
end; {End of TG1Binary.GetNullDataValue.}

function TG1Binary.GetUnits:string;
begin
  GetUnits:=FileInfo.Units;
end;

procedure TG1Binary.LoadForYearIndex(YearIndex:word);
var
  FileHandle:file;
begin
  with PG1YearRec(Items[YearIndex])^ do
  begin
    Data:=T2DWordArray.Create(366,FileInfo.NumPerDay);
    FileMode:=0;
    assignfile(FileHandle,Filename);
    reset(FileHandle,1);
    seek(FileHandle,FilePosition);
    if Data.LoadData(FileHandle)<>0 then
      raise EG1BinaryError.Create('the data for year '+inttostr(GetYear(YearIndex))+
        ' couldn''t be loaded from the file named '+Filename);
    closefile(FileHandle);
    FileMode:=2;
  end;
end; {End of TG1Binary.LoadForYearIndex.}

procedure TG1Binary.RemoveYear(AYear:word;ASource:string);
var
  I:integer; {Loop control variable.}
begin
  I:=-1;
  while I<(Count-1) do
  begin
    inc(I);
    if PG1YearRec(Items[I])^.Year=AYear then
      if ASource='' then Delete(I) else
        if ASource=PG1YearRec(Items[I])^.Source then Delete(I);
  end;
end; {End of TG1Binary.RemoveYear.}

procedure TG1Binary.PutValue(Year:integer;Day:word;Entry:longint;Value:double);
var
  YearIndex:integer;
  TestValue:double;
begin
  if Day<1 then raise EG1BinaryError.Create('the value for day was < 1.');
  if Day>366 then raise EG1BinaryError.Create('the value for day was greater than 366.');
  if Entry<1 then raise EG1BinaryError.Create('the value for entry was < 1.');
  if Entry>FileInfo.NumPerDay then raise EG1BinaryError.Create(
    'the value for entry was greater than '+inttostr(FileInfo.NumPerDay));
  if Year<=0 then
  begin
    YearIndex:=abs(Year);
    if YearIndex>(Count-1) then raise EG1BinaryError.Create(
      'the YearIndex was greater than the number of year records.');
  end
  else
  begin
    YearIndex:=GetYearIndex(Year);
    if YearIndex<0 then YearIndex:=AddYear(Year);
  end;
  with PG1YearRec(Items[YearIndex])^ do
  if Value=FileInfo.NullData then PutDataIndex(YearIndex,Day,Entry,0) else
  begin
    if Scaling=0.0 then TestValue:=0.0 else TestValue:=round((Value-Offset)/Scaling);
    if ((TestValue<1.0) or (TestValue>65535.0)) then Rescale(Year,Value);
    PutDataIndex(YearIndex,Day,Entry,round((Value-Offset)/Scaling));
  end;
end; {End of TG1Binary.PutValue.}

procedure TG1Binary.PutValue(DateTime:TDateTime;Value:double);
var
  Year:word;
  Day,
  Entry:longint;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  Entry:=CalcEntry(DateTime);
  PutValue(Year,Day,Entry,Value);
end; {End of TG1Binary.PutValue.}

function TG1Binary.GetYearIndex(AYear:word):integer;
var
  I:integer;
begin
  Result:=-1;
  if SourceForIO='' then
    if AYear=CurrentYear then Result:=CurrentIndex else
    begin
      for I:=0 to Count-1 do if PG1YearRec(Items[I])^.Year=AYear then
      begin
        Result:=I;
        CurrentYear:=AYear;
        CurrentIndex:=I;
      end;
    end
  else for I:=0 to Count-1 do
    with PG1YearRec(Items[I])^ do
      if ((Year=AYear) and (Source=copy(SourceForIO,1,40))) then Result:=I;
end; {End of TG1Binary.GetYearIndex.}

function TG1Binary.GetValue(Year:integer;Day:word;Entry:longint):double;
var
  YearIndex:integer;
begin
  if Year<=0 then YearIndex:=abs(Year) else YearIndex:=GetYearIndex(Year);
  if ((YearIndex<0) or (YearIndex>Count-1)) then Result:=FileInfo.NullData else
  begin
    with PG1YearRec(Items[YearIndex])^ do
    begin
      if Data=nil then LoadForYearIndex(YearIndex);
      if Data[Day,Entry]=0 then Result:=FileInfo.NullData
      else Result:=Data[Day,Entry]*Scaling+Offset;
    end;
  end;
end; {End of TG1Binary.GetValue.}

function TG1Binary.GetValue(DateTime:TDateTime):double;
var
  Entry,
  Year,
  Day:word;
begin
  Year:=YearOf(DateTime);
  Day:=DayOfTheYear(DateTime);
  Entry:=CalcEntry(DateTime);
  Result:=GetValue(Year,Day,Entry);
end; {End of TG1Binary.GetValue.}

function TG1Binary.GetWordValue(Year:integer;Day:word;Entry:longint):word;
var
  YearIndex:integer;
begin
  if Year<=0 then YearIndex:=abs(Year) else YearIndex:=GetYearIndex(Year);
  if ((YearIndex<0) or (YearIndex>Count-1)) then Result:=0 else
  begin
    with PG1YearRec(Items[YearIndex])^ do
    begin
      if Data=nil then LoadForYearIndex(YearIndex);
      Result:=Data[Day,Entry];
    end;
  end;
end; {End of TG1Binary.GetWordValue.}

procedure TG1Binary.PutWordValue(Year:integer;Day:word;Entry:longint;Value:word);
var
  YearIndex:integer;
begin
  if Day<1 then raise EG1BinaryError.Create('the value for day was < 1.');
  if Day>366 then raise EG1BinaryError.Create('the value for day was greater than 366.');
  if Entry<1 then raise EG1BinaryError.Create('the value for entry was < 1.');
  if Entry>FileInfo.NumPerDay then raise EG1BinaryError.Create(
    'the value for entry was greater than '+inttostr(FileInfo.NumPerDay));
  if Year<=0 then
  begin
    YearIndex:=abs(Year);
    if YearIndex>(Count-1) then raise EG1BinaryError.Create(
      'the YearIndex was greater than the number of year records.');
  end
  else
  begin
    YearIndex:=GetYearIndex(Year);
    if YearIndex<0 then YearIndex:=AddYear(Year);
  end;
  PutDataIndex(YearIndex,Day,Entry,Value);
end; {End of TG1Binary.PutWordValue.}

function TG1Binary.CalcEntry(DateTime:TDateTime):longint;
begin
  Result:=trunc(frac(DateTime)*FileInfo.NumPerDay)+1;
end;

function TG1Binary.GetDataIndex(YearIndex,Day:word;Entry:longint):word;
begin
  if YearIndex>Count-1 then raise EG1BinaryError.Create(
    'The Year passed to GetDataIndex was out of bounds');
  with PG1YearRec(Items[YearIndex])^ do
  begin
    if Data=nil then LoadForYearIndex(YearIndex);
    Result:=Data[Day,Entry];
  end;
end; {End of TG1Binary.GetDataIndex.}

procedure TG1Binary.PutDataIndex(YearIndex,Day:word;Entry:longint;Value:word);
begin
  if YearIndex>Count-1 then raise EG1BinaryError.Create(
    'The Year passed to PutDataIndex was out of bounds');
  with PG1YearRec(Items[YearIndex])^ do
  begin
    if Data=nil then LoadForYearIndex(YearIndex);
    Data[Day,Entry]:=Value;
    FChanged:=true;
  end;
end;

function TG1Binary.GetNumPerDay:word;
begin
  GetNumPerDay:=FileInfo.NumPerDay;
end; {End of TG1Binary.GetNumPerDay.}

function TG1Binary.GetYear(Index:integer):word;
begin
  if Index>Count-1 then Result:=65535 else Result:=PG1YearRec(Items[Index])^.Year;
end; {End of TG1Binary.GetYear.}

function TG1Binary.GetSource(Index:integer):string;
begin
  if Index>Count-1 then Result:='' else Result:=PG1YearRec(Items[Index])^.Source;
end; {End of TG1Binary.GetSource.}

procedure TG1Binary.OffsetBy(X:double);
var
  Day,
  Entry,
  I:longint;
  Value:double;
begin
  for I:=0 to Count-1 do
    for Day:=1 to 366 do
      for Entry:=1 to FileInfo.NumPerDay do
      if not IsNull(-I,Day,Entry) then
      begin
        Value:=GetValue(-I,Day,Entry)-X;
        PutValue(-I,Day,Entry,Value);
      end;
end;

function TG1Binary.GetMeanValue:double;
var
  Day,
  Entry,
  Num,
  I:longint;
begin
  Result:=0.0;
  Num:=0;
  for I:=0 to Count-1 do
    for Day:=1 to 366 do
      for Entry:=1 to FileInfo.NumPerDay do
        if not IsNull(-I,Day,Entry) then
        begin
          inc(Num);
          Result:=Result+GetValue(-I,Day,Entry);
        end;
  if Num=0 then raise EG1BinaryError.Create('there were no non-null data in the object.')
  else Result:=Result/Num;
end; {End of TG1Binary.GetMeanValue.}

procedure TG1Binary.ReScale(Year:integer;NewValue:double);
var
  NewScaling,
  NewOffset,
  Min,
  Max:double;
  YearIndex,
  I,J:integer;
begin
  if Year<=0 then YearIndex:=abs(Year) else YearIndex:=GetYearIndex(Year);
  if YearIndex<0 then exit;
  Min:=NewValue;
  Max:=NewValue;
  for I:=1 to 366 do
    for J:=1 to FileInfo.NumPerDay do
    if GetDataIndex(YearIndex,I,J)<>0 then
    begin
      if GetValue(-YearIndex,I,J)>Max then Max:=GetValue(-YearIndex,I,J);
      if GetValue(-YearIndex,I,J)<Min then Min:=GetValue(-YearIndex,I,J);
    end;
  if Max=Min then NewScaling:=1E-5 else NewScaling:=(Max-Min)/65534.0;
  NewOffset:=Min-NewScaling;
  for I:=1 to 366 do
    for J:=1 to FileInfo.NumPerDay do
    if PG1YearRec(Items[YearIndex])^.Data[I,J]<>0 then
      PG1YearRec(Items[YearIndex])^.Data[I,J]:=round((GetValue(-YearIndex,I,J)-NewOffset)/NewScaling);
  PG1YearRec(Items[YearIndex])^.Scaling:=NewScaling;
  PG1YearRec(Items[YearIndex])^.Offset:=NewOffset;
end; {End of TG1Binary.ReScale.}

procedure TG1Binary.SaveData;
var
  FileHandle:file;
  I:integer;
  DataRec:TFileDataRec;
  UnLoadedData:boolean;
begin
  UnLoadedData:=false;
  for I:=0 to Count-1 do
    if PG1YearRec(Items[I])^.Data=nil then UnLoadedData:=true;
  assignfile(FileHandle,Filename);
  if (UnLoadedData and FileExists(Filename)) then
  begin
    reset(FileHandle,1);
    for I:=0 to Count-1 do
      if PG1YearRec(Items[I])^.Data=nil then
      with PG1YearRec(Items[I])^ do
      begin
        Data:=T2DWordArray.Create(366,FileInfo.NumPerDay);
        seek(FileHandle,FilePosition);
        if Data.LoadData(FileHandle)<>0 then
        begin
           if MessageDlg('The data for year '+inttostr(GetYear(I))+' couldn''t '+
             'be loaded. Fill the array with null data?',mtWarning,[mbYes,mbNo],0)=mrYes then
             Data.Clear else raise EG1BinaryError.Create(
             'the data for year '+inttostr(GetYear(I))+' couldn''t be loaded');
        end;
      end;
    closefile(FileHandle);
  end;
  rewrite(FileHandle,1);
  blockwrite(FileHandle,FileInfo,sizeof(TG1Header));
  for I:=0 to Count-1 do
  if PG1YearRec(Items[I])^.Data<>nil then
  begin
    with DataRec do
    begin
      Year:=PG1YearRec(Items[I])^.Year;
      Source:=PG1YearRec(Items[I])^.Source;
      Offset:=PG1YearRec(Items[I])^.Offset;
      Scaling:=PG1YearRec(Items[I])^.Scaling;
    end;
    blockwrite(FileHandle,DataRec,sizeof(DataRec));
    if PG1YearRec(Items[I])^.Data.SaveData(FileHandle)<>0 then raise EG1BinaryError.Create(
      'the data for '+inttostr(DataRec.Year)+' couldn''t be saved');
  end;
  closefile(FileHandle);
  FChanged:=false;
end; {End of TG1Binary.SaveData.}

procedure TG1Binary.SetScalingAndOffset(Max,Min:double);
begin
  DefaultScaling:=(Max-Min)/65534.0;
  DefaultOffset:=Min-DefaultScaling;
end; {End of TG1Binary.SetScalingAndOffset}

function Compare(Item1,Item2:pointer):integer;
begin
  Result:=0;
  if PG1YearRec(Item1)^.Year<PG1YearRec(Item2)^.Year then Result:=-1;
  if PG1YearRec(Item1)^.Year>PG1YearRec(Item2)^.Year then Result:=1;
end;

procedure TG1Binary.SortByYear;
begin
  Sort(Compare);
  CurrentYear:=0;
end;

procedure TG1Binary.SetSource(ASource:string);
begin
  SourceForIO:=ASource;
end;

destructor TG1Binary.Destroy;
var
  I:integer; {loop control variable.}
begin
  for I:=0 to Count-1 do
  if Items[I]<>nil then
  begin
    if PG1YearRec(Items[I])^.Data<>nil then PG1YearRec(Items[I])^.Data.Free;
    freemem(Items[I]);
  end;
  inherited Destroy;
end; {End of TG1Binary.Destroy.}

function CountTotalNonNullData(Data:TG1Binary):longint;
{------------------------------------------------------------------------------}
{ This routine counts the total number of non-null data values in the time     }
{ series specified by 'Data'.                                                  }
{------------------------------------------------------------------------------}
var
  YearIndex,
  Day,
  I:integer;
begin
  Result:=0;
  for YearIndex:=0 to Data.Count-1 do
  begin
    if PG1YearRec(Data.Items[YearIndex])^.Data=nil then Data.LoadForYearIndex(YearIndex);
    for Day:=1 to DaysInAYear(Data.GetYear(YearIndex)) do
      for I:=1 to Data.GetNumPerDay do
        if PG1YearRec(Data.Items[YearIndex])^.Data[Day,I]=0 then inc(Result);
  end;
end; {End of CountTotalNonNullData.}

function CountTotalNonNullData(Data:TG1Binary;StartDate,EndDate:TDateTime):longint;
{------------------------------------------------------------------------------}
{ This routine counts the total number of non-null data values in the time     }
{ series specified by 'Data' between the dates and times specified by          }
{ StartDate and EndDate.                                                       }
{------------------------------------------------------------------------------}
var
  YearIndex,
  Day,
  I:integer;
  Year:word;
  TestDate:TDateTime;
begin
  Result:=0;
  for YearIndex:=0 to Data.Count-1 do
  begin
    if PG1YearRec(Data.Items[YearIndex])^.Data=nil then Data.LoadForYearIndex(YearIndex);
    Year:=Data.GetYear(YearIndex);
    for Day:=1 to DaysInAYear(Data.GetYear(YearIndex)) do
      for I:=1 to Data.GetNumPerDay do
      begin
        TestDate:=EncodeDateDay(Year,Day)+(I-0.5)/Data.GetNumPerDay;
        if ((TestDate>=StartDate) and (TestDate<=EndDate) and Data.IsNotNull(Year,Day,I)) then inc(Result);
      end;
  end;
end; {End of CountTotalNonNullData.}

function IsNZLocation(Data:TG1Binary):boolean;
var
  Lat,Long:double;
begin
  Lat:=Data.GetLatitude;
  Long:=Data.GetLongitude;
  if ((Lat>=-48.0) and (Lat<=-34.0) and (Long>166.0) and (Long<=180.0)) then
   Result:=true else Result:=false;
end;

procedure ThreePointMedianFilter(Data:TG1Binary);
{------------------------------------------------------------------------------}
{ This routine runs a three point median filter over the entire time series    }
{ specified by 'Data'.                                                         }
{------------------------------------------------------------------------------}
var
  ScanYear,
  ScanDay,
  ScanInDay,
  YearIndex,
  ScanYearIndex,
  Day,
  I,J:integer;
  Store:T3DDoubleArray;
  Hold:array[-1..1] of double;
begin
  Data.SortByYear;
  Store:=T3DDoubleArray.Create(Data.Count,366,Data.GetNumPerDay);
  try
    Store.Clear;
    for YearIndex:=0 to Data.Count-1 do
      for Day:=1 to DaysInAYear(Data.GetYear(YearIndex)) do
        for I:=1 to Data.GetNumPerDay do
          Store[YearIndex+1,Day,I]:=Data.GetValue(-YearIndex,Day,I);
    for YearIndex:=0 to Data.Count-1 do
      for Day:=1 to DaysInAYear(Data.GetYear(YearIndex)) do
        for I:=1 to Data.GetNumPerDay do
        begin
          for J:=-1 to 1 do
          begin
            ScanYear:=Data.GetYear(YearIndex);
            ScanDay:=Day;
            ScanInDay:=I+J;
            if ScanInDay<1 then
            begin
              ScanInDay:=ScanInDay+Data.GetNumPerDay;
              dec(ScanDay);
              if ScanDay<1 then
              begin
                dec(ScanYear);
                ScanDay:=DaysInAYear(ScanYear);
              end;
            end;
            if ScanInDay>Data.GetNumPerDay then
            begin
              ScanInDay:=ScanInDay-Data.GetNumPerDay;
              inc(ScanDay);
              if ScanDay>DaysInAYear(ScanYear) then
              begin
                inc(ScanYear);
                ScanDay:=1;
              end;
            end;
            ScanYearIndex:=Data.GetYearIndex(ScanYear);
            if ScanYearIndex<0 then Hold[J]:=Data.GetNullDataValue else
              Hold[J]:=Store[ScanYearIndex+1,ScanDay,ScanInDay];
          end;
          if Hold[-1]>Hold[0] then SwapDouble(Hold[-1],Hold[0]);
          if Hold[0]>Hold[1] then SwapDouble(Hold[0],Hold[1]);
          if Hold[-1]>Hold[0] then SwapDouble(Hold[-1],Hold[0]);
          Data.PutValue(-YearIndex,Day,I,Hold[0]);
        end;
  finally
    Store.Free;
  end;
end;

function TG1Binary.HasChanged:boolean;
begin
  Result:=FChanged;
end;

{$HINTS OFF}

end.

