//@***h* BDBP/gb_g1means.pas
// 
// NAME
//   gb_g1means.pas
//
// PURPOSE
//  Handles all the calculations of mean, maximum, minimum, total values from
//  data that is stored in g1binary format.
//
// DESCRIPTION
//  This unit handles all the calculations of mean, maximum, minimum, total values from
//  data that is stored in g1binary format.
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_g1means;

{$I UNITDIREC.INC}

interface

uses
  gb_g1binary;

type
  {This record holds the Daily, Monthly and Annual mean data for one year.}
  TMeans=record
    Daily:array[1..366] of double;
    Monthly:array[1..12] of double;
    Annual:double;
  end;

  TG1Mean=class(TG1Binary)
  private
    Max,
    Min,
    Totals,
    Means:array of TMeans;
  public
    MinDaily,
    MinMonthly,
    MinYearly:integer;
    procedure SetRestrictions(DailyMin,MonthlyMin,YearlyMin:integer);
    procedure FillMeans;
    function GetAnnualMean(Year:integer):double;
    function GetMonthlyMean(Year,Month: integer): double;
    function GetDailyMean(Year,Day:integer):double;
    function GetAnnualTotal(Year:integer):double;
    function GetMonthlyTotal(Year,Month: integer): double;
    function GetDailyTotal(Year,Day:integer):double;
    function GetAnnualMaximum(Year:integer):double;
    function GetMonthlyMaximum(Year,Month: integer): double;
    function GetDailyMaximum(Year,Day:integer):double;
    function GetAnnualMinimum(Year:integer):double;
    function GetMonthlyMinimum(Year,Month: integer): double;
    function GetDailyMinimum(Year,Day:integer):double;
    procedure Scale(ScalingFactor:double);
    procedure SortByYear; override;
  end;

implementation

uses
  DateUtils,
  SysUtils,
  gb_math;

procedure TG1Mean.SetRestrictions(DailyMin,MonthlyMin,YearlyMin:integer);
begin
  MinDaily:=DailyMin;
  MinMonthly:=MonthlyMin;
  MinYearly:=YearlyMin;
end; {End of TG1Mean.SetRestrictions.}

procedure TG1Mean.FillMeans;
var
  YearCount, {Loop control variable through years.}
  Year,      {Year associated with YearCount.}
  Date,      {Loop control variable through dates.}
  Day,       {Loop control variable through days.}
  Month,     {Loop control variable through months.}
  Hour,      {Loop control variable through hours.}
  Num,       {The number of values used to calculate the total.}
  I,J:word;  {General purpose loop control variables.}
  TotalAll,
  TotalValid,
  Scaling,
  Value:double; {The value returned from GetValue.}
  MonthClim:array[1..12] of double;
  MonthMask,
  MonthClimNum:array[1..12] of word;

begin
  SetLength(Means,Count);
  SetLength(Totals,Count);
  SetLength(Max,Count);
  SetLength(Min,Count);
  for I:=1 to 12 do MonthClim[I]:=0.0;
  for I:=1 to 12 do MonthClimNum[I]:=0;
  for YearCount:=0 to Count-1 do
  begin
    Year:=GetYear(YearCount);

    {Null the arrays}
    for I:=1 to 366 do
    begin
      Means[YearCount].Daily[I]:=0.0;
      Totals[YearCount].Daily[I]:=0.0;
      Max[YearCount].Daily[I]:=0.0;
      Min[YearCount].Daily[I]:=0.0;
    end;
    for I:=1 to 12 do
    begin
      Means[YearCount].Monthly[I]:=0.0;
      Totals[YearCount].Monthly[I]:=0.0;
      Max[YearCount].Monthly[I]:=0.0;
      Min[YearCount].Monthly[I]:=0.0;
    end;

    {Calculate the daily statistics.}
    for Day:=1 to DaysInAYear(Year) do
    begin
      Max[YearCount].Daily[Day]:=-1E30;
      Min[YearCount].Daily[Day]:=1E30;
      Totals[YearCount].Daily[Day]:=0.0;
      Num:=0;
      for Hour:=1 to FileInfo.NumPerDay do
      if not IsNull(-YearCount,Day,Hour) then
      begin
        inc(Num);
        Value:=GetValue(-YearCount,Day,Hour);
        Totals[YearCount].Daily[Day]:=Totals[YearCount].Daily[Day]+Value;
        if Value>Max[YearCount].Daily[Day] then Max[YearCount].Daily[Day]:=Value;
        if Value<Min[YearCount].Daily[Day] then Min[YearCount].Daily[Day]:=Value;
      end;
      if Num>=MinDaily then Means[YearCount].Daily[Day]:=Totals[YearCount].Daily[Day]/Num else
      begin
        Means[YearCount].Daily[Day]:=FileInfo.NullData;
        Totals[YearCount].Daily[Day]:=FileInfo.NullData;
      end;
      if Max[YearCount].Daily[Day]<-1E29 then Max[YearCount].Daily[Day]:=FileInfo.NullData;
      if Min[YearCount].Daily[Day]>1E29 then Min[YearCount].Daily[Day]:=FileInfo.NullData;
    end;

    {Calculate the monthly statistics.}
    for Month:=1 to 12 do
    begin
      Max[YearCount].Monthly[Month]:=-1E30;
      Min[YearCount].Monthly[Month]:=1E30;
      Totals[YearCount].Monthly[Month]:=0.0;
      Num:=0;
      for Date:=1 to DaysInAMonth(Year,Month) do
      begin
        Day:=DayOfTheYear(EncodeDate(Year,Month,Date));
        if ((not EquivDouble(Totals[YearCount].Daily[Day],FileInfo.NullData)) and
           (not EquivDouble(Totals[YearCount].Daily[Day],FileInfo.NullData))) then
        begin
          inc(Num);
          Value:=Totals[YearCount].Daily[Day];
          Totals[YearCount].Monthly[Month]:=Totals[YearCount].Monthly[Month]+Value;
          Value:=Means[YearCount].Daily[Day];
          Means[YearCount].Monthly[Month]:=Means[YearCount].Monthly[Month]+Value;
          Value:=Max[YearCount].Daily[Day];
          if Value>Max[YearCount].Monthly[Month] then Max[YearCount].Monthly[Month]:=Value;
          Value:=Min[YearCount].Daily[Day];
          if Value<Min[YearCount].Monthly[Month] then Min[YearCount].Monthly[Month]:=Value;
        end;
      end;
      if Num>=MinMonthly then
      begin
        Means[YearCount].Monthly[Month]:=Means[YearCount].Monthly[Month]/Num;
        MonthClim[Month]:=MonthClim[Month]+Means[YearCount].Monthly[Month];
        inc(MonthClimNum[Month]);
      end
      else
      begin
        Means[YearCount].Monthly[Month]:=FileInfo.NullData;
        Totals[YearCount].Monthly[Month]:=FileInfo.NullData;
      end;
      if Max[YearCount].Monthly[Month]<-1E29 then Max[YearCount].Monthly[Month]:=FileInfo.NullData;
      if Min[YearCount].Monthly[Month]>1E29 then Min[YearCount].Monthly[Month]:=FileInfo.NullData;
    end;
  end;

  {Generate the monthly climatology.}
  for I:=1 to 12 do
    if MonthClimNum[I]>0 then MonthClim[I]:=MonthClim[I]/MonthClimNum[I];

  for YearCount:=0 to Count-1 do
  begin
    Means[YearCount].Annual:=0.0;
    Max[YearCount].Annual:=-1E30;
    Min[YearCount].Annual:=1E30;
    Totals[YearCount].Annual:=0.0;
    Num:=0;

    for Month:=1 to 12 do
    if ((not EquivDouble(Totals[YearCount].Monthly[Month],FileInfo.NullData)) and
       (not EquivDouble(Totals[YearCount].Monthly[Month],FileInfo.NullData))) then
    begin
      inc(Num);
      MonthMask[Month]:=1;
      Value:=Totals[YearCount].Monthly[Month];
      Totals[YearCount].Annual:=Totals[YearCount].Annual+Value;
      Value:=Means[YearCount].Monthly[Month];
      Means[YearCount].Annual:=Means[YearCount].Annual+Value;
      Value:=Max[YearCount].Monthly[Month];
      if Value>Max[YearCount].Annual then Max[YearCount].Annual:=Value;
      Value:=Min[YearCount].Monthly[Month];
      if Value<Min[YearCount].Annual then Min[YearCount].Annual:=Value;
    end
    else MonthMask[Month]:=0;

    if Num>=MinYearly then
    begin
      if Num<>12 then {Need to worry about seasonal biasing.}
      begin
        TotalAll:=0.0;
        J:=0;
        for I:=1 to 12 do
          if MonthClim[I]>0.0 then
          begin
            inc(J);
            TotalAll:=TotalAll+MonthClim[I];
          end;
        TotalAll:=TotalAll/J;

        TotalValid:=0.0;
        J:=0;
        for I:=1 to 12 do
          if MonthMask[I]=1 then
          begin
            inc(J);
            TotalValid:=TotalValid+MonthClim[I];
          end;
        TotalValid:=TotalValid/J;
        Scaling:=TotalAll/TotalValid;
      end
      else Scaling:=1.0;

      Means[YearCount].Annual:=Scaling*Means[YearCount].Annual/Num;
    end
    else
    begin
      Means[YearCount].Annual:=FileInfo.NullData;
      Totals[YearCount].Annual:=FileInfo.NullData;
    end;
    if Max[YearCount].Annual<-1E29 then Max[YearCount].Annual:=FileInfo.NullData;
    if Min[YearCount].Annual>1E29 then Min[YearCount].Annual:=FileInfo.NullData;
  end;
end; {End of TG1Mean.FillMeans}

function TG1Mean.GetAnnualMean(Year:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Means[Year].Annual;
end; {End of TG1Mean.GetAnnualMean}

function TG1Mean.GetMonthlyMean(Year,Month:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Means[Year].Monthly[Month];
end; {End of TG1Mean.GetMonthlyMean}

function TG1Mean.GetDailyMean(Year,Day:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Means[Year].Daily[Day];
end; {End of TG1Mean.GetDailyMean}

function TG1Mean.GetAnnualTotal(Year:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Totals[Year].Annual;
end; {End of TG1Mean.GetAnnualTotal}

function TG1Mean.GetMonthlyTotal(Year,Month:integer): double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Totals[Year].Monthly[Month];
end; {End of TG1Mean.GetMonthlyTotal}

function TG1Mean.GetDailyTotal(Year,Day:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Totals[Year].Daily[Day];
end; {End of TG1Mean.GetDailyTotal}

function TG1Mean.GetAnnualMaximum(Year:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Max[Year].Annual;
end; {End of TG1Mean.GetAnnualMax}

function TG1Mean.GetMonthlyMaximum(Year,Month:integer): double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Max[Year].Monthly[Month];
end; {End of GetMonthlyMaximum}

function TG1Mean.GetDailyMaximum(Year,Day:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Max[Year].Daily[Day];
end; {End of GetDailyMaximum}

function TG1Mean.GetAnnualMinimum(Year:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Min[Year].Annual;
end; {End of GetMonthlyMaximum}

function TG1Mean.GetMonthlyMinimum(Year,Month:integer): double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Min[Year].Monthly[Month];
end; {End of GetMonthlyMaximum}

function TG1Mean.GetDailyMinimum(Year,Day:integer):double;
begin
  if Year<0 then Year:=-Year else Year:=GetYearIndex(Year);
  if Year<0 then Result:=FileInfo.NullData else Result:=Min[Year].Daily[Day];
end; {End of GetMonthlyMaximum}

procedure TG1Mean.Scale(ScalingFactor:double);
var
  YearCount,
  Hour,
  Day:integer;
begin
  for YearCount:=0 to Count-1 do
    for Day:=1 to 366 do
      for Hour:=1 to FileInfo.NumPerDay do
        if not IsNull(-YearCount,Day,Hour) then
          PutValue(-YearCount,Day,Hour,GetValue(-YearCount,Day,Hour)*ScalingFactor);
  FillMeans;
end; {End of Scale}

procedure TG1Mean.SortByYear;
begin
  inherited SortByYear;
  FillMeans;
end;

{$WARNINGS OFF}
end. {End of G1Means}
{$WARNINGS ON}
