//@***h* BDBP/gb_mercator.pas
// 
// NAME
//   gb_mercator.pas
//
// PURPOSE
//  Defines a class descendent from the T2DDoubleArray class defined 
//  in the gb_array unit which is used for manipulation of geographic data on 
//  a mercator projection. 
//
// DESCRIPTION
//  This unit defines a class descendent from the T2DDoubleArray class defined 
//  in the gb_array unit which is used for manipulation of geographic data on  
//  a mercator projection. The class's methods allow the creation of a geo     
//  graphic coordinate system and the extraction of data from the grid.        
//  Various mathematical and physical methods are also defined such as calcul- 
//  ating the area of any cell, calculating average values within any geo-     
//  graphic area etc.              
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_mercator;

{$I UNITDIREC.INC}

interface

uses
  sysutils, {Delphi system utilities unit.}
  gb_array, {The unit defining the large array object.}
  gb_math;  {The unit defining mathematical routines.}

type
  Orientation=(latitude,longitude); {Used for defining the primary file coordinate.}
  TMercator=class(T2DDoubleArray)
  private
    LongCellWidth, {The width of each longitude cell in degrees.}
    LatCellWidth,  {The width of each latitude cell in degrees.}
    WinStartLat,   {The start latitude of the data window.}
    WinStartLong,  {The start longitude of the data window.}
    WinEndLat,     {The end latitude of the data window.}
    WinEndLong,    {The end longitude of the data window.}
    StartLat,      {The start latitude of the data set.}
    StartLong,     {The start longitude of the data set.}
    EndLong,       {The end longitude of the data set.}
    EndLat:double; {The end latitude of the data set.}
  protected
    NumLong,         {The number of longitude cells in the data set.}
    NumLat:cardinal; {The number of latitude cells in the data set.}
    Null:double;   {The value denoting null data.}
  public
    constructor Create(NLong,NLat:cardinal);
    procedure   PlotSettings(WStartLat,WEndLat,WStartLong,WEndLong:double);
    procedure   GeogSettings(GStartLat,GEndLat,GStartLong,GEndLong:double);
    function    LongCellDeg(CellNum:cardinal):double;
    function    LatCellDeg(CellNum:cardinal):double;
    function    LongCellNum(Long:double):cardinal;
    function    LatCellNum(Lat:double):cardinal;
    function    GetNullDataValue:double;
    procedure   SetNullDataValue(NullVal:double);
    function    ReadOut(Lat,Long:double;Interpol:boolean):double;
    procedure   PlotData(X,Y,W,H:word);
    function    GetCellArea(LatCell:cardinal):double;
    function    GetZonalMean(Latitude,SLong,ELong:double):double;
    function    GetMeridMean(Longitude,SLat,ELat:double):double;
    function    GetZonalProfile(GStartLat,GEndLat:double):PMathArray;
    function    GetMeridProfile(GStartLong,GEndLong:double):PMathArray;
  end;
  EMercator=class(Exception);

implementation

uses
  gb_files;

constructor TMercator.Create(NLong,NLat:cardinal);
{------------------------------------------------------------------------------}
{ The `NLat' parameter specifies the number of latitude cells while 'NLong'    }
{ specifies the number of longitude cells. The constructor initialises the     }
{ data storage object, and fails if there was insufficient memory to create    }
{ the dynamic array.                                                           }
{------------------------------------------------------------------------------}
begin
  inherited Create(NLong,NLat);
  NumLat:=NLat;
  NumLong:=NLong;
end; {End of TMercator.Create.}

procedure TMercator.PlotSettings(WStartLat,WEndLat,WStartLong,WEndLong:double);
{------------------------------------------------------------------------------}
{ This routine is used for setting the geographical boundaries of the region   }
{ to be shown in the plot window. `WStartLat' and `WEndLat' are the start and  }
{ end latitude and 'WStartLong' and `WEndLong' are the start and end longitude }
{ of the region of the mercator area that is to be plotted.                    }
{------------------------------------------------------------------------------}
begin
  WinStartLat:=WStartlat;
  WinStartLong:=WStartLong;
  WinEndLat:=WEndLat;
  WinEndLong:=WEndLong;
end; {End of TMercator.Plotsettings.}

procedure TMercator.GeogSettings(GStartLat,GEndLat,GStartLong,GEndLong:double);
{------------------------------------------------------------------------------}
{ This routine defines the geographical coverage of the data set. `GStartLat'  }
{ and `GEndLat' are the starting and ending latitudes respectively. Note that  }
{ `GStartLat' must always be larger than `GEndLat'. The convention is positive }
{ latitude for the Northern Hemisphere and negative latitudes for the Southern }
{ Hemisphere. `GStartLong' and `GEndLong' are the starting and ending longit-  }
{ udes respectively. `GStartLong' must always be smaller than `GEndLong'. The  }
{ convention is negative for the Western Hemisphere and postive for the East-  }
{ ern Hemisphere.                                                              }
{------------------------------------------------------------------------------}
begin
  if (GStartLat<GEndLat) then SwapDouble(GStartLat,GEndLat);
  if (GStartLong>GEndLong) then SwapDouble(GStartLong,GEndLong);
  StartLat:=GStartLat;
  EndLat:=GEndLat;
  StartLong:=GStartLong;
  EndLong:=GEndLong;
  LongCellWidth:=(EndLong-StartLong)/NumLong;
  LatCellWidth:=(StartLat-EndLat)/NumLat;
end; {End of TMercator.GeogSettings.}

function TMercator.LongCellDeg(CellNum:cardinal):double;
{------------------------------------------------------------------------------}
{ This function returns the center longitude of the longitude cell specified   }
{ by `CellNum'.                                                                }
{------------------------------------------------------------------------------}
begin
  Result:=(CellNum*LongCellWidth)+(StartLong-(LongCellWidth/2.0));
end; {End of TMercator.LongCellDeg.}

function TMercator.LatCellDeg(CellNum:cardinal):double;
{------------------------------------------------------------------------------}
{ This function returns the center latitude of the latitude cell specified by  }
{ `CellNum'.                                                                   }
{------------------------------------------------------------------------------}
begin
  Result:=(StartLat+(LatCellWidth/2.0))-(CellNum*LatCellWidth);
end; {End of TMercator.LatCellDeg.}

function TMercator.LongCellNum(Long:double):cardinal;
{------------------------------------------------------------------------------}
{ This function returns the longitude cell number corresponding to `Long'.     }
{------------------------------------------------------------------------------}
begin
  if (Long=EndLong) then Result:=NumLong else
  Result:=trunc((Long-StartLong)/LongCellWidth)+1;
end; {End of TMercator.LongCellNum.}

function TMercator.LatCellNum(Lat:double):cardinal;
{------------------------------------------------------------------------------}
{ This function returns the latitude cell number corresponding to `Lat'.       }
{------------------------------------------------------------------------------}
begin
  if (Lat=EndLat) then Result:=NumLat else
  Result:=trunc((StartLat-Lat)/LatCellWidth)+1;
end; {End of TMercator.LatCellNum.}

function TMercator.GetNullDataValue:double;
{------------------------------------------------------------------------------}
{ This function returns the value denoting null data.                          }
{------------------------------------------------------------------------------}
begin
  Result:=Null;
end; {End of TMercator.GetNullDataValue.}

procedure TMercator.SetNullDataValue(NullVal:double);
{------------------------------------------------------------------------------}
{ This procedure sets the null data value to the value specified by 'NullVal'. }
{------------------------------------------------------------------------------}
begin
  Null:=NullVal;
end; {End of TMercator.SetNullDataValue.}

function TMercator.ReadOut(Lat,Long:double;Interpol:boolean):double;
{------------------------------------------------------------------------------}
{ This routine returns the data value corresponding to the geographical posit- }
{ ion given by the passed parameters `Lat' and `Long'. If `Interpol' is false, }
{ the routine finds the data cell corresponding to `Lat', `Long' and returns   }
{ that value. If `Interpol' is true, the routine makes use of an interpolation }
{ technique to determine the exact value between the cells. This interpolation }
{ algorithm is discussed further in the unit documentation. If `Lat' or `Long' }
{ are outside the geographical coverage of the data set, the routine returns   }
{ the null data value.                                                         }
{------------------------------------------------------------------------------}
type
  CellRec=record
    Data:double; {The data value in the cell.}
    X,Y:word;    {The cell coordinate.}
  end;
var
  NumNullData,   {The number of neighbouring cells having null data.}
  I,J,K,         {Loop control variables.}
  Xc,Yc:word;    {The cell coordinates corresponding to Long and Lat.}
  A,B:PMatrix33; {Used for solving the equation for the plane.}
  X,             {The X cell coordinates for the interpolation.}
  Y,             {The Y cell coordinates for the interpolation.}
  M,             {The slopes of the interpolation lines.}
  C,             {The intercepts for the interpolation lines.}
  D:array[1..3] of double;          {The data values from the interpolation.}
  Cell:array[1..2,1..2] of CellRec; {The 4 corner cell attributes.}
begin
  Result:=Null;
  Xc:=LongCellNum(Long);
  Yc:=LatCellNum(Lat);
  if ((Yc<1) or (Yc>NumLat) or (Xc<1) or (Xc>NumLong)) then exit;
  if Interpol then
  begin {beginning of interpolation.}
    {Take care of special cases first...}
    if (Data[Xc,Yc]=Null) then exit;

    {Check for locations on the `edge' of the data grid.}
    if ((Lat>LatCellDeg(1)) or (Lat<LatCellDeg(NumLat)) or
       (Long<LongCellDeg(1)) or (Long>LongCellDeg(NumLong))) then
    begin
      Result:=Data[Xc,Yc];
      exit;
    end;

    {Get the values of the neighbouring cells.}
    if Long<LongCellDeg(Xc) then {Location is westward of cell center.}
    begin
      Cell[1,1].X:=Xc-1;
      Cell[2,1].X:=Xc;
      Cell[1,2].X:=Xc-1;
      Cell[2,2].X:=Xc;
    end
    else {Location is eastward of cell center.}
    begin
      Cell[1,1].X:=Xc;
      Cell[2,1].X:=Xc+1;
      Cell[1,2].X:=Xc;
      Cell[2,2].X:=Xc+1;
    end;
    if Lat<LatCellDeg(Yc) then {Location is south of cell center.}
    begin
      Cell[1,1].Y:=Yc;
      Cell[2,1].Y:=Yc;
      Cell[1,2].Y:=Yc+1;
      Cell[2,2].Y:=Yc+1;
    end
    else {Location is north of cell center.}
    begin
      Cell[1,1].Y:=Yc-1;
      Cell[2,1].Y:=Yc-1;
      Cell[1,2].Y:=Yc;
      Cell[2,2].Y:=Yc;
    end;
    NumNullData:=0;
    for I:=1 to 2 do
      for J:=1 to 2 do
      begin
        Cell[I,J].Data:=Data[((Cell[I,J].X+287) mod 288)+1,Cell[I,J].Y];
        if Cell[I,J].Data=Null then inc(NumNullData);
      end;
    if NumNullData=0 then {All four data points present.}
    begin
      M[1]:=(Cell[2,2].Data-Cell[1,2].Data)/LongCellWidth;
      C[1]:=Cell[2,2].Data-M[1]*LongCellDeg(Cell[2,2].X);
      D[1]:=M[1]*Long+C[1];
      M[2]:=(Cell[2,1].Data-Cell[1,1].Data)/LongCellWidth;
      C[2]:=Cell[2,1].Data-M[2]*LongCellDeg(Cell[2,1].X);
      D[2]:=M[2]*Long+C[2];
      M[3]:=(D[2]-D[1])/LatCellWidth;
      C[3]:=D[1]-M[3]*LatCellDeg(Cell[2,2].Y);
      Result:=M[3]*Lat+C[3];
    end;
    if NumNullData>0 then {Get the available data points.}
    begin
      K:=0;
      for I:=1 to 2 do
        for J:=1 to 2 do
        if Cell[I,J].Data<>Null then
        begin
          inc(K);
          X[K]:=LongCellDeg(Cell[I,J].X);
          Y[K]:=LatCellDeg(Cell[I,J].Y);
          D[K]:=Cell[I,J].Data;
        end;
    end;
    if NumNullData=1 then {Use definition of a plane by 3 data points.}
    begin
      getmem(A,sizeof(TMatrix33));
      getmem(B,sizeof(TMatrix33));
      {First get the equation for the plane. Use C[1], C[2], C[3].}
      {beep; I put this here to remind me to check code performance.}
      for I:=1 to 3 do
      begin
        A^[I,1]:=D[I];
        A^[I,2]:=Y[I];
        A^[I,3]:=1;
      end;
      for I:=1 to 3 do
      begin
        B^[I,1]:=X[I];
        B^[I,2]:=Y[I];
        B^[I,3]:=1;
      end;
      C[1]:=CalcDet33(A)/CalcDet33(B);
      for I:=1 to 3 do
      begin
        A^[I,1]:=X[I];
        A^[I,2]:=D[I];
        A^[I,3]:=1;
      end;
      C[2]:=CalcDet33(A)/CalcDet33(B);
      for I:=1 to 3 do
      begin
        A^[I,1]:=X[I];
        A^[I,2]:=Y[I];
        A^[I,3]:=D[I];
      end;
      C[3]:=CalcDet33(A)/CalcDet33(B);
      Result:=C[1]*Long+C[2]*Lat+C[3];
      freemem(A,sizeof(TMatrix33));
      freemem(B,sizeof(TMatrix33));
    end;
    if NumNullData=2 then {Take a weighted average of the two available points.}
    begin
      {Use C[1] and C[2] to denote the distances from points 1 and 2.}
      C[1]:=sqrt(sqr(Long-X[1])+sqr(Lat-Y[1]));
      C[2]:=sqrt(sqr(Long-X[2])+sqr(Lat-Y[2]));
      ReadOut:=((C[2]/C[1])*D[1]/(1+(C[2]/C[1])))+(D[2]/(1+(C[2]/C[1])));
    end;
    if NumNullData=3 then Result:=D[1]; {Take the only available value.}
  end {end of interpolation.}
  else ReadOut:=Data[Xc,Yc];
end; {End of TMercator.ReadOut.}

procedure TMercator.PlotData;
{------------------------------------------------------------------------------}
{ The delphi code for this routine has not been written yet.                   }
{------------------------------------------------------------------------------}
begin
end; {End of TMercator.PlotData.}


function TMercator.GetCellArea(LatCell:cardinal):double;
{------------------------------------------------------------------------------}
{ This function returns the geograhical area of the cell at latitude `LatCell' }
{ in square meters.                                                            }
{------------------------------------------------------------------------------}
var
  PoleLat,         {The poleward latitude edge of the cell.}
  EquatLat:double; {The equatorward latitude edge of the cell.}
const
  EarthRadius=6.371E6; {Mean radius of the Earth in meters.}
begin
  PoleLat:=(LatCellDeg(LatCell)+(LatCellWidth/2.0))*pi/180.0;
  EquatLat:=(LatCellDeg(LatCell)-(LatCellWidth/2.0))*pi/180.0;
  if LatCellDeg(LatCell)<0.0 then SwapDouble(PoleLat,EquatLat);
  GetCellArea:=abs((LongCellWidth*pi/180)*sqr(EarthRadius)*(sin(PoleLat)-sin(EquatLat)));
end; {End of TMercator.GetCellArea.}

function TMercator.GetZonalMean(Latitude,SLong,ELong:double):double;
{------------------------------------------------------------------------------}
{ This routine returns the zonal mean between the longitudes (in degrees)      }
{ specified by `SLong' and `ELong'. The mean is calculated by summing all      }
{ non-null values along the latitude line specified by `Latitude' between the  }
{ two longitude limits. All errors are handled through exceptions.             }
{------------------------------------------------------------------------------}
var
  LongCount,    {The loop control variable.}
  Num:longint;  {The number of non-null values.}
  Total:double; {The total of all values.}
begin
  if SLong>ELong then SwapDouble(SLong,ELong);
  if SLong<StartLong then raise EMercator.Create('The start longitude passed to '+
    'GetZonalMean was outside the geographical area.');
  if ELong>EndLong then raise EMercator.Create('The end longitude passed to '+
    'GetZonalMean was outside the geographical area.');
  if ((Latitude<EndLat) or (Latitude>StartLat)) then raise EMercator.Create('The '+
    'latitude passed to GetZonalMean was outside the geographical area.');
  Num:=0;
  Total:=0.0;
  for LongCount:=LongCellNum(SLong) to LongCellNum(ELong) do
  if Data[LongCount,LatCellNum(Latitude)]<>Null then
  begin
    inc(Num);
    Total:=Total+Data[LongCount,LatCellNum(Latitude)];
  end;
  if Num=0 then raise EMercator.Create('All of the values at the specified '+
    'latitude were zero in the call to GetZonalMean.') else Result:=Total/Num;
end; {End of TMercator.GetZonalMean.}

function TMercator.GetMeridMean(Longitude,SLat,ELat:double):double;
{------------------------------------------------------------------------------}
{ This routine returns the meridional mean between the latitude (in degrees)   }
{ specified `SLat' and `ELat'. The mean is calculated by summing all non-null  }
{ values along the longitude line specified by `Longitude' between the two     }
{ latitude limits. All errors are managed by raising exceptions.               }
{------------------------------------------------------------------------------}
var
  LatCount,     {The loop control variable.}
  Num:longint;  {The number of non-null values.}
  Total:double; {The total of all values.}
begin
  if SLat<ELat then SwapDouble(SLat,ELat);
  if SLat>StartLat then raise EMercator.Create('The start latitude in the call '+
    'to GerMeridMean was outside the geographical area.');
  if ELat<EndLat then raise EMercator.Create('The end latitude in the call '+
    'to GerMeridMean was outside the geographical area.');
  if ((Longitude>EndLong) or (Longitude<StartLong)) then raise EMercator.Create(
    'The longitude in the call to GerMeridMean was outside the geographical area.');
  Num:=0;
  Total:=0.0;
  for LatCount:=LatCellNum(SLat) to LatCellNum(ELat) do
  if Data[LongCellNum(Longitude),LatCount]<>Null then
  begin
    inc(Num);
    Total:=Total+Data[LongCellNum(Longitude),LatCount];
  end;
  if Num=0 then raise EMercator.Create('All of the values at the specified '+
    'longitude were zero in the call to GetMeridMean.') else GetMeridMean:=Total/Num;
end; {End of TMercator.GetMeridMean.}

function TMercator.GetZonalProfile(GStartLat,GEndLat:double):PMathArray;
{------------------------------------------------------------------------------}
{ This routine returns the mean zonal profile between the latitudes (degrees)  }
{ passed by `GStartLat' (the northern most latitude) and `GEndLat' (the south- }
{ ern most latitude). The profile is returned through a MathArray pointer      }
{ which will contain NumLong elements. It is the responsibility of the calling }
{ routine to free the memory assigned to this pointer. All errors are managed  }
{ by raising exceptions. If all values in the meridian between GStartLat and   }
{ GEndLat are null, a null data entry is entered into the returned array.      }
{------------------------------------------------------------------------------}
var
  LatCount,     {The latitude loop control variable.}
  LongCount,    {The longitude loop control variable.}
  Num:longint;  {The number of non-null values.}
  Total:double; {The total of all values.}
begin
  getmem(Result,NumLong*sizeof(double));
  if GStartLat>StartLat then raise EMercator.Create('The start latitude passed to '+
    'GetZonalProfile was outside the geographical range.');
  if GEndLat<EndLat then raise EMercator.Create('The end latitude passed to '+
    'GetZonalProfile was outside the geographical range.');
  for LongCount:=1 to NumLong do
  begin
    Num:=0;
    Total:=0.0;
    for LatCount:=LatCellNum(GStartLat) to LatCellNum(GEndLat) do
    if Data[LongCount,LatCount]<>Null then
    begin
      inc(Num);
      Total:=Total+Data[LongCount,LatCount];
    end;
    if Num=0 then Result^[LongCount]:=Null else Result^[LongCount]:=Total/Num;
  end;
end; {End of TMercator.GetZonalProfile.}

function TMercator.GetMeridProfile(GStartLong,GEndLong:double):PMathArray;
{------------------------------------------------------------------------------}
{ This routine returns the mean meridional profile between the longitudes (in  }
{ degrees) passed by `GStartLong' (the western most longitude) and `GEndLong'  }
{ (the eastern most longitude). The profile is returned through a MathArray    }
{ pointer which will contain NumLat entries. It is the responsibility of the   }
{ calling routine to free the memory assigned to this array. All errors are    }
{ managed by raising excpetions. If all values in the zone between GStartLong  }
{ and GEndLong are null, a null data entry is entered into the returned array. }
{------------------------------------------------------------------------------}
var
  LatCount,     {The latitude loop control variable.}
  LongCount,    {The longitude loop control variable.}
  Num:longint;  {The number of non-null values.}
  Total:double; {The total of all values.}
begin
  getmem(Result,NumLat*sizeof(double));
  if GStartLong<StartLong then raise EMercator.Create('The start longitude '+
    'passed to GerMeridProfile was outside the geographical range.');
  if GEndLong>EndLong then raise EMercator.Create('The end longitude '+
    'passed to GerMeridProfile was outside the geographical range.');
  for LatCount:=1 to NumLat do
  begin
    Num:=0;
    Total:=0.0;
    for LongCount:=LongCellNum(GStartLong) to LongCellNum(GEndLong) do
    if Data[LongCount,LatCount]<>Null then
    begin
      inc(Num);
      Total:=Total+Data[LongCount,LatCount];
    end;
    if Num=0 then Result^[LatCount]:=Null else Result^[LatCount]:=Total/Num;
  end;
end; {End of TMercator.GetMeridProfile.}

{$WARNINGS OFF}
end. {End of mercator unit.}
{$WARNINGS ON}

