//@***h* BDBP/gb_merctoms.pas
// 
// NAME
//   gb_merctoms.pas
//
// PURPOSE
//  Is specifically designed to manipulate and display TOMS data.  
//
// DESCRIPTION
//  This unit defines a class descendent from the TMercator class defined in 
//  the gb_mercator unit. This class is specifically designed to manipulate 
//  and display TOMS data. The display methods that it makes use of are 
//  defined by the TMercator class.     
//
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/19/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_merctoms;


{$I UNITDIREC.INC}

interface

uses
  sysutils,    {The delphi system utilities unit.}
  gb_array,    {The unit defining the TLargeArray class.}
  gb_mercator, {The unit defining the TMercator class.}
  gb_synopticglobal; {The unit defining TSynopticGlobal.}

const
  NumSatSets=12;

type
  TTomsType=(N7T,M3T,EPT,ADT,GOA,GKN,DLR,N7S,N9S,N11S,N16S,OMI);
  TTomsList=array of TTomsType;
  THemis=(TomsNorth,TomsSouth,TomsGlobal);
  TIntStyle=(median,bestr2,combination);
  CorrFunction=function(Lat:double;Year,Day:word):double;

  TMercTOMS=class(TMercator)
  private
    MaxLongInterpol, {The maximum number of degrees over which longitudinal interpolation must be performed.}
    MaxLatInterpol:double; {The maximum number of degrees over which latitudinal interpolation must be performed.}
    OverThePoleLatCutoff, {The latitude interpolation distance constraint is relaxed poleward of this latitude.}
    DataAvailLatCutoff, {All data must be available poleward of this latitude.}
    MaxPasses:word; {The maximum number of passes in PatchToms.}
    ListIndex,     {The position on the priority list. First=1}
    DataType:byte; {The TOMS data file type.}
    DoNSSearch:boolean; {Set to true if meridional interpolation is permitted in PatchTOMS.}
    DataFilename:string; {The name of the file from which data were last loaded.}
    Coverage:THemis;     {Whether the data is for N or S hemisphere or global.}
  public
    constructor Create(Hemis:THemis);overload;
    constructor Create(TomsData:TMercToms);overload;
    procedure   LoadHemisData(Hemis:THemis;FullName:string);
    procedure   SaveHemisData(Hemis:THemis;FullName:string;Year,Day:integer);
    function    CalcMass(LongCell,LatCell:cardinal):double;
    procedure   SetPatchTOMSFactors(MaxLong,MaxLat:double;DataAvailCutoff,OverThePoleCutoff,MaxPass:word;NSSearch:boolean);
    procedure   PatchTOMS;
    procedure   LongitudeInterpol;
    procedure   GeophysInterpolTOMSPolar(GeoData:TSynopticGlobalArray;IntStyle:TIntStyle);
    function    Missing(Hemis:THemis):double;
    procedure   CorrectData(F:CorrFunction;Year,Day:word);
    function    GetDataType:byte;
    function    GetListIndex:byte;
    function    GetDataFilename:string;
    function    GetCoverage:THemis;
    procedure   PutDataType(Value:byte);
    procedure   PutListIndex(Value:byte);
    procedure   PutDataFilename(Value:string);
  end;
  EMercToms=class(Exception);

procedure SetNimbus7Dir(ADir:string);
procedure SetMeteor3Dir(ADir:string);
procedure SetDLRGomeDir(ADir:string);
procedure SetKNMIGomeDir(ADir:string);
procedure SetGOAGomeDir(ADir:string);
procedure SetEarthProbeDir(ADir:string);
procedure SetAdeosDir(ADir:string);
procedure SetNimbus7SBUVDir(ADir:string);
procedure SetNOAA9SBUVDir(ADir:string);
procedure SetNOAA11SBUVDir(ADir:string);
procedure SetNOAA16SBUVDir(ADir:string);
procedure SetOMIDir(ADir:string);

procedure LoadN7Coeff(Filename:string);
procedure LoadM3Coeff(Filename:string);
procedure LoadEPCoeff(Filename:string);
procedure LoadADCoeff(Filename:string);
procedure LoadGOCoeff(Filename:string);
procedure LoadGKCoeff(Filename:string);
procedure LoadGGCoeff(Filename:string);
procedure LoadN7SCoeff(Filename:string);
procedure LoadN9SCoeff(Filename:string);
procedure LoadN11SCoeff(Filename:string);
procedure LoadN16SCoeff(Filename:string);
procedure LoadOMICoeff(Filename:string);

function LoadTomsData(TomsData:TMercToms;Year,Day,StartIndex:word;CorrectData,
  LongInterpol,ApplyCutoffs,Composite:boolean):boolean; overload;
function LoadTomsData(TomsData:TMercToms;Year,Day,StartIndex:word;TomsList:TTomsList;
  CorrectData,LongInterpol,ApplyCutoffs,Composite:boolean):boolean; overload;

{==============================================================================}
implementation

uses
  DateUtils,
  Math,
  gb_datetime,
  gb_files,
  gb_ncepncar,
  gb_math,
  gb_strings;

type
  TInterpolReg=class(TMultiReg)
    BasisArray:array[1..50] of PMathArray;
    function CalcFunc(X:double):byte;override;
  end;

var
  N7Coeff,OMICoeff:array[1..22] of double;
  EPCoeff:array[1..23] of double;
  M3Coeff,
  ADCoeff,
  GOCoeff,
  GKCoeff,
  GGCoeff,
  N7SCoeff,
  N9SCoeff,
  N11SCoeff,
  N16SCoeff:array[1..6,1..90] of double;
  Nimbus7Dir,
  Meteor3Dir,
  DLRGomeDir,
  KNMIGomeDir,
  GOAGomeDir,
  EarthProbeDir,
  AdeosDir,
  Nimbus7SBUVDir,
  NOAA9SBUVDir,
  NOAA11SBUVDir,
  NOAA16SBUVDir,
  OMIDir:string;

function TInterpolReg.CalcFunc(X:double):byte;
var
  I:integer;
begin
  for I:=1 to NumCoeff-1 do BasisFunc^[I]:=BasisArray[I]^[round(X)];
  BasisFunc^[NumCoeff]:=1.0;
  Result:=0; {Tells the ancestor that the basis functions were filled.}
end;

constructor TMercTOMS.Create(Hemis:THemis);
{------------------------------------------------------------------------------}
{ This constructor creates the necessary data storage structure for the TOMS   }
{ data. The `Hemis' parameter specifies whether the data set will contain      }
{ Southern Hemisphere data, Northern Hemisphere data, or global data.          }
{------------------------------------------------------------------------------}
var
  LongDim:cardinal; {The longitude dimension which depends on Hemis.}
begin
  if Hemis=TomsGlobal then LongDim:=180 else LongDim:=90;
  inherited Create(288,LongDim);
  case Hemis of
    TomsGlobal:inherited GeogSettings(90,-90,-180,180);
    TomsSouth:inherited GeogSettings(0,-90,-180,180);
    TomsNorth:inherited GeogSettings(90,0,-180,180);
  end;
  Coverage:=Hemis;
  Null:=0.0;
  MaxLongInterpol:=30;
  MaxLatInterpol:=10;
  DataAvailLatCutoff:=60;
  DoNSSearch:=true;
  MaxPasses:=2;
  Clear;
end; {End of TMercTOMS.Create.}

constructor TMercToms.Create(TomsData:TMercToms);
{------------------------------------------------------------------------------}
{ This constructor creates the object as a direct copy of the object passed    }
{ via `TomsData'.                                                              }
{------------------------------------------------------------------------------}
var
  I,J:integer;
begin
  Create(TomsData.GetCoverage);
  for I:=1 to TomsData.NumLong do
    for J:=1 to TomsData.NumLat do
      Data[I,J]:=TomsData[I,J];
end; {End of TMercToms.Create.}

procedure TMercTOMS.LoadHemisData(Hemis:THemis;FullName:string);
{------------------------------------------------------------------------------}
{ This routine loads the data from the file specified by `FullName' into the   }
{ the storage array. The `Hemis' parameter specifies which hemisphere the data }
{ must be loaded into. This value must be either north or south. All errors    }
{ managed by raising exceptions.                                               }
{------------------------------------------------------------------------------}
type
  IntArray=array[1..(288*91)] of smallint;
  PIntArray=^IntArray;
var
  Year,Day,            {The year and day number read in from the TOMS file.}
  NLat,NLong:smallint; {The number of latitude and longitude values in the TOMS file.}
  TomsData:PIntArray;  {The pointer to the data read in from the TOMS file.}
  Source:file;         {The file handle for the source data file.}
  I,J,                 {Loop control variables.}
  BlockCount,          {Loop control variable for reading in the TOMS data.}
  LinesPerBlock,       {The number of lines of data per block.}
  NumMemBlocks:word;   {The number of memory blocks needed to read in the data.}
begin
  TomsData:=nil;
  LinesPerBlock:=288;
  if ((Hemis<>TomsSouth) and (Hemis<>TomsNorth)) then raise EMercToms.Create('The hemisphere '+
    'type passed to LoadHemisData wasn''t valid.');
  if not FileExists(FullName) then raise EMercToms.Create('The file named '+Fullname+
    ' couldn''t be found in the call to LoadHemisData.');
  assignfile(Source,FullName);
  FileMode:=0;
  reset(Source,sizeof(smallint));
  FileMode:=2;
  try
    if filesize(Source)<>(52424 div sizeof(smallint)) then raise EMercToms.Create(
      'The size of the file specified in the call to LoadHemisData was the wrong size.');
    blockread(Source,Year,1);
    blockread(Source,Day,1);
    blockread(Source,NLat,1);
    blockread(Source,NLong,1);
    repeat
      try getmem(TomsData,LinesPerBlock*91*sizeof(smallint)) except
        on E:EOutOfMemory do
        begin
          LinesPerBlock:=LinesPerBlock div 2;
          TomsData:=nil;
        end;
      end;
    until TomsData<>nil;
    if LinesPerBlock<9 then raise EMercToms.Create('There is insufficient memory '+
      'available in the call to LoadHemisData.');
    NumMemBlocks:=288 div LinesPerBlock;
    for BlockCount:=1 to NumMemBlocks do
    begin
      blockread(Source,TomsData^,LinesPerBlock*91);
      if ((Coverage=TomsSouth) and (Hemis=TomsSouth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do
          if TomsData^[((I-1)*91)+J+1]<=-200 then
            Data[(BlockCount-1)*LinesPerBlock+I,J]:=Null
          else Data[(BlockCount-1)*LinesPerBlock+I,J]:=TomsData^[((I-1)*91)+J+1];
      if ((Coverage=TomsNorth) and (Hemis=TomsNorth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do
          if TomsData^[((I-1)*91)+J+1]<=-200 then
            Data[(BlockCount-1)*LinesPerBlock+I,91-J]:=Null
          else Data[(BlockCount-1)*LinesPerBlock+I,91-J]:=TomsData^[((I-1)*91)+J+1];
      if ((Coverage=TomsGlobal) and (Hemis=TomsSouth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do
          if TomsData^[((I-1)*91)+J+1]<=-200 then
            Data[(BlockCount-1)*LinesPerBlock+I,J+90]:=Null
          else Data[(BlockCount-1)*LinesPerBlock+I,J+90]:=TomsData^[((I-1)*91)+J+1];
      if ((Coverage=TomsGlobal) and (Hemis=TomsNorth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do
          if TomsData^[((I-1)*91)+J+1]<=-200 then
            Data[(BlockCount-1)*LinesPerBlock+I,91-J]:=Null
          else Data[(BlockCount-1)*LinesPerBlock+I,91-J]:=TomsData^[((I-1)*91)+J+1];
    end;
  finally
    closefile(Source);
    if TomsData<>nil then freemem(TomsData,LinesPerBlock*91*sizeof(smallint));
  end;
end; {End of TMercTOMS.LoadHemisData.}

procedure TMercTOMS.SaveHemisData(Hemis:THemis;FullName:string;Year,Day:integer);
{------------------------------------------------------------------------------}
{ This routine saves the data in the data array to the file specified by       }
{ `FullName'. The data are saved as version 1 TOMS hemisphere data files. The  }
{ `Hemis' parameter specifies which hemisphere is to be saved. This value must }
{ be either north or south. To save both hemisphere of data, this routine must }
{ be called twice. The `Year' and `Day' parameters are also written to the     }
{ file. All errors are managed by raising exceptions.                          }
{-----------------------------------------------------------------------------}
type
  IntArray=array[1..(288*91)] of smallint;
  PIntArray=^IntArray;
var
  NLat,NLong:integer; {The number of latitude and longitude values in the TOMS file.}
  TomsData:PIntArray; {The pointer to the data read in from the TOMS file.}
  Dest:file;          {The file handle for the destination data file.}
  I,J,                {Loop control variables.}
  BlockCount,         {Loop control variables for writing out the TOMS data.}
  LinesPerBlock,      {The number of lines of data per block.}
  NumMemBlocks:word;  {The number of memory blocks needed to read in the data.}
begin
  TomsData:=nil;
  LinesPerBlock:=288;
  if ((Hemis<>TomsSouth) and (Hemis<>TomsNorth)) then raise EMercToms.Create('The hemisphere '+
    'type passed to SaveHemisData wasn''t valid.');
  try
    assignfile(Dest,FullName);
    {$I-}rewrite(Dest,sizeof(smallint));{$I+}
    if ioresult<>0 then raise EMercToms.Create('The output file couldn''t be created '+
      'in the call to SaveHemisData.');
    repeat
      try getmem(TomsData,LinesPerBlock*91*sizeof(smallint)) except
        on E:EOutOfMemory do
        begin
          LinesPerBlock:=LinesPerBlock div 2;
          TomsData:=nil;
        end;
      end;
    until TomsData<>nil;
    if LinesPerBlock<9 then raise EMercToms.Create('There is insufficient memory '+
      'available in the call to SaveHemisData.');
    NumMemBlocks:=288 div LinesPerBlock;
    NLat:=90;
    NLong:=288;
    blockwrite(Dest,Year,1);
    blockwrite(Dest,Day,1);
    blockwrite(Dest,NLat,1);
    blockwrite(Dest,NLong,1);
    for BlockCount:=1 to NumMemBlocks do
    begin
      if ((Coverage=TomsSouth) and (Hemis=TomsSouth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do TomsData^[((I-1)*91)+J+1]:=round(Data[I,J]);
      if ((Coverage=TomsNorth) and (Hemis=TomsNorth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do TomsData^[((I-1)*91)+J+1]:=round(Data[I,91-J]);
      if ((Coverage=TomsGlobal) and (Hemis=TomsSouth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do TomsData^[((I-1)*91)+J+1]:=round(Data[I,J+90]);
      if ((Coverage=TomsGlobal) and (Hemis=TomsNorth)) then
      for I:=1 to LinesPerBlock do
        for J:=1 to 90 do TomsData^[((I-1)*91)+J+1]:=round(Data[I,91-J]);
      {$I-}blockwrite(Dest,TomsData^,LinesPerBlock*91);{$I+}
      if ioresult<>0 then raise EMercToms.Create('The data couldn''t be written '+
        'to the output data file in SaveHemisData.');
    end;
  finally
    closefile(Dest);
    if TomsData<>nil then freemem(TomsData,LinesPerBlock*91*sizeof(smallint));
  end;
end; {End of TMercTOMS.SaveHemisData.}

function TMercToms.CalcMass(LongCell,LatCell:cardinal):double;
{------------------------------------------------------------------------------}
{ This routine calculates the mass in the TOMS cell indexed by `LongCell' and  }
{ `LatCell'.                                                                   }
{------------------------------------------------------------------------------}
begin
  Result:=2.11E-5*Data[LongCell,LatCell]*GetCellArea(LatCell);
end; {End of TMercToms.CalcMass.}

procedure TMercToms.SetPatchTOMSFactors(MaxLong,MaxLat:double;DataAvailCutoff,OverThePoleCutoff,MaxPass:word;NSSearch:boolean);
begin
  MaxLongInterpol:=MaxLong;
  MaxLatInterpol:=MaxLat;
  DataAvailLatCutoff:=DataAvailCutoff;
  OverThePoleLatCutoff:=OverThePoleCutOff;
  MaxPasses:=MaxPass;
  DoNSSearch:=NSSearch;
end; {End of TMercToms.SetPatchTOMSFactors.}

procedure TMercToms.PatchTOMS;
{------------------------------------------------------------------------------}
{ This function patches up gaps in the TOMS data files through longitudinally  }
{ weighted spatial interpolation, including across the polar cap. All errors   }
{ are managed by raising exceptions.                                           }
{------------------------------------------------------------------------------}
type
  SeekRec=record
    Pos:integer;
    Value:double;
  end;
var
  LatInterpolValue,
  LongInterpolValue,
  LongDist,
  LatDist:double;
  Pass,
  NorthSteps,
  SouthSteps,
  EastSteps,
  WestSteps,
  Counter,
  SeekLong,
  I,J,K:integer;
  Seek:array[1..4] of SeekRec;
  StartInd,EndInd:byte;
  StillMissing,
  NSSearchDone:boolean;
begin
  StartInd:=1;
  EndInd:=90;
  if Coverage=TomsGlobal then EndInd:=180;
  Pass:=0;
  repeat
    inc(Pass);
    StillMissing:=false;
    for Counter:=StartInd to EndInd do
    begin
      if ((Coverage=TomsGlobal) and (Counter<91)) then I:=91-Counter else I:=Counter;
      if Coverage=TomsNorth then I:=91-Counter;
      for J:=1 to 288 do
      if Data[J,I]<=1.0 then
      begin
        for K:=1 to 4 do Seek[K].Value:=0;
        for K:=1 to 4 do Seek[K].Pos:=-999;
        LatDist:=-999;
        LongDist:=-999;
        LatInterpolValue:=-999.0;
        LongInterpolValue:=-999.0;

        {First look east and west for values for (2) and (4)}
        {First look east}
        EastSteps:=0;
        Seek[2].Pos:=J;
        repeat
          inc(EastSteps);
          inc(Seek[2].Pos);
          Seek[2].Value:=Data[(Seek[2].Pos-1) mod 288+1,I];
        until ((Seek[2].Value>10) or ((Seek[2].Pos-1) mod 288+1=J));
        {Then look west}
        WestSteps:=0;
        Seek[4].Pos:=J+288;
        repeat
          inc(WestSteps);
          dec(Seek[4].Pos);
          Seek[4].Value:=Data[(Seek[4].Pos-1) mod 288+1,I];
        until ((Seek[4].Value>10) or ((Seek[4].Pos-1) mod 288+1=J));
        Seek[4].Pos:=Seek[4].Pos-288;

        {Now to use the east and west values they mustn't be at the same     }
        {location (this can occur if there was only 1 value at this latitude)}
        {and they mustn't be more than MaxLongInterpol degrees apart.        }
        if ((Seek[2].Value>0) and (Seek[4].Value>0) and (Seek[2].Pos<>Seek[4].Pos)) then
        begin
          LongDist:=EastSteps+WestSteps;
          if LongDist>MaxLongInterpol/1.25 then LongInterpolValue:=-999.0 else
            LongInterpolValue:=WestSteps*(Seek[2].Value-Seek[4].Value)/LongDist+Seek[4].Value;
        end;

        {Now look north and south for values for (1) and (3)}
        {First look north}
        if DoNSSearch then
        begin
          Seek[1].Pos:=I;
          SeekLong:=J;
          NSSearchDone:=false;
          NorthSteps:=0;
          repeat
            inc(NorthSteps);
            if SeekLong=J then dec(Seek[1].Pos) else inc(Seek[1].Pos);
            if Seek[1].Pos<1 then
            if Coverage=TomsSouth then NSSearchDone:=true else
            begin
              Seek[1].Pos:=1;
              if SeekLong<=144 then SeekLong:=SeekLong+144
              else SeekLong:=SeekLong-144;
            end;
            if Seek[1].Pos>EndInd then
            if Coverage=TomsNorth then NSSearchDone:=true else
            begin
              Seek[1].Pos:=EndInd;
              if SeekLong<=144 then SeekLong:=SeekLong+144
              else SeekLong:=SeekLong-144;
            end;
            if not NSSearchDone then
            begin
              Seek[1].Value:=Data[SeekLong,Seek[1].Pos];
              if Seek[1].Value>0 then NSSearchDone:=true;
              if ((SeekLong=J) and (Seek[1].Pos=I)) then NSSearchDone:=true; {Back at start.}
            end;
          until NSSearchDone;
          {Then look south.}
          Seek[3].Pos:=I;
          SeekLong:=J;
          NSSearchDone:=false;
          SouthSteps:=0;
          repeat
            inc(SouthSteps);
            if SeekLong=J then inc(Seek[3].Pos) else dec(Seek[3].Pos);
            if Seek[3].Pos>EndInd then
            if Coverage=TomsNorth then NSSearchDone:=true else
            begin
              Seek[3].Pos:=EndInd;
              if SeekLong<=144 then SeekLong:=SeekLong+144
              else SeekLong:=SeekLong-144;
            end;
            if Seek[3].Pos<1 then
            if Coverage=TomsSouth then NSSearchDone:=true else
            begin
              Seek[3].Pos:=1;
              if SeekLong<=144 then SeekLong:=SeekLong+144
              else SeekLong:=SeekLong-144;
            end;
            if not NSSearchDone then
            begin
              Seek[3].Value:=Data[SeekLong,Seek[3].Pos];
              if Seek[3].Value>0 then NSSearchDone:=true;
              if ((SeekLong=J) and (Seek[3].Pos=I)) then NSSearchDone:=true; {Back at start.}
            end;
          until NSSearchDone;
          if ((Seek[1].Value>0) and (Seek[3].Value>0)) then
          begin
            LatDist:=NorthSteps+SouthSteps;
            if ((abs(LatCellDeg(I))<OverThePoleLatCutoff) and (LatDist>MaxLatInterpol)) then
            LatInterpolValue:=-999.0 else LatInterpolValue:=SouthSteps*(Seek[1].Value-
            Seek[3].Value)/LatDist+Seek[3].Value;
          end;
        end;

        if LongInterpolValue>0 then
        begin
          if ((LatInterpolValue>0) and (LatDist*5<LongDist)) then Data[J,I]:=LatInterpolValue
          else Data[J,I]:=LongInterpolValue;
        end
        else if LatInterpolValue>0 then Data[J,I]:=LatInterpolValue;
        if Data[J,I]<1.0 then StillMissing:=true;
      end; {End of loop through J.}
    end; {End of loop through Counter.}
    if not StillMissing then Pass:=MaxPasses;
  until Pass>=MaxPasses;

  {Finally check to see if data are still missing.}
  for Counter:=StartInd to EndInd do
  begin
    if ((Coverage=TomsGlobal) and (Counter<91)) then I:=91-Counter else I:=Counter;
    if Coverage=TomsNorth then I:=91-Counter;
    for J:=1 to 288 do
    if ((Data[J,I]<=1.0) and (abs(LatCellDeg(I))>=DataAvailLatCutoff)) then
      raise EMercToms.Create('the value at longindex='+inttostr(J)+' and latindex='+
      inttostr(I)+' could not be filled');
  end;
end; {End of TMercToms.PatchToms.}

procedure TMercToms.LongitudeInterpol;
{-----------------------------------------------------------------------------}
{ This function interpolates across data gaps in longitude only.              }
{-----------------------------------------------------------------------------}
var
  L1,L2,D1,D2:integer;
  N,Ind,I,J:word;
begin
  if Coverage=TomsGlobal then Ind:=180 else Ind:=90;
  for I:=1 to Ind do
  begin
    N:=0;
    for J:=1 to 288 do if Data[J,I]<>Null then inc(N);
    if N>1 then
    for J:=1 to 288 do
    begin
      if Data[J,I]=Null then
      begin
        L1:=J;
        D1:=0;
        repeat
          dec(L1);
          inc(D1);
          if L1<1 then L1:=L1+288;
        until Data[L1,I]<>Null;
        L2:=J;
        D2:=0;
        repeat
          inc(L2);
          inc(D2);
          if L2>288 then L2:=L2-288;
        until Data[L2,I]<>Null;
        Data[J,I]:=round(D1*(Data[L2,I]-Data[L1,I])/(D1+D2)+Data[L1,I]);
      end;
    end;
  end;
end; {End of TMercToms.LongitudeInterpol.}

procedure TMercToms.GeophysInterpolTOMSPolar(GeoData:TSynopticGlobalArray;IntStyle:TIntStyle);
{------------------------------------------------------------------------------}
{ This routine interpolates the missing ozone data across the polar caps only. }
{ It makes use of the global two dimensional geophysical data passed via the   }
{ GeoData array, using the Extract method, as dynamical proxies. The following }
{ interpolation styles are used depending on the value passed via IntStyle:    }
{ median: An ozone value is estimated from all dynamical proxies and the       }
{         median value is selected.                                            }
{ bestr2: The ozone value is estimated from the dynamical proxy that produced  }
{         the best R squared value.                                            }
{ combination: a linear least squares regression model using all dynamical     }
{              proxies as basis functions is used to estimate the ozone value. }
{ It uses the available ozone data within 5 degrees of the terminator to train }
{ the various models.                                                          }
{------------------------------------------------------------------------------}
var
  M,C,
  Ozone,
  X,Y:PMathArray;
  PredOzone,
  Latitude,
  Longitude,
  RSquared,
  BestRSquared:double;
  ValueAvail:boolean;
  HemisCount,
  StartInd,
  EndInd:byte;
  GeoCount,
  LastLat,
  StopLat,
  R1,R2,R3:integer;
  BestGeo,
  NumGeoData,
  I,J:word;
  N:longint;
  Reg:TInterPolReg;
begin
  NumGeoData:=0;
  for I:=1 to 50 do if GeoData[I]<>nil then inc(NumGeoData);
  StartInd:=1;
  EndInd:=2;
  if Coverage=TomsSouth then EndInd:=1;
  if Coverage=TomsNorth then StartInd:=2;
  for HemisCount:=StartInd to EndInd do {south=1, north=2}
  begin
    LastLat:=0;
    R1:=1;
    StopLat:=90;
    if HemisCount=1 then
    begin
      if Coverage=TomsSouth then LastLat:=91 else LastLat:=181;
      if Coverage=TomsSouth then R1:=50 else R1:=140;
      StopLat:=LastLat-90;
    end;
    repeat
      LastLat:=LastLat+(HemisCount*2-3);
      ValueAvail:=false;
      for I:=1 to 288 do if Data[I,LastLat]>0.0 then ValueAvail:=true;
    until ((LastLat=StopLat) or ValueAvail);
    if ((LastLat>R1) and (LastLat<(R1+40))) then
    begin
      N:=0;
      if HemisCount=1 then R2:=LastLat-4 else R2:=1;
      if HemisCount=1 then R3:=R1+40 else R3:=LastLat+4;
      for I:=1 to 288 do
        for J:=R2 to R3 do
          if Data[I,J]>0.0 then inc(N);
      if N>MaxMathArray then N:=MaxMathArray;
      M:=nil;
      C:=nil;
      X:=nil;
      Y:=nil;
      Ozone:=nil;
      BestGeo:=0;
      if IntStyle in [median,bestr2] then
      begin
        getmem(X,N*sizeof(double));
        getmem(Y,N*sizeof(double));
        getmem(M,NumGeoData*sizeof(double));
        getmem(C,NumGeoData*sizeof(double));
        if IntStyle=median then getmem(Ozone,NumGeoData*sizeof(double));
      end;
      if IntStyle in [combination] then
      begin
        Reg:=TInterPolReg.Create(NumGeoData+1,N);
        for I:=1 to NumGeoData do getmem(Reg.BasisArray[I],N*sizeof(double));
      end
      else Reg:=nil;
      N:=0;
      for I:=1 to 288 do
        for J:=R2 to R3 do
        if Data[I,J]>0.0 then
        begin
          if N<MaxMathArray then inc(N);
          if IntStyle in [median,bestr2] then Y^[N]:=Data[I,J];
          if IntStyle in [combination] then
          begin
            Latitude:=LatCellDeg(J);
            Longitude:=LongCellDeg(I);
            Reg.X^[N]:=N;
            Reg.Y^[N]:=Data[I,J];
            Reg.Sigma^[N]:=1.0;
            for GeoCount:=1 to NumGeoData do Reg.BasisArray[GeoCount]^[N]:=
              GeoData[GeoCount].BiLinear(Latitude,Longitude);
          end;
        end;
      if IntStyle in [median,bestr2] then
      begin
        BestRSquared:=0.0;
        for GeoCount:=1 to NumGeoData do
        begin
          N:=0;
          for I:=1 to 288 do
            for J:=R2 to R3 do
            if Data[I,J]>0.0 then
            begin
              Latitude:=LatCellDeg(J);
              Longitude:=LongCellDeg(I);
              if N<MaxMathArray then inc(N);
              X^[N]:=GeoData[GeoCount].BiLinear(Latitude,Longitude);
            end;
          RSquared:=sqr(Regression(X,Y,N,M^[GeoCount],C^[GeoCount],false));
          if RSquared>BestRSquared then
          begin
            BestRSquared:=RSquared;
            BestGeo:=GeoCount;
          end;
        end;
      end;
      if IntStyle in [combination] then Reg.Run;
      if HemisCount=1 then R2:=LastLat+1 else R2:=1;
      if HemisCount=1 then R3:=R1+40 else R3:=LastLat-1;
      for I:=1 to 288 do
        for J:=R2 to R3 do
        if Data[I,J]<=0.0 then
        begin
          Latitude:=LatCellDeg(J);
          Longitude:=LongCellDeg(I);
          PredOzone:=0.0;
          if IntStyle=median then
          begin
            for GeoCount:=1 to NumGeoData do Ozone^[GeoCount]:=
              M^[GeoCount]*GeoData[GeoCount].BiLinear(Latitude,Longitude)+C^[GeoCount];
            PredOzone:=CalcMedian(Ozone,NumGeoData);
          end;
          if IntStyle=bestr2 then PredOzone:=M^[BestGeo]*GeoData[BestGeo].
            BiLinear(Latitude,Longitude)+C^[BestGeo];
          if IntStyle=combination then
          begin
            PredOzone:=Reg.A^[Reg.NumCoeff];
            for GeoCount:=1 to NumGeoData do PredOzone:=PredOzone+Reg.A^[GeoCount]*
              GeoData[GeoCount].BiLinear(Latitude,Longitude);
          end;
          if ((PredOzone<1000.0) and (PredOzone>10.0)) then Data[I,J]:=PredOzone;
        end;
      if IntStyle in [median,bestr2] then
      begin
        freemem(X,N*sizeof(double));
        freemem(Y,N*sizeof(double));
        freemem(M,NumGeoData*sizeof(double));
        freemem(C,NumGeoData*sizeof(double));
        if IntStyle=median then freemem(Ozone,NumGeoData*sizeof(double));
      end;
      if IntStyle in [combination] then
      begin
        for I:=1 to NumGeoData do freemem(Reg.BasisArray[I],N*sizeof(double));
        Reg.Free;
      end;
    end;
  end;
end; {End of TMercToms.GeophysInterpolTOMSPolar.}

function TMercToms.Missing(Hemis:THemis):double;
{-----------------------------------------------------------------------------}
{ This function returns the fraction of data missing in the hemisphere        }
{ specified by 'Hemis'.                                                       }
{-----------------------------------------------------------------------------}
var
  I,J,
  Index1,
  Index2,
  NumNull:word;
begin
  Index1:=1;
  Index2:=90;
  if ((Coverage=TomsGlobal) and (Hemis=TomsSouth)) then Index1:=91;
  if ((Coverage=TomsGlobal) and (Hemis=TomsSouth)) then Index2:=180;
  if ((Coverage=TomsNorth) and (Hemis=TomsSouth)) then Index2:=0;
  if ((Coverage=TomsSouth) and (Hemis=TomsNorth)) then Index2:=0;
  NumNull:=25920;
  for I:=Index1 to Index2 do
    for J:=1 to 288 do
      if not EquivDouble(Data[J,I],Null) then dec(NumNull);
  Result:=NumNull/25920.0;
end; {End of TMercToms.Missing.}

procedure TMercToms.CorrectData(F:CorrFunction;Year,Day:word);
var
  I,J:integer;
  Corr:double;
begin
  for I:=1 to NumLat do
  begin
    Corr:=F(LatCellDeg(I),Year,Day);
    for J:=1 to NumLong do if Data[J,I]>10.0 then Data[J,I]:=Data[J,I]+Corr;
  end;
end; {End of TMercToms.CorrectData.}

function TMercToms.GetDataType:byte;
begin
  Result:=DataType;
end;

function TMercToms.GetListIndex:byte;
begin
  Result:=ListIndex;
end;

function TMercToms.GetDataFilename:string;
begin
  Result:=DataFilename;
end;

function TMercToms.GetCoverage:THemis;
begin
  Result:=Coverage;
end;

procedure TMercToms.PutDataType(Value:byte);
begin
  DataType:=Value;
end;

procedure TMercToms.PutListIndex(Value:byte);
begin
  ListIndex:=Value;
end;

procedure TMercToms.PutDataFilename(Value:string);
begin
  DataFilename:=Value;
end;

procedure SetNimbus7Dir(ADir:string);
const
  CheckFilename='n7t1980001.syn';
begin
  Nimbus7Dir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(Nimbus7Dir,'1980')+'\'+CheckFilename))) then
    raise EMercToms.Create('The Nimbus 7 TOMS directory '+Nimbus7Dir+' is mostly likely wrong');
end;

procedure SetMeteor3Dir(ADir:string);
const
  CheckFilename='m3t1991234.syn';
begin
  Meteor3Dir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(Meteor3Dir,'1991')+'\'+CheckFilename))) then
    raise EMercToms.Create('The Meteor 3 TOMS directory '+Meteor3Dir+' is mostly likely wrong');
end;

procedure SetDLRGomeDir(ADir:string);
const
  CheckFilename='DLRGome1995178.syn';
begin
  DLRGomeDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(DLRGomeDir,'1995')+'\'+CheckFilename))) then
    raise EMercToms.Create('The DLR GOME directory '+Meteor3Dir+' is mostly likely wrong');
end;

procedure SetKNMIGomeDir(ADir:string);
const
  CheckFilename='knmi_togomi_2000003.syn';
begin
  KNMIGomeDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(KNMIGomeDir,'2000')+'\'+CheckFilename))) then
    raise EMercToms.Create('The KNMI GOME directory '+KNMIGomeDir+' is mostly likely wrong');
end;

procedure SetGOAGomeDir(ADir:string);
const
  CheckFilename='KNMI_gome_200000100.syn';
begin
  GOAGomeDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(GOAGomeDir,'2000')+'\'+CheckFilename))) then
    raise EMercToms.Create('The GOA GOME directory '+GOAGomeDir+' is mostly likely wrong');
end;

procedure SetEarthProbeDir(ADir:string);
const
  CheckFilename='epc2004001.syn';
begin
  EarthProbeDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(EarthProbeDir,'2004')+'\'+CheckFilename))) then
    raise EMercToms.Create('The Earth Probe TOMS directory '+EarthProbeDir+' is mostly likely wrong');
end;

procedure SetAdeosDir(ADir:string);
const
  CheckFilename='adt1996255.syn';
begin
  AdeosDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(AdeosDir,'1996')+'\'+CheckFilename))) then
    raise EMercToms.Create('The Adeos TOMS directory '+AdeosDir+' is mostly likely wrong');
end;

procedure SetNimbus7SBUVDir(ADir:string);
const
  CheckFilename='SBUV_N7_1984001.syn';
begin
  Nimbus7SBUVDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(Nimbus7SBUVDir,'1984')+'\'+CheckFilename))) then
    raise EMercToms.Create('The Nimbus 7 SBUV directory '+Nimbus7SBUVDir+' is mostly likely wrong');
end;

procedure SetNOAA9SBUVDir(ADir:string);
const
  CheckFilename='SBUV_N9_1987001.syn';
begin
  NOAA9SBUVDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(NOAA9SBUVDir,'1987')+'\'+CheckFilename))) then
    raise EMercToms.Create('The NOAA 9 SBUV directory '+NOAA9SBUVDir+' is mostly likely wrong');
end;

procedure SetNOAA11SBUVDir(ADir:string);
const
  CheckFilename='SBUV_N11_1991002.syn';
begin
  NOAA11SBUVDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(NOAA11SBUVDir,'1991')+'\'+CheckFilename))) then
    raise EMercToms.Create('The NOAA 11 SBUV directory '+NOAA11SBUVDir+' is mostly likely wrong');
end;

procedure SetNOAA16SBUVDir(ADir:string);
const
  CheckFilename='SBUV_N16_2003002.syn';
begin
  NOAA16SBUVDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(NOAA16SBUVDir,'2003')+'\'+CheckFilename))) then
    raise EMercToms.Create('The NOAA 16 SBUV directory '+NOAA16SBUVDir+' is mostly likely wrong');
end;

procedure SetOMIDir(ADir:string);
const
  CheckFilename='omi2009001.syn';
begin
  OMIDir:=ADir;
  if ((ADir<>'') and (not FileExists(InsertYear(OMIDir,'2009')+'\'+CheckFilename))) then
    raise EMercToms.Create('The OMI directory '+OMIDir+' is mostly likely wrong');
end;

procedure LoadN7Coeff(Filename:string);
var
  Source:textfile;
  I:byte;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    for I:=1 to 22 do readln(Source,N7Coeff[I]);
    closefile(Source);
  end
  else raise EMercToms.Create('the N7 coefficient correction file named '+Filename+
    'couldn''t be found');
end;

function CorrectN7(Lat:double;Year,Day:word):double;
var
  DecYear,
  TestLat:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  TestLat:=Lat;
  if TestLat>65.0 then TestLat:=65.0;
  TestLat:=DegToRad(TestLat);
  Result:=N7Coeff[1];
  Result:=Result+N7Coeff[2]*sin(TestLat);
  Result:=Result+N7Coeff[3]*cos(TestLat);
  Result:=Result+N7Coeff[4]*sin(2*TestLat);
  Result:=Result+N7Coeff[5]*cos(2*TestLat);
  Result:=Result+N7Coeff[6]*sin(3*TestLat);
  Result:=Result+N7Coeff[7]*cos(3*TestLat);
  Result:=Result+N7Coeff[8]*DecYear;
  Result:=Result+N7Coeff[9]*sin(TestLat)*DecYear;
  Result:=Result+N7Coeff[10]*cos(TestLat)*DecYear;
  Result:=Result+N7Coeff[11]*sin(DecYear*2*Pi);
  Result:=Result+N7Coeff[12]*sin(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+N7Coeff[13]*cos(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+N7Coeff[14]*cos(DecYear*2*Pi);
  Result:=Result+N7Coeff[15]*sin(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+N7Coeff[16]*cos(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+N7Coeff[17]*sin(DecYear*4*Pi);
  Result:=Result+N7Coeff[18]*sin(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+N7Coeff[19]*cos(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+N7Coeff[20]*cos(DecYear*4*Pi);
  Result:=Result+N7Coeff[21]*sin(TestLat)*cos(DecYear*4*Pi);
  Result:=Result+N7Coeff[22]*cos(TestLat)*cos(DecYear*4*Pi);
end;

{$HINTS OFF}
procedure LoadM3Coeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,M3Coeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The M3 coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectM3(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  Result:=M3Coeff[1,Ind]+M3Coeff[2,Ind]*DecYear+M3Coeff[3,Ind]*sin(2*pi*
    frac(DecYear))+M3Coeff[4,Ind]*cos(2*pi*frac(DecYear))+M3Coeff[5,Ind]*
    sin(4*pi*frac(DecYear))+M3Coeff[6,Ind]*cos(4*pi*frac(DecYear));
end;

procedure LoadEPCoeff(Filename:string);
var
  Source:textfile;
  I:byte;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    for I:=1 to 22 do readln(Source,EPCoeff[I]);
    closefile(Source);
  end
  else raise EMercToms.Create('the EP coefficient correction file named '+Filename+
    'couldn''t be found');
end;

function CorrectEP(Lat:double;Year,Day:word):double;
var
  DecYear,
  TestLat:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  TestLat:=Lat;
  if TestLat>65.0 then TestLat:=65.0;
  TestLat:=DegToRad(TestLat);
  Result:=EPCoeff[1];
  Result:=Result+EPCoeff[2]*sin(TestLat);
  Result:=Result+EPCoeff[3]*cos(TestLat);
  Result:=Result+EPCoeff[4]*sin(2*TestLat);
  Result:=Result+EPCoeff[5]*cos(2*TestLat);
  Result:=Result+EPCoeff[6]*sin(3*TestLat);
  Result:=Result+EPCoeff[7]*cos(3*TestLat);
  Result:=Result+EPCoeff[8]*DecYear;
  Result:=Result+EPCoeff[9]*sin(TestLat)*DecYear;
  Result:=Result+EPCoeff[10]*cos(TestLat)*DecYear;
  Result:=Result+EPCoeff[11]*sin(DecYear*2*Pi);
  Result:=Result+EPCoeff[12]*sin(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+EPCoeff[13]*cos(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+EPCoeff[14]*cos(DecYear*2*Pi);
  Result:=Result+EPCoeff[15]*sin(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+EPCoeff[16]*cos(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+EPCoeff[17]*sin(DecYear*4*Pi);
  Result:=Result+EPCoeff[18]*sin(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+EPCoeff[19]*cos(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+EPCoeff[20]*cos(DecYear*4*Pi);
  Result:=Result+EPCoeff[21]*sin(TestLat)*cos(DecYear*4*Pi);
  Result:=Result+EPCoeff[22]*cos(TestLat)*cos(DecYear*4*Pi);
end;

{$HINTS OFF}
procedure LoadADCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,ADCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The AD coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectAD(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  Result:=ADCoeff[1,Ind]+ADCoeff[2,Ind]*DecYear+ADCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+ADCoeff[4,Ind]*cos(2*pi*frac(DecYear))+ADCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+ADCoeff[6,Ind]*cos(4*pi*frac(DecYear));
end;

{$HINTS OFF}
procedure LoadGOCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,GOCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The DLR GOME coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectGO(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=GOCoeff[1,Ind]+GOCoeff[2,Ind]*DecYear+GOCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+GOCoeff[4,Ind]*cos(2*pi*frac(DecYear))+GOCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+GOCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadGKCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,GKCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The KNMI GOME coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectGK(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=GKCoeff[1,Ind]+GKCoeff[2,Ind]*DecYear+GKCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+GKCoeff[4,Ind]*cos(2*pi*frac(DecYear))+GKCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+GKCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadGGCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,GGCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The KNMI GOME coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectGG(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=GGCoeff[1,Ind]+GGCoeff[2,Ind]*DecYear+GGCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+GGCoeff[4,Ind]*cos(2*pi*frac(DecYear))+GGCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+GGCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadN7SCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,N7SCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The Nimbus 7 SBUV coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectN7S(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=N7SCoeff[1,Ind]+N7SCoeff[2,Ind]*DecYear+N7SCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+N7SCoeff[4,Ind]*cos(2*pi*frac(DecYear))+N7SCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+N7SCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadN9SCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,N9SCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The NOAA 9 SBUV coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectN9S(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=N9SCoeff[1,Ind]+N9SCoeff[2,Ind]*DecYear+N9SCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+N9SCoeff[4,Ind]*cos(2*pi*frac(DecYear))+N9SCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+N9SCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadN11SCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,N11SCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The NOAA 11 SBUV coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectN11S(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=N11SCoeff[1,Ind]+N11SCoeff[2,Ind]*DecYear+N11SCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+N11SCoeff[4,Ind]*cos(2*pi*frac(DecYear))+N11SCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+N11SCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadN16SCoeff(Filename:string);
var
  Source:textfile;
  I,J:integer;
  Lat:double;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    DeleteHeader(Source);
    for I:=1 to 90 do
    begin
      read(Source,Lat);
      for J:=1 to 6 do read(Source,N16SCoeff[J,I]);
      readln(Source);
    end;
    closefile(Source);
  end
  else raise EMercToms.Create('The NOAA 16 SBUV coefficient correction file couldn''t be found');
end;
{$HINTS ON}

function CorrectN16S(Lat:double;Year,Day:word):double;
var
  Ind:integer;
  DecYear:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  Ind:=trunc((Lat+90)/2.0)+1;
  if Ind>90 then Ind:=90;
  if Ind=1 then Ind:=2;
  if Ind=90 then Ind:=89;
  Result:=N16SCoeff[1,Ind]+N16SCoeff[2,Ind]*DecYear+N16SCoeff[3,Ind]*sin(2*pi*
    frac(DecYear))+N16SCoeff[4,Ind]*cos(2*pi*frac(DecYear))+N16SCoeff[5,Ind]*
    sin(4*pi*frac(DecYear))+N16SCoeff[6,Ind]*cos(4*pi*frac(DecYear));
  if Result>30 then Result:=30.0;
  if Result<-30 then Result:=30.0;
end;

{$HINTS OFF}
procedure LoadOMICoeff(Filename:string);
var
  Source:textfile;
  I:byte;
begin
  if FileExists(Filename) then
  begin
    assignfile(Source,Filename);
    reset(Source);
    for I:=1 to 22 do readln(Source,OMICoeff[I]);
    closefile(Source);
  end
  else raise EMercToms.Create('the OMI coefficient correction file named '+Filename+'couldn''t be found');
end;
{$HINTS ON}

function CorrectOMI(Lat:double;Year,Day:word):double;
var
  DecYear,
  TestLat:double;
begin
  DecYear:=Year+(Day-0.5)/DaysInAYear(Year)-1980;
  TestLat:=Lat;
  if TestLat>65.0 then TestLat:=65.0;
  TestLat:=DegToRad(TestLat);
  Result:=OMICoeff[1];
  Result:=Result+OMICoeff[2]*sin(TestLat);
  Result:=Result+OMICoeff[3]*cos(TestLat);
  Result:=Result+OMICoeff[4]*sin(2*TestLat);
  Result:=Result+OMICoeff[5]*cos(2*TestLat);
  Result:=Result+OMICoeff[6]*sin(3*TestLat);
  Result:=Result+OMICoeff[7]*cos(3*TestLat);
  Result:=Result+OMICoeff[8]*DecYear;
  Result:=Result+OMICoeff[9]*sin(TestLat)*DecYear;
  Result:=Result+OMICoeff[10]*cos(TestLat)*DecYear;
  Result:=Result+OMICoeff[11]*sin(DecYear*2*Pi);
  Result:=Result+OMICoeff[12]*sin(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+OMICoeff[13]*cos(TestLat)*sin(DecYear*2*Pi);
  Result:=Result+OMICoeff[14]*cos(DecYear*2*Pi);
  Result:=Result+OMICoeff[15]*sin(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+OMICoeff[16]*cos(TestLat)*cos(DecYear*2*Pi);
  Result:=Result+OMICoeff[17]*sin(DecYear*4*Pi);
  Result:=Result+OMICoeff[18]*sin(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+OMICoeff[19]*cos(TestLat)*sin(DecYear*4*Pi);
  Result:=Result+OMICoeff[20]*cos(DecYear*4*Pi);
  Result:=Result+OMICoeff[21]*sin(TestLat)*cos(DecYear*4*Pi);
  Result:=Result+OMICoeff[22]*cos(TestLat)*cos(DecYear*4*Pi);
end;

function LoadTomsData(TomsData:TMercToms;Year,Day,StartIndex:word;CorrectData,LongInterpol,ApplyCutoffs,Composite:boolean):boolean;
var
  TomsList:TTomsList;
begin
  setlength(TomsList,NumSatSets);
  {Here is where the default satellite data set order is set.}
  TomsList[0]:=N7T;
  TomsList[1]:=M3T;
  TomsList[2]:=OMI;
  TomsList[3]:=EPT;
  TomsList[4]:=ADT;
  TomsList[5]:=GOA;
  TomsList[6]:=GKN;
  TomsList[7]:=DLR;
  TomsList[8]:=N7S;
  TomsList[9]:=N9S;
  TomsList[10]:=N11S;
  TomsList[11]:=N16S;
  Result:=LoadTomsData(TomsData,Year,Day,StartIndex,TomsList,CorrectData,LongInterpol,ApplyCutoffs,Composite);
end;

function LoadTomsData(TomsData:TMercToms;Year,Day,StartIndex:word;TomsList:TTomsList;CorrectData,LongInterpol,ApplyCutoffs,Composite:boolean):boolean;
var
  Fields:array[1..12] of TMercToms;
  Hold:TMercToms;
  YearStr,
  DayStr,
  Filename:string;
  ListCount,
  LatInd,
  I,J,K,L,M:integer;
  LongInd:cardinal;
  HoldValue,
  CutOff:double;
  DLROK,
  Done:boolean;
  SynopticGlobalData:TSynopticGlobal;
  TestArray:array[-1..1,-1..1] of double;
begin
  Done:=false;
  TomsData.Clear;
  YearStr:=IntToStrF(Year,4,true);
  DayStr:=IntToStrF(Day,3,true);
  ListCount:=StartIndex-2;
  if Composite then for I:=1 to 12 do Fields[I]:=nil;
  repeat
    inc(ListCount);
    case TomsList[ListCount] of
      N7T:Filename:=InsertYear(Nimbus7Dir,YearStr)+'\n7t'+YearStr+DayStr+'.syn';
      M3T:Filename:=InsertYear(Meteor3Dir,YearStr)+'\m3t'+YearStr+DayStr+'.syn';
      EPT:Filename:=InsertYear(EarthProbeDir,YearStr)+'\epc'+YearStr+DayStr+'.syn';
      ADT:Filename:=InsertYear(AdeosDir,YearStr)+'\adt'+YearStr+DayStr+'.syn';
      OMI:Filename:=InsertYear(OMIDir,YearStr)+'\omi'+YearStr+DayStr+'.syn';
      GOA:Filename:=InsertYear(GOAGomeDir,YearStr)+'\KNMI_gome_'+YearStr+DayStr+'00.syn';
      GKN:Filename:=InsertYear(KNMIGomeDir,YearStr)+'\knmi_togomi_'+YearStr+DayStr+'.syn';
      DLR:Filename:=InsertYear(DLRGomeDir,YearStr)+'\DLRGome'+YearStr+DayStr+'.syn';
      N7S:Filename:=InsertYear(Nimbus7SBUVDir,YearStr)+'\SBUV_N7_'+YearStr+DayStr+'.syn';
      N9S:Filename:=InsertYear(NOAA9SBUVDir,YearStr)+'\SBUV_N9_'+YearStr+DayStr+'.syn';
      N11S:Filename:=InsertYear(NOAA11SBUVDir,YearStr)+'\SBUV_N11_'+YearStr+DayStr+'.syn';
      N16S:Filename:=InsertYear(NOAA16SBUVDir,YearStr)+'\SBUV_N16_'+YearStr+DayStr+'.syn';
    end;
    if FileExists(Filename) then
    try
      SynopticGlobalData:=TSynopticGlobal.Create(Filename);
      if Composite then Fields[ListCount+1]:=TMercToms.Create(TomsData.GetCoverage) else Done:=true;
      for I:=1 to TomsData.NumLat do
        for J:=1 to TomsData.NumLong do
        begin
          HoldValue:=-999.99;
          try
            HoldValue:=SynopticGlobalData.GetValueByCoords(TomsData.LongCellDeg(J),TomsData.LatCellDeg(I),false);
            if HoldValue>1E10 then HoldValue:=-999.99;
          except
            on E:ESynopticGlobal do; {Do nothing since HoldValue has been set to -999.99.}
          end;
          if Composite then Fields[ListCount+1][J,I]:=HoldValue else TomsData[J,I]:=HoldValue;
        end;
    finally
      FreeAndNil(SynopticGlobalData);
    end;

    {There are some glitches in the DLR files see e.g. NH for day 228 of 1995.}
    if ((TomsList[ListCount]=DLR) and FileExists(Filename)) then
    begin
      Hold:=TMercToms.Create(TomsData.Coverage);
      try
        for J:=1 to TomsData.NumLat do
          for K:=1 to TomsData.NumLong do
          if Composite then Hold[K,J]:=Fields[ListCount+1][K,J] else
            Hold[K,J]:=TomsData[K,J];

        for K:=1 to TomsData.NumLong do
          for J:=1 to TomsData.NumLat do
          if Hold[K,J]<50 then
          begin
            if Composite then Fields[ListCount+1][K,J]:=-999.99 else
            TomsData[K,J]:=-999.99;
          end
          else
          begin
            {Collect neighbours}
            for L:=-1 to 1 do
            begin
              LongInd:=K+L;
              if LongInd<1 then LongInd:=TomsData.NumLong;
              if LongInd>TomsData.NumLong then LongInd:=1;
              for M:=-1 to 1 do
              begin
                LatInd:=J+M;
                if LatInd<1 then LatInd:=1;
                if cardinal(LatInd)>TomsData.NumLat then LatInd:=TomsData.NumLat;
                if Hold[LongInd,LatInd]>10.0 then TestArray[L,M]:=Hold[LongInd,LatInd] else TestArray[L,M]:=-999.99;
              end;
            end;

            {If any non-null entry in TestArray is more than 50DU away from the center entry, we}
            {set the value to null.}
            DLROK:=true;
            for L:=-1 to 1 do
              for M:=-1 to 1 do
                if ((TestArray[L,M]>-999) and (abs(TestArray[L,M]-TestArray[0,0])>50)) then
                  DLROK:=false;

            {If within 30 degrees of the north pole and less than 180DU, then kill. Note}
            {this isn't a big issue since its just DLR data we are messing with.}
            if (DLROK and (J<=30) and (TestArray[0,0]<180)) then
              DLROK:=false;

            {If the value is more than 650DU, then kill. Note}
            {this isn't a big issue since its just DLR data we are messing with.}
            if (DLROK and (TestArray[0,0]>650)) then
              DLROK:=false;

            if not DLROK then
              if Composite then Fields[ListCount+1][K,J]:=-999.99 else TomsData[K,J]:=-999.99;
          end;
      finally
        Hold.Free;
      end;
    end;
    if ApplyCutoffs then
    begin
      CutOff:=0.66;
      if TomsList[ListCount] in [GOA,GKN] then CutOff:=0.95; {DLR GOME and TOGOMI}
      if TomsList[ListCount] in [N7S,N9S,N11S,N16S] then CutOff:=0.996142; {SBUV}
      if (Composite and (Fields[ListCount+1]<>nil)) then
      begin
        if Fields[ListCount+1].Missing(TomsData.Coverage)>Cutoff then Fields[ListCount+1].Clear;
      end
      else if TomsData.Missing(TomsData.Coverage)>Cutoff then Done:=false;
    end;
  until ((ListCount=High(TomsList)) or Done);

  {First correct all of the data.}
  if CorrectData then
  begin
    if Composite then
    begin
      for I:=1 to 12 do
      if Fields[I]<>nil then
      begin
        case TomsList[I-1] of
          N7T:Fields[I].CorrectData(CorrectN7,Year,Day);
          M3T:Fields[I].CorrectData(CorrectM3,Year,Day);
          EPT:Fields[I].CorrectData(CorrectEP,Year,Day);
          ADT:Fields[I].CorrectData(CorrectAD,Year,Day);
          OMI:Fields[I].CorrectData(CorrectOMI,Year,Day);
          GOA:Fields[I].CorrectData(CorrectGG,Year,Day);
          GKN:Fields[I].CorrectData(CorrectGK,Year,Day);
          DLR:Fields[I].CorrectData(CorrectGO,Year,Day);
          N7S:Fields[I].CorrectData(CorrectN7S,Year,Day);
          N9S:Fields[I].CorrectData(CorrectN9S,Year,Day);
          N11S:Fields[I].CorrectData(CorrectN11S,Year,Day);
          N16S:Fields[I].CorrectData(CorrectN16S,Year,Day);
        end;
      end;
    end;
    if ((not Composite) and Done) then
    begin
      case TomsList[ListCount] of
        N7T:TomsData.CorrectData(CorrectN7,Year,Day);
        M3T:TomsData.CorrectData(CorrectM3,Year,Day);
        EPT:TomsData.CorrectData(CorrectEP,Year,Day);
        ADT:TomsData.CorrectData(CorrectAD,Year,Day);
        OMI:TomsData.CorrectData(CorrectOMI,Year,Day);
        GOA:TomsData.CorrectData(CorrectGG,Year,Day);
        GKN:TomsData.CorrectData(CorrectGK,Year,Day);
        DLR:TomsData.CorrectData(CorrectGO,Year,Day);
        N7S:TomsData.CorrectData(CorrectN7S,Year,Day);
        N9S:TomsData.CorrectData(CorrectN9S,Year,Day);
        N11S:TomsData.CorrectData(CorrectN11S,Year,Day);
        N16S:TomsData.CorrectData(CorrectN16S,Year,Day);
      end;
    end;
  end;

  {Now generate the composite data set if selected.}
  if Composite then
  begin
    Done:=false;
    for I:=1 to 12 do
    if Fields[I]<>nil then
    begin
      for J:=1 to TomsData.NumLat do
        for K:=1 to TomsData.NumLong do
          if ((Fields[I][K,J]>10.0) and (TomsData[K,J]<=10.0)) then
          begin
            Done:=true;
            TomsData[K,J]:=Fields[I][K,J];
          end;
      Fields[I].Free;
    end;
  end;

  {Do the final jobs.}
  if Done then
  begin
    if Composite then
    begin
      TomsData.PutDataType(0);
      TomsData.PutListIndex(0);
      TomsData.PutDataFilename('Composite');
    end
    else
    begin
      TomsData.PutDataType(ord(TomsList[ListCount])+1);
      TomsData.PutListIndex(ListCount+1);
      TomsData.PutDataFilename(Filename);
    end;
    if LongInterpol then TomsData.LongitudeInterpol;
  end;
  Result:=Done;
end;

end. {End of merctoms unit.}

