//@***h* BDBP/gb_strings.pas
// 
// NAME
//   gb_strings.pas
//
// PURPOSE
//  Provides fundamental string manipulation routines that are not provided by 
//  Delphi.
//
// DESCRIPTION
//  This unit provides fundamental string manipulation routines that are not provided by 
//  Delphi.
//  
// AUTHOR
//  Birgit Hassler, CIRES / NOAA ESRL CSD
//
// COPYRIGHT 
//   THIS SOFTWARE AND ITS DOCUMENTATION ARE CONSIDERED TO BE IN THE PUBLIC
//   DOMAIN AND THUS ARE AVAILABLE FOR UNRESTRICTED PUBLIC USE. THEY ARE
//   FURNISHED "AS IS." THE AUTHORS, THE UNITED STATES GOVERNMENT, ITS
//   INSTRUMENTALITIES, OFFICERS, EMPLOYEES, AND AGENTS MAKE NO WARRANTY,
//   EXPRESS OR IMPLIED, AS TO THE USEFULNESS OF THE SOFTWARE AND
//   DOCUMENTATION FOR ANY PURPOSE. THEY ASSUME NO RESPONSIBILITY (1) FOR
//   THE USE OF THE SOFTWARE AND DOCUMENTATION; OR (2) TO PROVIDE TECHNICAL
//   SUPPORT TO USERS.
//
// REVISION HISTORY
//   6/16/2014 Initial Version prepared for NCDC.
// 
//@***** 


unit gb_strings;

{$I UNITDIREC.INC}

interface

uses
  Classes;

type
  TGBStringList=class(TStringList)
    function YesNo(NameStr:string):boolean;
  end;
  TCharSet=set of AnsiChar;

{Here is an alphabetical list of the routines available in this unit.}
function BackBuffer(St:string;Ch:char;Num:byte):string;
function BooleanToYesNo(Val,UCase:boolean):string;
function CharCount(St:string;Ch:char):byte;
function EraseEndChar(St:string;Ch:char):string;
function EraseFrontAndEndChar(St:string;Ch:char):string;
function EraseFrontChar(St:string;Ch:char):string;
function FrontBuffer(St:string;Ch:char;Num:byte):string;
function GetWord(St:string;Ch:char;Num:word;IgnoreEmptyStrings:boolean=true):string;
function InsertHour(PathStr,HourStr:string):string;
function InsertDate(PathStr,DateStr:string):string;
function InsertMonth(PathStr,MonthStr:string):string;
function InsertDay(PathStr,DayStr:string):string;
function InsertYear(PathStr,YearStr:string):string;
function InsertLevel(PathStr,LevelStr:string):string;
function IntToStrF(X:integer;NumDigits:byte;ReplaceSpaces:boolean):string;
function LocateChar(St:string;Ch:char;Start:byte):byte;
function Replace(St:string;ChOld,ChNew:char):string;
function ReplaceChars(St,OldSubStr,NewSubStr:string):string;
function ReverseStr(St:string):string;
function WordCount(St:string;Ch:char):word;
function YesNoToBoolean(St:string):boolean;

{$IFDEF DELPHI2006}
function CharInSet(Ch:char;const TestSet:TCharSet):boolean;
{$ENDIF}

implementation

uses
  sysutils;

function TGBStringList.YesNo(NameStr:string):boolean;
begin
  Result:=YesNoToBoolean(Values[NameStr]);
end; {End of TGBStringList.YesNo.}

function BackBuffer(St:string;Ch:char;Num:byte):string;
{------------------------------------------------------------------------------}
{ This routine adds character `Ch' to the end of string `St', until the length }
{ of `St' is equal to `Num'. If Num is less than length of St, the string is   }
{ unaltered.                                                                   }
{------------------------------------------------------------------------------}
begin
  while length(St)<Num do St:=St+Ch;
  BackBuffer:=St;
end; {End of BackBuffer.}

function BooleanToYesNo(Val,UCase:boolean):string;
{------------------------------------------------------------------------------}
{ This function returns 'YES' or 'yes' if Val is true, otherwise it returns    }
{ 'NO' or 'no'. The case of the returned string is set by UCase.               }
{------------------------------------------------------------------------------}
begin
  if Val then
    if UCase then BooleanToYesNo:='YES' else BooleanToYesNo:='yes'
  else if UCase then BooleanToYesNo:='NO' else BooleanToYesNo:='no';
end; {End of BooleanToYesNo.}

function CharCount(st:string;ch:char):byte;
{------------------------------------------------------------------------------}
{ This routine returns the number of times the character `Ch' appears in the   }
{ string `St'.                                                                 }
{------------------------------------------------------------------------------}
var
  I:byte; {Loop control variable.}
begin
  Result:=0;
  for I:=1 to length(St) do if St[I]=Ch then inc(Result);
end; {End of CharCount.}

function EraseEndChar(St:string;Ch:char):string;
{-----------------------------------------------------------------------------}
{ This function accepts a string and removes characters from the end, speci-  }
{ fied by `Ch', until a different character is encountered. If the last       }
{ character in the string does not match `Ch', the string is not changed. The }
{ modified/unmodified string is returned.                                     }
{-----------------------------------------------------------------------------}
begin
  while ((length(St)<>0) and  (St[length(St)]=Ch)) do
    St:=copy(St,1,length(St)-1);
  EraseEndChar:=St;
end; {End of EraseEndChar.}

function EraseFrontChar(St:string;Ch:char):string;
{-----------------------------------------------------------------------------}
{ This routine start at the beginning of the string `St' and removes consec-  }
{ utive characters specified by `Ch' until a different character is encount-  }
{ ered. If the first character in the string does not match `Ch', the string  }
{ is not changed. The modified/unmodified string is returned.                 }
{-----------------------------------------------------------------------------}
begin
  while ((length(St)<>0) and (St[1]=Ch)) do St:=copy(St,2,length(St));
  EraseFrontChar:=St;
end; {End of EraseFrontChar.}

function EraseFrontAndEndChar(St:string;Ch:char):string;
{-----------------------------------------------------------------------------}
{ This routine erases characters specified by `St' from the front and end of  }
{ the string specified by `St'. The modified string is returned.              }
{-----------------------------------------------------------------------------}
begin
  while ((length(St)<>0) and (St[1]=Ch)) do St:=copy(St,2,length(St));
  while ((length(St)<>0)  and (St[length(St)]=Ch)) do St:=copy(St,1,length(St)-1);
  EraseFrontAndEndChar:=St;
end; {End of EraseFrontAndEndChar.}

function FrontBuffer(St:string;Ch:char;Num:byte):string;
{-----------------------------------------------------------------------------}
{ This routine adds character `Ch' to the front of string `St' until the      }
{ length of St is equal to Num. If Num is less than the length of St, the     }
{ string is not altered.                                                      }
{-----------------------------------------------------------------------------}
begin
  while length(St)<Num do St:=Ch+St;
  FrontBuffer:=St;
end; {End of FrontBuffer.}

function GetWord(St:string;Ch:char;Num:word;IgnoreEmptyStrings:boolean=true):string;
{-----------------------------------------------------------------------------}
{ This routine returns the Num'th word in a string (numbering left to right), }
{ where a word is defined as any text separated by the character `Ch'.        }
{-----------------------------------------------------------------------------}
var
  Start,
  Count,
  I,J:integer;
begin
  Count:=1;
  Start:=1;
  J:=length(St);
  Result:='';
  for I:=1 to J do
  if St[I]=Ch then
  begin
    if ((I=Start) and IgnoreEmptyStrings) then inc(Start) else
    if Count=Num then
    begin
      Result:=copy(St,Start,I-Start);
      Exit;
    end
    else
    begin
      inc(Count);
      if Count>Num then Exit;
      Start:=I+1;
    end;
  end;
  if Count=Num then Result:=copy(St,Start,MaxInt);
end; {End of GetWord.}

function InsertHour(PathStr,HourStr:string):string;
var
  X:word;
begin
  while pos('HH',PathStr)>0 do
  begin
    X:=pos('HH',PathStr);
    Delete(PathStr,X,2);
    Insert(HourStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function InsertDate(PathStr,DateStr:string):string;
var
  X:word;
begin
  while pos('DD',PathStr)>0 do
  begin
    X:=pos('DD',PathStr);
    Delete(PathStr,X,2);
    Insert(DateStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function InsertMonth(PathStr,MonthStr:string):string;
var
  X:word;
begin
  while pos('MM',PathStr)>0 do
  begin
    X:=pos('MM',PathStr);
    Delete(PathStr,X,2);
    Insert(MonthStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function InsertDay(PathStr,DayStr:string):string;
var
  X:word;
begin
  while pos('DDD',PathStr)>0 do
  begin
    X:=pos('DDD',PathStr);
    Delete(PathStr,X,3);
    Insert(DayStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function InsertYear(PathStr,YearStr:string):string;
var
  X:word;
begin
  while pos('YYYY',PathStr)>0 do
  begin
    X:=pos('YYYY',PathStr);
    Delete(PathStr,X,4);
    Insert(YearStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function InsertLevel(PathStr,LevelStr:string):string;
var
  X:word;
begin
  while pos('LLLL',PathStr)>0 do
  begin
    X:=pos('LLLL',PathStr);
    Delete(PathStr,X,4);
    Insert(LevelStr,PathStr,X);
  end;
  Result:=PathStr;
end;

function IntToStrF(X:integer;NumDigits:byte;ReplaceSpaces:boolean):string;
begin
  str(X:NumDigits,Result);
  if ReplaceSpaces then Result:=Replace(Result,' ','0');
end;

function LocateChar(St:string;Ch:char;Start:byte):byte;
{-----------------------------------------------------------------------------}
{ This routine searches through a string specified by `St', starting at the   }
{ character numbered by `Start', until it finds the character matching `Ch'.  }
{ The position of `Ch' in the string is returned. If `Start' exceeds the      }
{ string's length, the function returns the value of `Start'. If the          }
{ character is not found, the string's length+1 is returned. This is very     }
{ useful for some purposes, see the GetWord procedure defined in this unit    }
{ for an application of this result when the word requested is the last on    }
{ the line.                                                                   }
{-----------------------------------------------------------------------------}
begin
  while (St[Start]<>Ch) and (Start<=length(St)) do inc(Start);
  LocateChar:=Start;
end; {End of LocateChar.}

function Replace(St:string;ChOld,ChNew:char):string;
{-----------------------------------------------------------------------------}
{ This routine replaces all instances of `ChOld' in string `St' with the      }
{ character `ChNew'.                                                          }
{-----------------------------------------------------------------------------}
var
  I:cardinal; {Loop control variable.}
begin
  {$WARNINGS OFF}
  for I:=1 to length(St) do
    if St[I]=ChOld then St[I]:=ChNew;
  Replace:=St;
  {$WARNINGS ON}
end; {End of Replace.}

function ReplaceChars(St,OldSubStr,NewSubStr:string):string;
{------------------------------------------------------------------------------}
{ This function replaces all instance of 'OldSubStr' in 'St' with 'NewSubStr'. }
{------------------------------------------------------------------------------}
begin
  Result:='';
  while pos(OldSubStr,St)<>0 do
  begin
    Result:=Result+copy(St,1,pos(OldSubStr,St)-1)+NewSubStr;
    St:=copy(St,pos(OldSubStr,St)+length(OldSubStr),length(St));
  end;
  Result:=Result+St;
end; {End of ReplaceChars.}

function ReverseStr(St:string):string;
{------------------------------------------------------------------------------}
{ This function returns the reverse of the string specified by St.             }
{------------------------------------------------------------------------------}
var
  I:word;
begin
  Result:='';
  for I:=1 to length(St) do Result:=Result+St[length(St)-I+1]
end; {End of ReverseStr.}

function WordCount(St:string;Ch:char):word;
{------------------------------------------------------------------------------}
{ This function counts the number of words in the string passed via 'St' where }
{ a word is seperated from the previous and next words by 'Ch' characters.     }
{------------------------------------------------------------------------------}
begin
  Result:=0;
  while GetWord(St,Ch,Result+1)<>'' do inc(Result);
end; {End of WordCount.}

function YesNoToBoolean(St:string):boolean;
{------------------------------------------------------------------------------}
{ If the uppercase of string passed to this function is 'YES' the function     }
{ returns true, otherwise the function returns false.                          }
{------------------------------------------------------------------------------}
begin
  if Uppercase(St)='YES' then Result:=true else Result:=false;
end; {End of YesNoToBoolean.}

{$IFDEF DELPHI2006}
function CharInSet(Ch:char;const TestSet:TCharSet):boolean;
begin
  Result:=Ch in TestSet;
end;
{$ENDIF}

{$WARNINGS OFF}
end. {End of gb_strings unit.}
{$WARNINGS ON}

